/*
Gargoyle components (dlcore.js) copyright (c) 2007 - 2011 Henrik Bechmann, bechmann.ca, 
	Bechmann Software Services, Toronto Canada, all rights reserved.
website: www.dlml.org 
licence: BSD
dl = "dynamic layout"
*/
/*
Dev tasks:
- Add debugger function to display errors such as JSON failure.
- dlaction: action of actions, such as selection (not necessarily rendering).
- Support "conditions" of state: eg. "open/closed" "ready/active/selected"
- explicitly delete actionpack?
- autogenerate critical errors (like JSON parsing) to first element in body - create this element if necessary, id="dlerror"
- error handling, for example there's a jquery restore error with object that has incorrect dltype.
*/
//=========================================================
//------------------[ PROTOTYPE FACTORY ]------------------
// args = {$extends, $include, $constructor, $prototype} // $include = mixin/delegate array
//=========================================================
DLPrototype = function(args) 
{
	args = args || {};
	var 
		$constructor = args.$constructor || function() {},
		$prototype = args.$prototype || {},
		$base =	args.$extends,
		$include = args.$include || [];
	var length = $include.length, mixin, slot;
	for (var i = 0; i < length; i++) {
		mixin = $include[i];
		for (slot in mixin) $prototype[slot] = $.extend({},mixin[slot]);
	}
	// if there is no extension, then enhance constructor and exit
	if (!$base) {
		$constructor.prototype = $prototype;
		$constructor.prototype.constructor = $constructor; // restore constructor
		return $constructor;
	} 
	else { //if there is an extension, apply extension and inheritance
		var dlconstructor, surrogateconstructor, prototypeObject;
		// create closure for dlconstructor, with merged slots at create time
		dlconstructor = function() 
		{
			// apply future arguments to base and new object constructors
			$base.apply(this,arguments);
			$constructor.apply(this,arguments);
		}
		// create surrogate constructor for inheritance; inherit instanceof
		surrogateconstructor = function () {}
		surrogateconstructor.prototype = $base.prototype; // lose any constructor code
		prototypeObject = new surrogateconstructor();
		// eclipse $base properties, anticipate extend...
		for (var slot in $prototype) 
			prototypeObject[slot]=$prototype[slot];
		// extend local copy...
		prototypeObject=dlservices$extend(prototypeObject,$base.prototype); 
		// for ancestor method calls...
		prototypeObject.base=$base.prototype; 
		// set inheritance...
		dlconstructor.prototype = prototypeObject;
		// establish correct dlconstructor...
		dlconstructor.prototype.constructor = dlconstructor;
		dlconstructor.isDLPrototype = true;
		return dlconstructor;
	}
}
//=========================================================
//-----------------------[ NATIVE PROTOTYPES]--------------
//=========================================================
// standard prototypes for introspection 
// generic functionality: action subscription and notification
DLNative = DLPrototype({
$constructor: function() {
	this.subscriptions = {};
},
$prototype:{
	// subscribe to action
	subscribe:function (action,object) {
		var subscriptions = this.subscriptions;
		subscriptions[action] || (subscriptions[action] = []);
		subscriptions[action].push(object);
	},
	// unsubscribe from action
	unsubscribe:function (action,object) {
		try {
			this.subscriptions[action].removeItem(object);
		} catch(e) {}
	},
	// notify subscribers of action
	notify:function (action,data) { // subcriber objects require act(action) method
		data = data || {};
		var subscribers = this.subscriptions[action];
		if (subscribers) {
			data.notifier = this;
			var length = subscribers.length;
			for (var i=0; i<length; i++) {
				var subscriber = subscribers[i];
				subscriber.act(action, data);
			};
		}
	},
	// stub for act, should be overloaded
	act:function (action, data) {	
		return this;
	}
}
});
// register module prototypes with DLManager; provide receptacle for platform list
DLModule = DLPrototype({
$extends:DLNative,
$constructor: function () {
	this.platforms = [];
	DLManager.register(this);
},
$prototype: {
	register: function (list) {
		(list instanceof Array) || (list = [list]);
		var module = this;
		var length = list.length;
		for (var i=0; i<length; i++) {
			var proto = list[i];
			proto.prototype.module = module; // instances have access to module
			DLManager.prototypes[proto.prototype.type] = proto;
		};
	}
}});
// state processing: setstate, changestate, renderstate, render
DLObject = DLPrototype({
$extends:DLNative,
$constructor: function () {
	this.flags = {};
},
$prototype:{
	setstate: function (state) { // set lastdlstate from dlstate, and dlstate from argument
		if (this.dlstate !== state) {
			this.lastdlstate = this.dlstate;
			this.dlstate = state;
			this.statewaschanged=true;
		}
		return this;
	},
	changestate: function(action) { // find next state name from action as index; setstate
		var state = this.dlstate, newstate;
		this.statewaschanged = false; // for renderstate
		try {
			newstate = this.transitions[action][state];
		} catch(e) {}
		if (newstate) this.setstate(newstate);
		return this;
	},
	renderstate: function (targetstate) { // render to target state, as passed or previously set
		if (targetstate || this.statewaschanged) {
			var state = (targetstate || this.dlstate)
			if (state && this.states[state] && this.states[state].index) {
				var index = this.states[state].index;
				this.render(index);
				this.statewaschanged = false;
			}
		}
		return this;
	},
	render: function (index) { 
		var htmlelement, renderargs;
		// capture default element
		var frame = this.elements.frame; // !this.elements = fatal error
		// must have default element to which to assign styling
		if (!frame) return this;
		// obtain styling data from index, as passed, or default
		var args = (index || this.index || {});
		for (var htmlindex in args) {
			htmlelement = this.elements[htmlindex] || frame; // target element for styling
			renderargs = args[htmlindex] || {}; // styling data
			dlobject$renderelement(htmlelement,renderargs);
		}
		return this;
	}
}
});
//private
dlobject$renderelement = function (htmlelement,renderargs){
	var $element = $(htmlelement);
	$element.stop(true,true); // anticipate reset of animation queue
	var styles = renderargs.styles;
	var classes = renderargs.classes;
	var animations = renderargs.animations;
	var effects = renderargs.effects;
	var attributes = renderargs.attributes;
	var triggers = renderargs.triggers;
	if (styles) {
		if (styles instanceof Array)
			$.fn.css.apply($element,styles);
		else
			$element.css(styles); // must be map
	}
	if (classes) {
		if (classes.removeClass !== undefined) {
			if (!classes.removeClass)
				$element.removeClass();
			else
				$element.removeClass(classes.removeClass);
		}
		if (classes.addClass !== undefined)
			$element.addClass(classes.addClass);
	}
	if (animations) {
		$.fn.animate.apply($element,animations);
	}
	if (attributes) {
		$.each(attributes,function(index,value) {
			$element.attr(index,value);
		});
	}
	if (effects && DLData.lists.effects) {
		for (var effect in effects) {
			if (DLData.lists.effects.inArray(effect)) {
				var args = effects[effect];
				(args instanceof Array) || (args = [args]);
				$.fn[effect].apply($element,args);
			}
		}
	}
	if (triggers) {
		for (var trigger in triggers) {
			if (this.triggers && this.triggers[trigger])
				this.triggers[trigger].apply(htmlelement,triggers[trigger]);
		}
	}
}
// container and platform for introspection
DLContainer = DLPrototype(
{$extends:DLObject,
$constructor: function () {
	this.components = {};
},
$prototype:{
	registercomponent: function (index, component) {
		this.components[index] = component;
	}
}
});

DLPlatform = DLPrototype(
{$extends:DLContainer
});

//=========================================================
//----------------------[action MODEL]--------------------
// *** needs work
//=========================================================
// new DLAction generated for notifications
DLAction = function (name,properties)
{
	this.name = name;
	var propertyobject = (properties ||{});
	var object = this;
	$.each(propertyobject,function(key,value) {
		object[key] = value;
	});
}
// generic event function factory, returned from closure to accommodate actionname returned in action object
// must be attached to outer element (frame)
function DLEventTerminal(action,data) // returns false, therefore stops bubbling
{
	data || (data = {});
	return function (event) {
			data.event = event;
			DLManager.objects[$(this).attr("dlobjectid")]
				.act(action,new DLAction(action,data));
			return false;
		}
}

function DLEventBubble(action,data) // continues to bubble
{
	data || (data = {});
	return function (event) {
			data.event = event;
			DLManager.objects[$(this).attr("dlobjectid")]
				.act(action, new DLAction(action,data));
		}
}
function DLCallTrigger(action,data){
	var self = this;
	data || (data = {});
	return function (event) {
		data.event = event;
		self[action](data);
	}
}
function DLActionTrigger(action,data) // can be called: DLActionTrigger.call(this,action);
{
	var self = this;
	data || (data = {});
	return function (event) {
			event.preventDefault();
			event.stopPropagation();
			event.stopImmediatePropagation();
			data.event = event;
			self[action](data);
			return false;
		}
}
//==============================================================
//---------------[ FUNCTIONAL ARRAY EXTENSIONS ]----------------
//==============================================================

Array.prototype.indexOf = function (value) 
{
	var length = this.length;
	for (var i=0; i<length; i++) 
		if (this[i]===value) return i;
	return -1;
}

Array.prototype.inArray = function (value) 
{
	return this.indexOf(value)>=0;
}

Array.prototype.removeItem = function (item) 
{
	var index =	this.indexOf(item);
	if (index>=0) this.splice(index,1);
}

Array.prototype.shallowCopy = function () 
{
	var newarray = [], length = this.length;
	for (var i=0; i<length; i++) 
		newarray[i]=this[i];
	return newarray;
}

//==============================================================================
//-------------[ FUNCTION LIBRARY ADDITIONS FOR INSTROSPECTION]-----------------
//==============================================================================

//returns object, element, array, null, or native type...
function typeOfVar(value) 
{
	var type = typeof value;
	if (type == 'object') {
		if (value) {
			if (value.nodeName) 
				type = 'element';
			else 
				if (value instanceof Array) type = 'array';
		} else 
			type = 'null';
	}
	return type;
}

//returns element, text, whitespace, or document if exists
function typeOfElement(element) 
{
	if (!element) return;
	switch (element.nodeType) {
		case 1: return 'element';
		case 3: return 
				element.nodeValue.match(/\S/)
					? 'text'
					: 'whitespace';
		case 9: return 'document';
	}
}

//returns dlobject, dlmodule, dlprototype, function, or object if exists
function typeOfObject(object) 
{
	if (!object) return;
	return 
		(object instanceof DLObject)
			? "dlobject"
			: object instanceof DLModule
				? "dlmodule"
				: $.isFunction(object) && object.isDLPrototype
					? "dlprototype"
					: $.isFunction(object)
						? "function"
						: "object";
}

//=======================================================================
// ----------------------------[ SUPPORT ]------------------------------
//=======================================================================
//DLSupport (chain of command mixins - $include for DLPrototype: drag, drop, select, sort, ajax)
// implemented as prototpe mixin's (include array)
/*
pattern:
{namespace:{ // eg. "drag"
	fn: function(widget,args) { // widget is caller's this
		var delegate = this; // namespace object; place for vars
		...
	}
}
*/
DLSupport = {
	// object
	DropManager:function (){},
	// function
	getDropManager:function(){}, // singleton
	// prototypes
	Drag:{drag:{}},
	Drop:{drop:{}},
	Select:{select:{}}, // overlay dotted div highlight underlying selections, + Ctrl-click, Shift-click
	Sort:{sort:{}},
	Ajax:{ajax:{}},
	Resize:{resize:{}},
	Expose:{expose:{}},
	Sync:{sync:{}},
	Transfer:{transfer:{}} // doubleclick between trays
}

//=======================================================================
// ----------------------------[ SERVICES ]------------------------------
//=======================================================================
DLServices = {
	clone:dlservices$clone, //(object)
	update:dlservices$update, // (object, update) update precedence
	extend:dlservices$extend, // (object, extension) object precedence
	encodehtml:dlservices$encodehtml,
	initializechildren:dlservices$initializechildren, // (object)
	initializebranch:dlservices$initializebranch // (object)
}
// initialize children, not self
function dlservices$initializechildren(object)
{
	var children = object.children;
	if (!children) return;
	var length = children.length;
	for (var i = 0; i < length; i++)
		DLServices.initializebranch(children[i]);
}
// initialize self, then children
function dlservices$initializebranch(object)
{
	if (!object.flags.initialized) {
		object.initialize && object.initialize();
		object.flags.initialized = true;
		object.render(); // to assign attributes for jquery tools selector, esp. history
	}
	dlservices$initializechildren(object);
}
// deep copy
function dlservices$clone (object)
{
	var type = typeOfVar(object), item, i, value, key;
	if (type=='array') {
		item = [];
		$.each(object,function (key,value) {
			item.push(dlservices$clone(value));
		});
	} 
	else if (type=='object') {
		item = {};
		$.each(object,function (key,value) {
			item[key] = dlservices$clone(value);
		});
	} 
	else 
		item=object;
	return item;
}

// update properties take precedence...
// if update is an array and update.immutable then a clone of update is returned
function dlservices$update (object, update)
{
	if (update && update.immutable) {
		update = DLServices.clone(update);
		delete update.immutable;
		return update;
	}
	var sourcetype = typeOfVar(update);
	if (sourcetype=='array') {
		((object===undefined) || (typeOfVar(object)!='array')) && (object=[]);
		$.each(update,function (key,value) {
			if (object.indexOf(value) ==-1) object.push(value);
		});
	}
	else if (sourcetype=='object') {
		((object===undefined) || (typeOfVar(object)!='object')) && (object={});
		$.each(update, function (key, value) {
			if (object[key] && (!(object.hasOwnProperty(key)))) 
				object[key] = dlservices$clone(object[key]);
			object[key]=dlservices$update(object[key],value);
			});
	}
	else 
		object=update;
	return object;
}

// object properties takes precedence...
// if extension is an object and extension.immutable then a clone of extension is returned.
function dlservices$extend (object, extension)
{
	var sourcetype = typeOfVar(extension);
	if (sourcetype=='array') {
		if (object===undefined) {
			object=dlservices$clone(extension);
			return object;
		}
		if (typeOfVar(object)!='array') return object;
		if (extension && extension.immutable) {
			if (extension.length == 0) {
				extension = DLServices.clone(extension);
				delete extension.immutable;
				return object = extension;
			} else return object;
		}
		$.each(extension, function (key,value) {
			if (object.indexOf(value)==-1) object.push(value);
		});
	} 
	else if (sourcetype=='object') {
		if (object===undefined) {
			object=dlservices$clone(extension);
			return object;
		}
		if (typeOfVar(object)!='object') return object;
		if (extension && extension.immutable) {
			if ($.isEmptyObject(object)) {
				extension = DLServices.clone(extension);
				delete extension.immutable;
				return object = extension;
			} else return object;
		}
		$.each(extension, function (key, value) {
			if (object[key]===undefined) 
				object[key]=dlservices$clone(value);
			else {
				if (object[key] && (!(object.hasOwnProperty(key)))) 
					object[key] = dlservices$clone(object[key]);
				object[key]=dlservices$extend(object[key],value);
			}
		});
	} 
	else {
		if (object===undefined) object=extension;
	}
	return object;
}

function dlservices$encodehtml(str) {
 str=str.toString();
 str = str.replace(/\&/g,"&amp;");
 str = str.replace(/\</g,"&lt;");
 str = str.replace(/\>/g,"&gt;");
 str = str.replace(/\"/g,"&quot;");
 return str;
}


//=======================================================================
// ----------------------------[ PARSER LIBRARY ]-------------------------
//=======================================================================
// *** add parsers for setstate and setmode??
// parsers are .call'ed, so passed 'this'; 
// they return a value as output based on a string argument as input. 
// Side effects (like this.property = ...) are allowed.
DLParse = {
	//parsers are called with .call(string), therefore 'this' var (calling object) is available within parsers
	"boolean":dlparse$boolean,
	"date":dlparse$date,
	"number":dlparse$number,
	//parseInteger
	"dlstyle":dlparse$renderfunction('styles'),//dlparse$dlstyle, // assigned to this.index.index.styles
	"dlclass":dlparse$renderfunction('classes'), // assigned to this.index[index].classes
	"dleffect":dlparse$renderfunction('effects'), // assigned to this.index[index].effects
	"dlanimation":dlparse$renderfunction('animations'), // assigned to this.index[index].animations
	"dlattribute":dlparse$renderfunction('attributes'), // // assigned to this.index[index].attributes
	"dltrigger":dlparse$renderfunction('triggers'), // assigned to this.index[index].triggers
	"dlaction":dlparse$dlaction, // assigned to this.actions[action]
	"dloptions":dlparse$dloptions, // assigned to this.options
	jsonproperty:dlparse$jsonproperty // (targetpropertyname)
}
// create closures for render parsers: dlstyle etc.
function dlparse$renderfunction(property) // assigns parsed string/object to property
{
	return function (string) {
		return dlparse$parseJSONrenderproperty.call(this,property,string);
	}
}

//common object attribute parsers...
function dlparse$boolean (string) 
{
	if (string == "true" || string == "on" || string=="yes") return true; 
	else return false;
}

function dlparse$number (string) 
{
	return +string;
}

function dlparse$date (string) 
{
	return new Date(string);
}
// private
function dlparse$parseJSONrenderproperty(rendertype,string)
{
	var index,object;
	if (!string) return string;
	try {
		object = JSON.parse(string);
	}
	catch (e) {
//		debugger; // should automatically report errors
		return string;
	}
	if (!object.index) {
		var tempobject = object;
		object = {index:{frame:tempobject}};
	}
	this.index || (this.index = {});
	for (index in object.index) {
		this.index[index] || (this.index[index] = {});
		this.index[index][rendertype] = object.index[index];
	}
	return string;
}
function dlparse$jsonproperty(property)
{
	return function (string)
	{
		if (!string) return string;
		var json;
		try {json = JSON.parse(string);}
		catch (e) {return string;}
		if (typeof json === 'object') {
			this[property] || (this[property] = {});
			$.extend(this[property],json);
		} else {
			this[property] = json;
		}
		return string;
	}
}
function dlparse$dlaction(string)
{
	if (!string) return string;
	try {object = JSON.parse(string);}
	catch (e) {return string;}
	if (typeof json === 'object') {
		this.actions || (this.actions = {});
		$.extend(this.actions,object);
	}
	return string;
}

function dlparse$dloptions (string) {
	if (!this.setupoptions) this.setupoptions = {};
	var options = string.split(',');
	var object = this;
	var length = options.length;
	for (var i=0;i<length; i++) {
		var value = options[i];
		object.options[$.trim(value)]= true;
	};
	return string;
}
//=======================================================================
// ----------------------------[ DATA ]------------------------------
//=======================================================================

// self-referential structure requires new object
DLData = new (function () {
	// reference structures...
	// standard style metadata
	this.styleattributetree = 
		{
			ATTRIBUTES:
				['dlstyle','dlclass','dleffect','dlanimation','dlattribute','dltrigger'],
			PARSERS:{
					dlstyle:DLParse.dlstyle, 
					dlclass:DLParse.dlclass,
					dleffect:DLParse.dleffect, 
					dlanimation:DLParse.dlanimation,
					dlattribute:DLParse.dlattribute,
					dltrigger:DLParse.dltrigger
				}
		};
	this.stateattributetree = 
		DLServices.update(
			DLServices.update(
				DLServices.clone(this.styleattributetree),
				{ATTRIBUTES:['dlstate','dlmode']}
			),
			{
				SETTINGS: {
					states:{
						COLLECTIONS: {
							states: this.styleattributetree
						}
					}
				},
				COLLECTIONS: {
					states: this.styleattributetree
				}
			}
		);
	this.commonattributetree = DLServices.update(this.stateattributetree,
		{ATTRIBUTES:['dlinherits'],PARSERS:{dlinherits:dlparse$boolean}});
	this.defaultstyleattributetree = 
		{	SETTINGS:{
				defaults:{
					ATTRIBUTES:['dlapplyto'],
					COLLECTIONS:{
						defaults:this.styleattributetree
					}
				}
			},
			COLLECTIONS:{
				defaults:this.styleattributetree
			}
		};
	this.defaultstateattributetree = 
		{	SETTINGS:{
				defaults:{
					ATTRIBUTES:['dlapplyto'],
					COLLECTIONS:{
						defaults:this.stateattributetree
					}
				}
			},
			COLLECTIONS:{
				defaults:this.stateattributetree
			}
		};
	this.lists = [];
	this.register = function (listname,members) {
		if (!this.lists[listname]) this.lists[listname]=[];
		this.lists[listname] = this.lists[listname].concat(members);
	}
})();

//=========================================================
//-----------------------[ DLML OBJECT ]-------------------
//=========================================================

//dlml: dynamic layout markup language
var dlml =  
{
	xmlelements:
		{
			//parameters for named modules
			module:"dl:module", 
			//anonymous placeholders for dhtml structures
			object:"dl:widget", 
			//typed sets of attributes; compound properties
			settings:"dl:settings",
			//user-named, typed sets of attributes (sets of typed, named properties)
			marker:"dl:marker"
		}
};
dlml.testtag = document.createElement('dl:widget');
if (dlml.testtag.tagName.toLowerCase()=='widget') {// ie
	dlml.xmltagnames = 
		{
			module:"module", 
			object:"widget", 
			settings:"settings",
			marker:"marker"
		}
} else {
	dlml.xmltagnames = 
		{
			module:"dl:module", 
			object:"dl:widget", 
			settings:"dl:settings",
			marker:"dl:marker"
		}
}
delete dlml.testtag;

//=========================================================
//---------[ Dynamic Layout Manager: DLManager ]-----------
//=========================================================

DLManager = new (DLPrototype({
$extends:DLNative,
$prototype:{
	register:dlmanager$register, // register module, from DLModule
	//=========================================
	//methods
	bind:dlmanager$bind, // (eventname,functionref)
	unbind:dlmanager$unbind, // (eventname,functionref)
	setdatamodel:dlmanager$setdatamodel, // (context)
	destroyobject:dlmanager$destroyobject, // (dlobject)
	act:dlmanager$act, // (action, data)
	actions:{ // standard action lookup, by actions[action]
		ready:dlmanager$ready // (action)
	},
	//=========================================
	//data properties
	// heaps:
	modules: {}, 
	prototypes: {}, 
	objects: [], 
	//dlobjects with no dlobject parents
	trees: [], 
	//next addition to objects[]
	objectindex: 0, 
	freeobjectindexes: [] // for re-use, in case of thrashing (ajax)
}}))();

// ================[ register module]=======================
function dlmanager$register(module)
{
	DLManager.modules[module.name]=module;
}
//============[ Event and Action handling]=================

function dlmanager$bind(eventname,functionref)
{
	if (eventname=='ready') $(functionref)
	else $(window).bind(eventname,functionref);
}

function dlmanager$unbind(eventname,functionref)
{
	$(window).unbind(eventname,functionref);
}

function dlmanager$act (action, data)
{
	DLManager.actions[action](data);
	return this;
}

// bind common window events, objects can subscribe to "ready","load","unload" and "select"
// allow DLUpdate object if it exists to update prototypes before instances created
// set data model; 
// notify objects subscribers that "ready" state has been achieved
function dlmanager$ready(action)
{ 
	DLManager.bind("load",function(event) {
		var action = new DLAction("load",{event:event});
		DLManager.notify(action.name,action);
	});
	DLManager.bind("unload",function(event) {
		var action = new DLAction("unload",{event:event});
		DLManager.notify(action.name,action);
	});
	DLManager.bind("click",function(event) {
		var action = new DLAction("select",{event:event});
		DLManager.notify(action.name,action);
	});
	if (typeof(DLUpdate) != "undefined") 
		if (DLUpdate.update) DLUpdate.update();
	DLManager.setdatamodel(document);
	DLManager.notify(action.name,action);
}
//=================[ Create object environment ]====================

//update modules, then create objects, and render objects
function dlmanager$setdatamodel(rootelement) 
{
	var objects = [], modules = [], length, i;
	DLManager.newobjects = [];
	// initialize modules
	var $modules = rootelement.getElementsByTagName(dlml.xmltagnames.module);
	length = $modules.length;
	for (i=0; i<length; i++) modules[i] = $modules[i];
	for (i=0; i<length; i++) dlmanager$initializemodule(modules[i]);
	// create objects
	var $objects = rootelement.getElementsByTagName(dlml.xmltagnames.object);
	length = $objects.length;
	for (i=0; i<length; i++) objects[i] = $objects[i];
	for (i=0; i<length; i++) dlmanager$createobject(objects[i]);
	length = DLManager.newobjects.length;
	// initialize and render successfully created objects
	for (i=0; i<length; i++) dlmanager$initializeobject(DLManager.newobjects[i]);
	delete DLManager.newobjects;
}

//initialize module. Events: initialize 
dlmanager$initializemodule = function (element) 
{
	var modulename = element.getAttribute("dlname"),
		object = DLManager.modules[modulename];
	if (!object) return;
	
	var attributeobject = {};
	dlmanager$assignattributes(attributeobject,element,object.metadata);
	dlmanager$assignsettings(attributeobject,element,object.metadata);
	DLServices.update(object,attributeobject);

	$(element).remove();
	object.initialize && object.initialize();
}

dlmanager$createobject = function(xmlelement) 
{
	var object,	type = xmlelement.getAttribute("dltype");
	try {object = new DLManager.prototypes[type]();}
	catch (e) {return;}
	object.metadata || (object.metadata = {});
	if (xmlelement.id) 
		object.elementid = xmlelement.id;
	else {
		var id = xmlelement.getAttribute("dlid");
		id && (object.elementid = id);
	}
	object.xmlelement = xmlelement;
	dlmanager$registerobject(object);
	dlmanager$setrelations(object);
	dlmanager$setdefaults(object);
	
	var attributeobject = {};
	dlmanager$assignattributes(attributeobject,xmlelement,object.metadata);
	dlmanager$assignsettings(attributeobject,xmlelement,object.metadata);
	DLServices.update(object,attributeobject); // local copies
	
	dlmanager$createhtmlobject(object);
}

dlmanager$initializeobject = function(object) 
{
	if (object.initialize && (!object.flags.initialized)) {
		object.initialize(); // may use renderstate()
		object.flags.initialized = true;
	}
	object.render(); // using styles, not states.
}

dlmanager$registerobject = function(object) 
{
	if (!object.metadata) return;
	var index = dlmanager$getobjectheapnumber();
	object.dlobjectid = index;
	DLManager.objects[index] = object;
	DLManager.newobjects.push(object);
}

dlmanager$getobjectheapnumber = function() 
{
	if (DLManager.freeobjectindexes.length>0)
		return DLManager.freeobjectindexes.pop();
	else
		return ++DLManager.objectindex;
}

// for upcoming remove object heap item (removedlobject)
dlmanager$deleteobjectheapitem = function(itemindex) 
{
	if (!DLManager.objects[itemindex]) return;
	delete DLManager.objects[itemindex]; // available for re-use
	DLManager.freeobjectindexes.push(itemindex);
}

/*set object relations: 
	.parent, 
	.parent.children[]
	.platform, 
	.container, 
*/
dlmanager$setrelations = function(object) 
{
	var parentnode,	objectid, parentobject;
	if (object instanceof DLPlatform) {
		object.module.platforms.push(object);
		object.platform=object; // own platform
	}
	parentnode=object.xmlelement.parentNode;
	// find first ancestor parent
	while (parentnode && parentnode.tagName && (parentnode.tagName.toLowerCase()!='body')) {
		objectid = parentnode.getAttribute("dlobjectid");
		if (objectid) break;
		parentnode=parentnode.parentNode;
	}
	// if objectid, then ancestor parent found
	if (objectid) { // assign parent, parent.children, container, and platform if possible
		parentobject = object.parent = DLManager.objects[objectid];
		parentobject.children || (parentobject.children=[]);
		parentobject.children.push(object);
		if (object.module===parentobject.module) { // same module
			if (!(object instanceof DLPlatform)) // might have been set for instance of DLPlatform  above
				 object.platform = parentobject.platform; // should always be set, if not set above
		}
		if (parentobject instanceof DLContainer) 
			object.container=parentobject;
		else if (parentobject.container) 
			object.container=parentobject.container;
		var dlindex = object.xmlelement.getAttribute('dlindex');
		if (dlindex && object.container) {
			object.dlindex = dlindex;
			object.container.registercomponent(dlindex,object);
		}
	}
	object.parent || DLManager.trees.push(object);
}

// looks through containers ancestors for default settings. Current container, 
//and more distant containers with default dlapplyto of "branch", are included, as well as platform
dlmanager$setdefaults = function(object) 
{
	if ((object instanceof DLModule) || (!object.container)) return;
	if (object.dlinherits === false) return;
	
	var type = object.type, typelist = [], testobject,
		container, platform, containerspan = 0,
		defaults, defaultset = {},	applyto;

	if (!type) return;
	
	typelist.push(type);
	testobject=object;
	// check inherited types as well...
	while (testobject.base) {
		testobject=testobject.base;
		testobject.type && typelist.push(testobject.type);
	}
	container=object.container;
	platform=object.platform;
	applyto = null;
	if (container && (container!==platform)) { // check container chain not including platform
		do {
			containerspan += 1;
			if (container.defaults) {
				var length = typelist.length;
				for (var i=0; i<length; i++) {
					type=typelist[i];
					if (container.defaults[type]) {
						defaults=container.defaults[type];
						if (defaults) {
							applyto = defaults.dlapplyto;
							if ((containerspan > 1) && (!applyto || (applyto && (applyto != "branch"))))
								continue;
							dlmanager$extenddefaults(defaultset,dlservices$clone(defaults));
						}
					}
				}
			}
			container=container.container;
		} while (container && (!(container instanceof DLPlatform)));
	}
	if (platform && platform.defaults) { // check platform
		var length = typelist.length, type;
		for (var i=0; i<length; i++) {
			type=typelist[i];
			if (platform.defaults[type]) {
				defaults = dlservices$clone(platform.defaults[type]);
				dlmanager$extenddefaults(defaultset,defaults);
			}
		}
	}
	DLServices.update(object,defaultset);
}

dlmanager$extenddefaults = function (object,defaults) 
{
	for (var slot in defaults) {
		//eclipse prototype structure
		if (object[slot]) { 
			if (!object.hasOwnProperty(slot)) 
				object[slot]=dlservices$clone(object[slot]);
			object[slot]=dlservices$extend(object[slot],defaults[slot]);
		} 
		else 
			object[slot]=defaults[slot];
	}
}

dlmanager$assignattributes = function (object,element,metadata)
{
	if (!metadata) return;
	var attributes = metadata.ATTRIBUTES; 
	if (!attributes) return;
	var parsers = metadata.PARSERS || {};
	var length = attributes.length;
	for (var i=0;i<length;i++) {
		var attributename = attributes[i];
		var attribute = element.getAttribute(attributename);
		if (attribute) {
			if (parsers[attributename]) 
				object[attributename] = parsers[attributename].call(object,attribute);
			else 
				object[attributename] = attribute;
		}
	};
}

dlmanager$assignsettings = function(object,element,metadata) 
{
	if (!metadata) return;
	var settings = metadata.SETTINGS || {};
	var child = element.firstChild, children = [];
	while (child) {
		child.tagName && (child.tagName.toLowerCase() == dlml.xmltagnames.settings) && (children.push(child));
		child = child.nextSibling;
	}
	var collections = metadata.COLLECTIONS || {};
	var length = children.length;
	for (var i=0; i<length; i++) {
		var node = children[i];
		var type, index;
		type = node.getAttribute("dltype");
		index = node.getAttribute("dlindex");
		if (!index) { // attribute settings
			if (type) {
				if (settings[type]) {
					if (!object[type]) 
						object[type]={};
					else 
						if (!object.hasOwnProperty(type)) // anticipate augmentation
							object[type] = dservicesl$clone(object[type]);
					dlmanager$assignattributes(object[type],node,settings[type]);
					dlmanager$assignsettings(object[type],node,settings[type]);
				}
			}
		}
		else { // collection settings
			var settingstype = element.getAttribute("dltype"); // inherit from parent dlsettings
			var collectionstype = type;
			if (!type) type = settingstype;
			if (index && type) {
				if (collections[type]) {
					if ((!collectionstype) && settingstype && (type == settingstype)) { // apply to parent type slot
						if (!object[index]) 
							object[index]={};
						else if (!object.hasOwnProperty(index)) //anticipate augmentation
							object[index] = dlservices$clone(object[index]);
					
						dlmanager$assignattributes(object[index],node,collections[type]);
						dlmanager$assignsettings(object[index],node,collections[type]);
					}
					else { // apply to called object type slot
						if (!object[type]) object[type]={};
						if (!object[type][index]) 
							object[type][index]={};
						else if (!object[type].hasOwnProperty(index)) //anticipate augmentation
							object[type][index] = dlservices$clone(object[type][index]);
					
						dlmanager$assignattributes(object[type][index],node,collections[type]);
						dlmanager$assignsettings(object[type][index],node,collections[type]);
					}
				}
			}
		}
		$(node).remove();
	}; // each
}

// dlindex="insertmarker" element causes all xmlelement contents to be inserted before that item.
dlmanager$createhtmlobject = function(object)
{
	var $objectelements = $(object.html), insertmarker;
	object.elements={};
	var frame = object.elements.frame = $objectelements.get(0);
	if (!frame) return;
	frame.setAttribute("dlobjectid",object.dlobjectid.toString());
	frame.setAttribute("dltype",object.type);
	object.elementid && (frame.id=object.elementid);
	var $elements = $objectelements.add($('*',frame));
	object.elements.liner = $elements.get($elements.length - 1); // can be overridden next
	$elements.each(function (key,element)	{
		var index = element.getAttribute("dlindex");
		index && (object.elements[$.trim(index)]=element);
	});
	var liner = object.elements.liner;
	if (insertmarker = object.elements.insertmarker) {
		while (object.xmlelement.firstChild) // transfer dlobject contents to html object
			liner.insertBefore(object.xmlelement.firstChild,insertmarker); // allow div with clear:both as last element
	} else {
		while (object.xmlelement.lastChild) // transfer dlobject contents to html object
			liner.insertBefore(object.xmlelement.lastChild,liner.firstChild); // allow div with clear:both as last element
	}
	object.xmlelement.parentNode.replaceChild(frame,object.xmlelement);
	delete object.xmlelement;
}

/*call object's custom destroy() method, and destroys the following: 
	.parent, 
	.parent.children[]
	.platform, 
	.container, 
*/
// *** add unsubscribe check
function dlmanager$destroyobject (dlobject) {
//	debugger;
	if (!(dlobject instanceof DLObject)) return;
	var children = dlobject.children;
	if (children) {
		var length = children.length;
		for (var i=0; i<length; i++) {
			dlmanager$destroyobject(children[i]);
		}
	}
	var $frame = $(dlobject.elements.frame);
	dlobject.destroy && dlobject.destroy();
	if (!dlobject.parent) {
		this.trees.removeItem(dlobject);
	} else {
		delete dlobject.parent;
	}
	if (dlobject.dlindex)
		delete dlobject.container.components[dlobject.dlindex];
	if (dlobject instanceof DLPlatform)
		dlobject.module.platforms.removeItem(dlobject);
	delete dlobject.children;
	delete dlobject.platform;
	delete dlobject.container;
	delete dlobject.elements;
	delete dlobject.module;
	dlmanager$deleteobjectheapitem(dlobject.dlobjectid);
	$frame.remove();
}
//=========================================================
//---------[ INTIALIZE ]-----------
//=========================================================

DLData.register("effects",["effect","fadeIn","fadeOut","slideDown","slideUp","show","hide","animate"]);

DLManager.bind("ready",function(event) {
		var action = new DLAction("ready"); // jQuery 'ready' event returns jQuery constructor instead of event, so event not passed on
		DLManager.act(action.name, action);
	}
);
