<?php 

namespace Muster;

use Kohana, Upload, Image, Validation;
use System;

defined('SYSPATH') or die('No direct script access.');
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net
/*
	Usage:
	$filer
		->import($_FILES['file'])
		->as_image()
		->targetbase(System::instance()->UploadBase)
		->to('subdirectory')
		->as_name('picname.ext')
		->go();
	Note: one operation at a time on one file at a time
	changes:
		auto-detect file type
		reject file name with blanks
		error_handling
		over_write flag
*/
class Filer {
	const FILETYPE_IMAGE = 0;
	const FILETYPE_XDOC = 1;
	
	const ACTION_IMPORT = 0;
	const ACTION_MODIFY = 1;
	// pending...
	const ACTION_DELETE = 2;
	const ACTION_COPY = 3;
	const ACTION_MOVE = 4;
	const ACTION_CREATE_VERSIONS = 5;
	
	const OPERATION_RESIZE = 1;
	
	protected static $_instance; // Filer instance
	
	/*
		$file_info in $_FILES single file format:
		$_file_info => array(
			'name' => NAME,
			'type' => TYPE,
			'size'=> SIZE,
			'tmp_name'=> TMP_NAME, 
			'error' => ERROR,
		)
	*/
	protected $_file_info; // format: $_FILES[$index]['name']...
	protected $_file_type; // one of constants above
	protected $_quality; // for jpeg
	
	protected $_action; // one of constants above
	
	protected $_root; // typically DOCROOT/public/, from config file 'filer'
	
	// source root subdirectory
	protected $_source_base; // eg. uploads/
	protected $_from; // source subdirectory, from (not including) source_base
	protected $_source_filebasename; // eg. picture.jpg
	/* pathinfo in pathinfo format
		$pathinfo = array(
			'dirname' => DIRNAME, // path
			'basename' => BASENAME, // filename . extension
			'extension' => EXTENSION, // if any
			'filename' => FILENAME, // basename without extension
		)
	*/
	protected $_source_pathinfo;
	protected $_source_filespec; // dirname . basename
	
	// target root subdirectory
	protected $_target_base;
	protected $_to; // targetbase subdirectory
	protected $_target_filebasename;
	protected $_target_pathinfo;
	protected $_target_filespec;
	protected $_replace_existing = FALSE;
	protected $_auto_convert = FALSE;
	
	protected $_config; // imported
	
	protected $_validation; // validation object, if used
	protected $_errors; // list array of message strings
	protected $_messages; // for create_versions
	
	public static function instance()
	{
		if (!self::$_instance)
			self::$_instance = new Filer();
		return self::$_instance;
	}
	
	public function __construct()
	{
		$config = $this->_config = (object) Kohana::$config->load('filer');
		$this->_root = System::instance()->PublicBase; //$config->root; // nothing can be uploaded above this directory
		$this->_quality = $config->default_quality;
	}
	public function reset()
	{
		$this->_file_info = NULL;
		$this->_file_type = NULL;
		$this->_action = NULL;
		$this->_operation = NULL;
		$this->_quality = $this->_config->default_quality;

		// leave source_base unchanged
		$this->_from = NULL; // path
		$this->_source_filebasename = NULL; // filename.extension
		$this->_source_pathinfo = NULL; // from pathinfo()
		$this->_source_filespec = NULL; // from.filebasename

		// leave target_base unchanged
		$this->_to = NULL; // path
		$this->_target_filebasename = NULL;
		$this->_target_pathinfo = NULL;
		$this->_target_filespec = NULL;

		// control data
		$this->_validation = NULL;
		$this->_errors = NULL;
		$this->_messages = NULL;
		$this->_auto_convert = FALSE;
		$this->_replace_existing = FALSE;
		return $this;
	}
	# actions - choose one
	public function import($file_info)
	{
		$this->_action = self::ACTION_IMPORT;
		$this->_file_info = $file_info;
		return $this;
	}
	public function modify($filespec)
	{
		$this->_action = self::ACTION_MODIFY;
		$this->_source_filespec = $filespec;
		$this->_target_filespec = NULL;
		return $this;
	}
	public function create_versions($filespec)
	{
		$this->_action = self::ACTION_CREATE_VERSIONS;
		$this->_source_filespec = $filespec;
		$this->_target_filespec = NULL;
		return $this;
	}
	public function delete($filespec)
	{
		$this->_action = self::ACTION_DELETE;
		$this->_source_filespec = $filespec;
		$this->_target_filespec = NULL;
		return $this;
	}
	public function copy($filespec)
	{
		$this->_action = self::ACTION_COPY;
		$this->_source_filespec = $filespec;
		$this->_target_filespec = NULL;
		return $this;
	}
	public function move($filespec)
	{
		$this->_action = self::ACTION_MOVE;
		$this->_source_filespec = $filespec;
		$this->_target_filespec = NULL;
		return $this;
	}
	# operations
	public function resize($width = NULL, $height = NULL, $master = NULL)
	{
		$this->_operation = self::OPERATION_RESIZE;
		$this->_arguments = compact('width','height','master');
		return $this;
	}
	# assert filetypes
	public function as_image()
	{
		$this->_file_type = self::FILETYPE_IMAGE;
		return $this;
	}
	public function as_docfile()
	{
		$this->_file_type = self::FILETYPE_XDOC;
		return $this;
	}
	public function get_filetype($filespec)
	{
//		$file_data = pathinfo($filespec);
//		if (!isset($file_data['extension'])) return FALSE;
//		$extension = $file_data['extension'];
		$extension = pathinfo($filespec,PATHINFO_EXTENSION);
		if (empty($extension)) return FALSE;
		$extension = strtolower($extension);
		$types = $this->_config->types;
		foreach ($types as $type) {
			if (in_array($extension,$type['extensions']))
				return $type['type_code'];
		}
		return FALSE;
	}
	# paths and property accessors
	// get or set both source and target base
	public function base($base = NULL)
	{
		if (is_null($base)) return $this->_source_base;
		$this->_source_base = $this->_target_base = $this->_assert_trailing_slash($base);
		return $this;
	}
	// get or set source base
	public function sourcebase($base = NULL)
	{
		if (is_null($base)) return $this->_source_base;
		$this->_source_base = $this->_assert_trailing_slash($base);
		return $this;
	}
	// get or set target base
	public function targetbase($base = NULL)
	{
		if (is_null($base)) return $this->_target_base;
		$this->_target_base = $this->_assert_trailing_slash($base);
		return $this;
	}
	public function as_name($filebasename = NULL)
	{
		return $this->filebasename($filebasename);
	}
	public function filename($filebasename = NULL)
	{
		return $this->filebasename($filebasename);
	}
	// get or set target file base name (name.ext)
	public function filebasename($filebasename = NULL)
	{
		if (is_null($filebasename)) return $this->_target_filebasename;
		$this->_target_filebasename = $filebasename;
		return $this;
	}
	// get or set target filespec (path.name.ext)
	public function as_filespec($filespec) {
		$this->filespec($filespec);
		return $this;
	}
	public function filespec($filespec = NULL)
	{
		if (is_null($filespec)) return $this->_target_filespec;
		$this->_target_filespec = $filespec;
		return $this;
	}
	public function filetype()
	{
		return $this->_file_type;
	}
	// get source filebasename
	public function source_filebasename($filebasename = NULL)
	{
		if (is_null($filebasename)) return $this->_source_filebasename;
		$this->_source_filebasename = $filebasename;
		return $this;
	}
	// get or set from (path)
	public function from($from = NULL)
	{
		if (is_null($from)) return $this->_from;
		$this->_from = $this->_assert_trailing_slash($from);
		return $this;
	}
	// get or set quality
	public function quality($quality = NULL)
	{
		if (is_null($quality)) return $this->_quality;
		$this->_quality = $quality;
		return $this;
	}
	// get or set to (path)
	public function subdirectory($to = NULL)
	{
		if (is_null($to)) return $this->_to;
		$this->_to = $this->_assert_trailing_slash($to);
		return $this;
	}
	public function to($to = NULL)
	{
		if (!is_null($to)) $this->subdirectory($to); // set subdirectory
		return $this; // in andy case
	}
	// get source filespec
	public function source_filespec()
	{
		return $this->_source_filespec;
	}
	// get target filespec
	public function target_filespec()
	{
		return $this->_target_filespec;
	}
	// get validation object
	public function validation()
	{
		return $this->_validation;
	}
	// alias for errors
	public function error_info()
	{
		return $this->errors();
	}
	// get errors
	public function errors()
	{
		return $this->_errors;
	}
	public function messages()
	{
		return $this->_messages;
	}
	# execution
	public function execute()
	{
		$action = $this->_action;
		switch ($action) {
			case self::ACTION_IMPORT: {
				return $this->_action_import();
				break;
			}
			case self::ACTION_MODIFY: {
				return $this->_action_modify();
				break;
			}
			case self::ACTION_CREATE_VERSIONS: {
				return $this->_action_create_versions();
				break;
			}
			// pending...
			case self::ACTION_DELETE: {
				return $this->_action_delete();
				break;
			}
			case self::ACTION_COPY: {
				return $this->_action_copy();
				break;
			}
			case self::ACTION_MOVE: {
				return $this->_action_move();
				break;
			}
			default: {
				return FALSE; // ** set error message
			}
		}
	}
	public function go()
	{
		return $this->execute();
	}
	public function replace_existing($value) 
	{
		if (!is_bool($value)) die ('must have boolean value for replace_existing in filer.php');
		$this->_replace_existing = $value;
		return $this;
	}
	public function auto_convert($value) 
	{
		if (!is_bool($value)) die ('must have boolean value for auto_convert in filer.php');
		$this->_auto_convert = $value;
		return $this;
	}
	// do import action
	protected function _action_create_versions()
	{
		$success = TRUE;
		$sourcefilespec = $this->_source_filespec;
		if (!is_file($sourcefilespec)) {
			$this->_errors[] = 'File does not exist';
			return FALSE;
		}
		$versions = (array) System::instance()->PictureSizes;
		$fileinfo = getimagesize($sourcefilespec);
		if (empty($fileinfo)) {
			$this->_errors[] = 'Could not get image information';
			return FALSE;
		}
		$originalwidth = $fileinfo[0];
		if (!$originalwidth) {
			$this->_errors[] = 'Could not get image width - not a known image type';
			return FALSE;
		}
		$pathinfo = pathinfo($sourcefilespec);
		if (empty($pathinfo['extension'])) {
			$this->_errors[] = 'File must have extension';
			return FALSE;
		}
		$filetype = $this->get_filetype($sourcefilespec);
		if ($filetype !== self::FILETYPE_IMAGE) {
			$this->_errors[] = 'File must be image type';
			return FALSE;
		}
		$leftpart = $pathinfo['dirname'].'/'.$pathinfo['filename'];
		$rightpart = '.'.$pathinfo['extension'];
		$errorlist = array();
		$message = '';
		foreach ($versions as $key => $value) {
			if (is_null($value)) continue;
			if ($originalwidth < $value) continue;
			$targetfilespec = $leftpart.'.'.$key.$rightpart;
			$this->reset();
			$result = $this->modify($sourcefilespec)->resize($value)->as_filespec($targetfilespec)->go();
			if (!$result) {
				if (is_array($this->_errors)) {
					$errorlist = array_merge($errorlist,$this->errors());
				}
				$success = FALSE;
			} else {
				$message .= ' ' . $key;
			}
		}
		$this->reset();
		if (!$success) {
			$this->_errors = $errorlist;
		}
		if (!empty($message)) {
			$this->_messages[] = 'versions created:' . $message;
		}
		return $success;
	}
	protected function _action_import()
	{
		$filename = $this->_file_info['name'];
		// discover filetype from extension
		$errorsubs = array(':filename' => $filename);
		$filetype = $this->get_filetype($filename);
		if ($filetype === FALSE) {
			$this->_errors[] = strtr(':filename unknown file type',$errorsubs);
			return FALSE;
		}
		if ((!is_null($this->_file_type)) and ($this->_file_type !== $filetype)) {
			$this->_errors[] = strtr(':filename wrong file type',$errorsubs);
			return FALSE;
		} elseif (is_null($this->_file_type)) {
			$this->_file_type = $filetype;
		}
		switch ($filetype) {
			case self::FILETYPE_IMAGE:
				$success = $this->_validate_image();
				break;
			case self::FILETYPE_XDOC:
				$success = $this->_validate_xdoc();
				break;
			default:
				die('unexpected failure to recognize filetype in filer.php');
		}
		if (!$success) return $success;
		
		$source_filename = $filename;
		$target_filename = $this->_target_filebasename;
		if (!$target_filename) {
			$target_filename = $source_filename;
		} else { // check extension validity
			$target_info = pathinfo($target_filename);
			$source_info = pathinfo($source_filename);
			if ((!isset($target_info['extension'])) and (isset($source_info['extension']))) {
				$target_info['extension'] = $source_info['extension'];
				$target_filename = $this->_target_filebasename = $target_info['filename'] . '.' . $target_info['extension'];
			}
//			$target = $target_info['filename'] . '.' . $source_info['extension']; // guarantee matching extensions
			if (strtolower($target_info['extension']) != strtolower($source_info['extension'])) {
				$this->_errors[] = strtr(':filename mismatched file extensions',$errorsubs);
				return FALSE;
			}
		}
//		$target = preg_replace('/[\[{ ]/','_',$target_filename);
		$target = strtr($target_filename,' ','_');
		if ($target != $target_filename) {
			if ($this->_auto_convert) {
				$this->_target_filebasename = $target_filename = $target;
			} else {
				$this->_errors[] = strtr(':filename invalid characters in file name',$errorsubs);
				return FALSE;
			}
		}
		$directory = $this->_root . $this->_target_base . $this->_to;
		$filespec = $this->_target_filespec = $directory . $target_filename;
		// assert directory
		if (!($success = is_dir($directory))) {
			$success = mkdir($directory,$this->_config->access_mode,TRUE); // true for recursive
		}
		// check for overwrite
		if ($success and (!$this->_replace_existing)) {
			$success = !is_file($filespec);
			if (!$success) {
				$this->_errors[] = strtr(':filename already exists; upload cancelled',$errorsubs);
			}
		}
		if ($success) {
			$filespec = $this->_target_filespec = Upload::save($this->_file_info, $target_filename, $directory);
			$success = (bool) $filespec;
			if (!$success) $this->_errors[] = strtr(Kohana::message('filer','failure'),$errorsubs);
		}
		return $success;
	}
	protected function _action_modify()
	{
		$filetype = $this->_file_type;
		switch ($filetype) {
			case self::FILETYPE_IMAGE: {
				$success = $this->_modify_image();
				break;
			}
			default: $success = FALSE;
		}
		return $success;
	}
	protected function _modify_image()
	{
		$source_filespec = $this->_source_filespec;
		$filename = pathinfo($source_filespec,PATHINFO_BASENAME);
		$target_filespec = $this->_target_filespec;
		if (!$target_filespec) 
			$target_filespec = $this->_target_filespec = $source_filespec;
		$operation = $this->_operation;
		extract($this->_arguments); // $width, $height, $master
		switch($operation) {
			case self::OPERATION_RESIZE:
				try {
					Image::factory($source_filespec)->resize($width,$height,$master)->save($target_filespec,$this->_quality);
				} catch(Exception $e) {
					$this->_errors[] = $filename . ': '. $e->getMessage();
					return FALSE;
				}
				break;
			default: return FALSE;
		}
		return TRUE;
	}
	protected function _action_delete()
	{
	}
	protected function _action_copy()
	{
	}
	protected function _action_move()
	{
	}
	# support
	protected function _assert_trailing_slash($string)
	{
		if (is_null($string) or ($string == '')) return $string;
		if (substr($string,0,1) == "/")
			$string = substr($string,1);
		if ($string == '' or $string == '/') return '';
		if (substr($string,strlen($this->_to)-1,1) != '/')
			$string .= '/';
		return $string;
	}
	protected function _validate_image()
	{
		$index = 'file';
		$file_data = $this->_file_info;
		$errorsubs = array(':filename' => $file_data['name']);
		$validation = $this->_validation = Validation::factory(array('file' => $file_data))
			->rule($index, 'Upload::not_empty')
			->rule($index, 'Upload::valid',array($file_data))
			->rule($index, 'Upload::type',array($file_data,$this->_config->types['image']['extensions']))
			->rule($index, 'Upload::size',array($file_data,$this->_config->types['image']['max_size']));
		$success = $validation->check();
		if (!$success) {
			$errors = $this->_validation->errors('filer');
			foreach ($errors as $error) {
				$this->_errors[] = strtr($error,$errorsubs);
			}
		}
		return $success;
	}
	protected function _validate_xdoc()
	{
		$index = 'file';
		$file_data = $this->_file_info;
		$errorsubs = array(':filename' => $file_data['name']);
		$validation = $this->_validation = Validation::factory(array('file' => $file_data))
			->rule($index, 'Upload::not_empty')
			->rule($index, 'Upload::valid',array($file_data))
			->rule($index, 'Upload::type',array($file_data,$this->_config->types['xdoc']['extensions']))
			->rule($index, 'Upload::size',array($file_data,$this->_config->types['xdoc']['max_size']));

		$success = $validation->check();
		if (!$success) {
			$errors = $this->_validation->errors('filer');
			foreach ($errors as $error) {
				$this->_errors[] = strtr($error,$errorsubs);
			}
		}
		return $success;
	}
}
