<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net
namespace Muster;

use StdClass;
use Kohana, Validation, Database, DB, Security;
use Access, Point, User;
use 
	Muster\Point\Block, 
	Muster\Point\Blocklist,
	Muster\Point\Blockselect, 
	Muster\Point\Meta, 
	Muster\Point\Data,
	Muster\Node;

defined('SYSPATH') or die('No direct script access.');
/*
	Frames hold the metadata and the whiteboard data for the current data operation.
	
	The format of each value depends on the state of processing. 
	Input defers formatting to the finishinput after validate process.
	
	*** Design: allow multiple constraints per point?
*/

class Gateway {
	# actions
	const RETRIEVE = 'retrieve';
	const OUTPUT = 'output';
	const INPUT = 'input'; // pause for validation
	const VALIDATE = 'validate';
	const RECOVERINPUT = 'recoverinput'; // specialized process when recovery required, after finishinput
	const FINISHINPUT = 'finishinput'; // after validation
	const STORE = 'store';
	# the post array in datapool->arrays[]
	const POST_ARRAY = 'post';

	// caller/controller
	public $datamodel;
	
	// transient values
	protected $_frame;
	protected $_frame_type; // dataframe, controlframe, containerframe, set by point::constant
	protected $_process_type; // one of the constants above: retrieve, output, input, validate, finishinput, store
	protected $_database; // index from gateway config
	protected $_dateformat; // from gateway config
	protected $_datetimeformat; // from gateway config
	protected $_errordictionaryindex;
	protected $_errormessageindex;
	protected $_input_default_recover_state;
	
	public function __construct($datamodel)
	{
		$this->datamodel = $datamodel;
		$config = (object) Kohana::$config->load('gateway');
		$this->_database = $config->database;
		$this->_dateformat = $config->dateformat;
		$this->_datetimeformat = $config->datetimeformat;
	}

	#===============[ CONFIGURATION ]===============

	# Select type of processing
	public function process_retrieve()
	{
		$this->_process_type = self::RETRIEVE;
		return $this;
	}
	public function process_output()
	{
		$this->_process_type = self::OUTPUT;
		return $this;
	}
	public function process_input($default_recover_state = NULL)
	{
		$this->_input_default_recover_state = $default_recover_state;
		$this->_process_type = self::INPUT;
		return $this;
	}
	public function process_validate()
	{
		$this->_process_type = self::VALIDATE;
		return $this;
	}
	public function process_finishinput()
	{
		$this->_process_type = self::FINISHINPUT;
		return $this;
	}
	public function process_recoverinput()
	{
		$this->_process_type = self::RECOVERINPUT;
		return $this;
	}
	public function process_store()
	{
		$this->_process_type = self::STORE;
		return $this;
	}
	public function errordictionaryindex($index)
	{
		$this->_errordictionaryindex = $index;
		return $this;
	}
	public function errormessageindex($file)
	{
		$this->_errormessageindex = $file;
		return $this;
	}
	// collect frame (data and metadata) to process
	public function with($frame)
	{
		$this->_frame = $frame;
		$this->_frame_type = $frame->type;
		return $this;
	}
	#===============[ DISPATCH ]===============
	// all operations are transient
	public function execute()
	{
		if (!($this->_frame and $this->_process_type)) return;
		$this->datamodel->errorstate = 0; //reset
		$method = '_process_' . strtolower($this->_frame_type);// controlframe, dataframe, containerframe
		$this->$method();
		$frame = $this->_frame;
		$this->_frame = NULL;
		$this->_frame_type = NULL;
		$this->_process_type = NULL;
		return $frame;
	}
	#========================================================================
	#-------------------------[ CONTROLFRAME ]-------------------------------
	#========================================================================
	// controller
	protected function _process_controlframe()
	{
		$frame = $this->_frame;
		$pointlist = &$frame->pointlist;
		$metadatalist = $frame->metadata;
		$process_type = $this->_process_type;
		switch ($process_type) {
			case self::INPUT: {
				$this->_input_control($metadatalist,$pointlist);
				break;
			}
		}
	}
	#========================= CONTROLFRAME INPUT ==================================
	protected function _input_control($metadatalist,&$pointlist)
	{
		foreach ($metadatalist as $pointname => $metapoint) {
			if (property_exists($metapoint,'constraint')) {
				$constraint = $metapoint->constraint;
				if (!$this->datamodel->constraints($constraint)) 
					continue;
			}
			if (property_exists($metapoint,'notconstraint')) {
				$constraint = $metapoint->notconstraint;
				if ($this->datamodel->constraints($constraint)) 
					continue;
			}
			if (property_exists($metapoint,'access')) {
				$accessdata = (object) $metapoint->access;
				if (isset($accessdata->token)) {
					$token = $accessdata->token;
					$requires = $accessdata->requires;
					$access = Access::instance();
					if (!$access->permission($token,$requires,$this->datamodel->ownerid,$this->datamodel->articleid)) {
						continue;
					}
				} elseif (isset($accessdata->permission)) {
					$permission = 
						Node::get_permission($this->datamodel,$$accessdata->permission);
					if ($permission===FALSE)
						continue;
				}
			}
			if (property_exists($metapoint,'input')) {
				if ($metapoint instanceof Meta) {
					$this->_input_control_point($pointname,$metapoint,$pointlist);
				}
			}
		}
	}
	// add control point to point name according to metapoint, indexed by point name
	protected function _input_control_point($pointname,$metapoint,&$pointlist)
	{
		$input = (object) $metapoint->input;
		$method = $input->method;
		$datatype = $metapoint->datatype;
		$datapool = $this->datamodel->datapool;
		$iscommand = TRUE;
		switch ($method) {
			case Point::DATASOURCE_ARRAY: {
				$name = $input->name;
				$index = $input->index;
				$array = $datapool->arrays[$name];
				// command
				if (($datatype == Point::DATATYPE_BOOLEAN) and property_exists($metapoint,'command')) {
					if (isset($array[$index])) {
						$commandvalue = $array[$index];
					} else {
						$commandvalue = NULL;
					}
					$value = (boolean) (isset($array[$index]) and !empty($array[$index]));
					if ($value) $value = $metapoint->command;
				// ... or parameter
				} else {
					$iscommand = FALSE;
					if (isset($array[$index])) {
						$value = $array[$index];
					} else {
						$value = NULL;
					}
				}
				break;
			}
			case Point::DATASOURCE_ARRAYS: {
				$names = $input->names;
				$index = $input->index;
				$value = NULL;
				if (!empty($names)) {
					foreach ($names as $name) {
						$array = $datapool->arrays[$name];
						if (isset($array[$index])) {
							$value = $array[$index];
							$commandvalue = $value;
							break;
						}
					}
				}
				// command
				if (($datatype == Point::DATATYPE_BOOLEAN) and property_exists($metapoint,'command')) {
					$value = (boolean) (isset($value) and !empty($value));
					if ($value) $value = $metapoint->command;
				// parameter
				} else {
					$iscommand = FALSE;
				}
				break;
			}
		}
		if ($iscommand) {
			$point = $pointlist[$pointname] = Point::get_command($value,$metapoint);
			isset($commandvalue) or ($commandvalue = NULL);
			$point->commandvalue = $commandvalue;
		} else {
			$pointlist[$pointname] = Point::get_parameter($value,$metapoint);
		}
	}
	#========================================================================
	#-------------------------[ DATAFRAME ]----------------------------------
	#========================================================================
	// assemble data into $frame->data structure based on metadata
	// controller
	protected function _process_dataframe()
	{
		$frame = $this->_frame;
		$process_type = $this->_process_type;
		switch ($process_type) {
			# outgoing data
			// add datapoints to pointlist according to metadata
			case self::RETRIEVE: // internal, normally from database
				$metadatalist = $frame->metadata;
				$pointlist = &$frame->pointlist; // pointer allows accumulation to array
				if ($metadatalist)
					$this->_retrieve_data($metadatalist,$pointlist);
				break;
			// add data items to dictionary according to point list
			case self::OUTPUT: // create data dictionary for generator
				$pointlist = $frame->pointlist;
				$dictionary = array();
				$this->_output_data($pointlist, $dictionary); // pointer allows accumulation to array
				$this->datamodel->dictionary = $dictionary;
				break;
			# incoming data
			// add datapoints to pointlist according to metadata
			case self::INPUT: // normally from browser
				$metadatalist = $frame->metadata;
				$pointlist = &$frame->pointlist; // pointer allows accumulation to array
				if ($metadatalist)
					$this->_input_data($metadatalist,$pointlist);
				break;
			// add datapoints to pointlist according to metadata
			case self::VALIDATE: // previously collected input data
				$pointlist = $frame->pointlist; // pointer allows accumulation to array
				$error = !$this->_validate_data($pointlist);
				if ($error) $this->datamodel->errorstate = Point::ERRORSTATE_ERROR;
				break;
			// finish data transformations for valid data points
			case self::FINISHINPUT: // response to validation, possibly inluding point creation
				$pointlist = $frame->pointlist; // pointer allows accumulation to array
				$this->_finishinput_data($pointlist);
				break;
			// add datapoints to pointlist according to metadata
			case self::RECOVERINPUT: // create data points from specified input recover points
				$metadatalist = $frame->metadata;
				$pointlist = &$frame->pointlist; // pointer allows accumulation to array
				if ($metadatalist)
					$this->_recoverinput_data($metadatalist,$pointlist);
				break;
			// store data items from pointlist, if validated
			case self::STORE: // normally to cache, from which can be saved to database
				$pointlist = $frame->pointlist;
				$this->_store_data($pointlist);
				break;
		}
	}
	#========================= DATAFRAME RETRIEVE ==================================
	protected function _retrieve_data($metadatalist,&$pointlist)
	{
		foreach ($metadatalist as $pointname => $metapoint) {
			if (property_exists($metapoint,'constraint')) {
				$constraint = $metapoint->constraint;
				if (!$this->datamodel->constraints($constraint)) 
					continue;
			}
			if (property_exists($metapoint,'notconstraint')) {
				$constraint = $metapoint->notconstraint;
				if ($this->datamodel->constraints($constraint))
					continue;
			}
			if (property_exists($metapoint,'access')) {
				$accessdata = (object) $metapoint->access;
				if (isset($accessdata->token)) {
					$token = $accessdata->token;
					$requires = $accessdata->requires;
					$access = Access::instance();
					if (!$access->permission($token,$requires,$this->datamodel->ownerid,$this->datamodel->articleid)) {
						continue;
					}
				} elseif (isset($accessdata->permission)) {
					$permission = 
						Node::get_permission($this->datamodel,$accessdata->permission);
					if ($permission===FALSE)
						continue;
				}
			}
			if ($metapoint instanceof Meta) {
				if (property_exists($metapoint,'retrieve')) {
					$this->_retrieve_data_point($pointname,$metapoint,$pointlist);
				}
			} elseif ($metapoint instanceof Blockselect) {
				$this->_retrieve_data_select($pointname,$metapoint,$pointlist);
			} elseif ($metapoint instanceof Blocklist) {
				$this->_retrieve_data_list($pointname,$metapoint,$pointlist);
			} elseif ($metapoint instanceof Block) {
				$this->_retrieve_data_block($pointname,$metapoint,$pointlist);
			}
		}
	}
	#-----------------------[ Datapoint Retrieve ]-------------------------------
	protected function _retrieve_data_point($pointname,$metapoint,&$pointlist)
	{
		if (isset($pointlist[$pointname])) return;
		
		$value = $this->_get_datavalue($metapoint,$pointlist);
		$pointlist[$pointname] = Point::data($value, $metapoint);
	}
	// can be called directly; pointlist passed to allow reference to indexed values
	// pointlist is not modified here so is not passed by reference
	protected function _get_datavalue($metapoint,$pointlist)
	{
		$retrieve = (object) $metapoint->retrieve;
		$method = $retrieve->method;
		$datamodel = $this->datamodel;
		$datapool = $datamodel->datapool;
		if ($datamodel->recordcaching()) {
			$cache = $datamodel->cache;
		}
		$value = NULL;
		switch ($method) {
			case Point::DATASOURCE_ARRAY:
				$value = isset($datapool->arrays[$retrieve->name][$retrieve->index])
					? $datapool->arrays[$retrieve->name][$retrieve->index]
					: NULL;
				break;
			case Point::DATASOURCE_VARIABLE:
				$value = isset($datapool->variables[$retrieve->name])
					? $datapool->variables[$retrieve->name]
					: NULL;
				break;
			case Point::DATASOURCE_DEFAULT:
				$value = property_exists($metadata,'default')?$metadata->default:NULL;
				break;
			case Point::DATASOURCE_SELF:
				$value = property_exists($retrieve,'self')?$retrieve->self:NULL;
				break;
			case Point::DATASOURCE_FUNCTION:
				break;
			case Point::DATASOURCE_RECORD:
				if ($datamodel->recordcaching()) {
					if (isset($cache->records[$retrieve->recordindex])) {
						$record = $cache->records[$retrieve->recordindex];
					} else {
						$record = NULL;
					}
				} else {
					if (isset($datapool->records[$retrieve->recordindex])) {
						$record = $datapool->records[$retrieve->recordindex];
					} else {
						$record = NULL;
					}
				}
				if ($record) {
					$field = $retrieve->field;
					if (property_exists($record,$field)) {
						$value = $record->$field;
					} else {
						$value = NULL;
					}
				}
				break;
			case Point::DATASOURCE_ACTIVERECORD:
				if ($datamodel->recordcaching()) {
					if (isset($cache->activerecords[$retrieve->recordindex])) {
						$record = $cache->activerecords[$retrieve->recordindex]->current();
					} else {
						$record = NULL;
					}
				} else {
					if (isset($datapool->activerecords[$retrieve->recordindex])) {
						$record = $datapool->activerecords[$retrieve->recordindex]->current();
					} else {
						$record = NULL;
					}
				}
				if ($record) {
					$field = $retrieve->field;
					if (property_exists($record,$field)) {
						$value = $record->$field;
					} else {
						$value = NULL;
					}
				}
				break;
			case Point::DATASOURCE_RECORD_JSON:
				if ($datamodel->recordcaching() and isset($cache->records[$retrieve->recordindex])) {
					$record = $cache->records[$retrieve->recordindex];
				} else {
					$record = $datapool->records[$retrieve->recordindex];
				}
				$field = $retrieve->field;
				$json = $record->$field;
				if ($json) {
					if (is_string($json)) {
						$json = json_decode($json);
					}
					$jsonindex = $retrieve->json;
					if (isset($json->$jsonindex)) {
						$value = $json->$jsonindex;
					}
				}
				break;
			case Point::DATASOURCE_ACTIVERECORD_JSON:
				if ($datamodel->recordcaching() and isset($cache->activerecords[$retrieve->recordindex])) {
					$record = $cache->activerecords[$retrieve->recordindex];
				} else {
					$record = $datapool->activerecords[$retrieve->recordindex];
				}
				$field = $retrieve->field;
				$json = $record->$field;
				if ($json) {
					if (is_string($json)) {
						$json = json_decode($json);
					}
					$jsonindex = $retrieve->json;
					if (isset($json->$jsonindex)) {
						$value = $json->$jsonindex;
					}
				}
				break;
			case Point::DATASOURCE_MATCH:
				$arguments = (object) $retrieve->match;
				$method = $arguments->method;
				switch ($method) {
					case Point::DATASOURCE_VARIABLE: {
						$left = $datapool->variables[$arguments->name];
					}
					break;
				}
				$right = $pointlist[$arguments->blockindex]->value;
				if (is_null($left) or is_null($right))
					$value = NULL;
				else
					$value = ($left == $right);
				break;
			case Point::DATASOURCE_OBJECT:
				$name = $retrieve->name;
				$property = $retrieve->property;
				$value = $datapool->objects[$name]->$property;
				break;
			case Point::DATASOURCE_NOT_PERMITTED: // ** deprectated, should test permissions instead
				$accessdata = (object) $retrieve->access;
				$token = $accessdata->token;
				$requires = $accessdata->requires;
				$access = Access::instance();
				$value = !$access->permission($token,$requires,$this->datamodel->ownerid,$this->datamodel->articleid);
				break;
			case Point::DATASOURCE_PERMITTED: // ** deprectated, should test permissions instead
				$accessdata = (object) $retrieve->access;
				$token = $accessdata->token;
				$requires = $accessdata->requires;
				$access = Access::instance();
				$value = $access->permission($token,$requires,$this->datamodel->ownerid,$this->datamodel->articleid);
				break;
			case Point::DATASOURCE_BLOCKVALUE:
				if (property_exists($retrieve,'path')) {
					$path = $retrieve->path;
					$reflist = $this->_frame->pointlist;
					foreach ($path as $index) {
						$reflist = $reflist[$index]->pointlist;
					}
				} else {
					$reflist = $pointlist;
				}
				$name = $retrieve->name;
				$value = $reflist[$name]->value;
				break;
			case Point::DATASOURCE_LOOKUP_PARTIAL:
				try {
					$recordtypeindex = $retrieve->recordtypeindex;
					$arrayindex = $retrieve->arrayname;
					$recordtype = $pointlist[$recordtypeindex]->value;
					if ($recordtype)
						$value = $datapool->arrays[$arrayindex][$recordtype];
				} catch(Exception $e) {
					var_dump($e->getMessage());
					var_dump($e->getLine());
					die($e->getMessage() . ' ' . $e->getLine() . ' ' . $e->getFile());
				}
				break;
			case Point::DATASOURCE_LOOKUP_LAYOUT_PARTIAL:
				try {
					$layouttypeindex = $retrieve->layouttypeindex;
					$recordtypeindex = $retrieve->recordtypeindex;
					$arrayindex = $retrieve->arrayname;
					$recordtype = $pointlist[$recordtypeindex]->value;
					$layouttype = $this->_frame->pointlist[$layouttypeindex]->value;
					if ($recordtype and $layouttype)
						$value = $datapool->arrays[$arrayindex][$layouttype][$recordtype];
				} catch(Exception $e) {
					var_dump($e->getMessage());
					var_dump($e->getLine());
					die($e->getMessage() . ' ' . $e->getLine() . ' ' . $e->getFile());
				}
				break;
			case Point::DATASOURCE_LOOKUP_LAYOUT_IMAGEDATA:
				$layouttypeindex = $retrieve->layouttypeindex;
				$recordtypeindex = $retrieve->recordtypeindex;
				$imageidindex = $retrieve->imageidindex;
				$recordidindex = $retrieve->recordidindex;
				$recordtype = (!empty($pointlist[$recordtypeindex]->value))?$pointlist[$recordtypeindex]->value:NULL;
				$layouttype = (!empty($this->_frame->pointlist[$layouttypeindex]->value))?$this->_frame->pointlist[$layouttypeindex]->value:NULL;
				$imageid = (!empty($pointlist[$imageidindex]->value))?$pointlist[$imageidindex]->value:NULL;
				$recordid = (!empty($pointlist[$recordidindex]->value))?$pointlist[$recordidindex]->value:NULL;
				if ($recordtype == 'picture') {
					$imageid = $recordid;
				}
				if (is_numeric($imageid)) {
					$operation = DB::select('ObjectType','Directory','FileName')
						->from('Nodes_View')
						->where('ID','=',$imageid)
						->as_object()
						->prepare()
						->execute();
					if ($operation->success()) {
						$record = $operation->fetch();
						if ($record) {
							if ($record->ObjectType == 'picture') {
								$system = System::instance();
								$originalfilespec = $system->PublicBase.$system->UploadBase.$record->Directory.$record->FileName;
								$version = NULL;
								switch ($layouttype) {
									case 'planks':
									case 'mediumtiles':
										$version = 'smallest';
										break;
									case 'largetiles':
										$version = 'small';
										break;
									case 'gallery':
										$version = 'medium';
									break;
								}
								if (!empty($version)) {
									$pathinfo = pathinfo($originalfilespec);
									$filespec = $pathinfo['dirname'].'/'.$pathinfo['filename'].'.'.$version.'.'.$pathinfo['extension'];
								}
								if (!empty($version) and !empty($filespec)) {
									if (is_file($filespec)) {
										$value = $filespec;
									} elseif (is_file($originalfilespec)) { // there may be a smaller version available
										$filespec = $this->_find_smaller_version($originalfilespec,$version);
										if (!empty($filespec)) {
											$value = $filespec;
										}
									}
								}
							}
						}
					}
				}
				break;
			case Point::DATASOURCE_LOOKUP_RECORD_IMAGEDATA:
				$recordtypeindex = $retrieve->recordtypeindex;
				$imageidindex = $retrieve->imageidindex;
				$recordidindex = $retrieve->recordidindex;
				$recordtype = (!empty($pointlist[$recordtypeindex]->value))?$pointlist[$recordtypeindex]->value:NULL;
				$imageid = (!empty($pointlist[$imageidindex]->value))?$pointlist[$imageidindex]->value:NULL;
				$recordid = (!empty($pointlist[$recordidindex]->value))?$pointlist[$recordidindex]->value:NULL;
				if ($recordtype == 'picture') {
					break;
				}
				if (is_numeric($imageid)) {
					$operation = DB::select('ObjectType','Directory','FileName','Title','Caption','Description')
						->from('Nodes_View')
						->where('ID','=',$imageid)
						->as_object()
						->prepare()
						->execute();
					if ($operation->success()) {
						$record = $operation->fetch();
						if ($record) {
							if ($record->ObjectType == 'picture') {
								$system = System::instance();
								$originalfilespec = $system->PublicBase.$system->UploadBase.$record->Directory.$record->FileName;
								$pathinfo = pathinfo($originalfilespec);
								$filespec = $pathinfo['dirname'].'/'.$pathinfo['filename'].'.small.'.$pathinfo['extension'];
								if (!is_file($filespec)) {
									if (is_file($originalfilespec)) {
										$filespec = $this->_find_smaller_version($originalfilespec,'small');
									} else {
										$filespec = NULL;
									}
								}
								if (!empty($filespec)) {
									$smallfilespec = $pathinfo['dirname'].'/'.$pathinfo['filename'].'.smallest.'.$pathinfo['extension'];
									if (!is_file($smallfilespec))
										$smallfilespec = NULL;
									$value = $filespec;
									$imagetitleindex = $retrieve->imagetitleindex;
									$imagecaptionindex = $retrieve->imagecaptionindex;
									$imagedescriptionindex = $retrieve->imagedescriptionindex;
									$smallimageurlindex = $retrieve->smallimageurlindex;
									if (isset($pointlist[$imagetitleindex]->value) or is_null($pointlist[$imagetitleindex]->value))
										$pointlist[$imagetitleindex]->value = $record->Title;
									if (isset($pointlist[$imagecaptionindex]->value) or is_null($pointlist[$imagecaptionindex]->value))
										$pointlist[$imagecaptionindex]->value = $record->Caption;
									if (isset($pointlist[$imagedescriptionindex]->value) or is_null($pointlist[$imagedescriptionindex]->value))
										$pointlist[$imagedescriptionindex]->value = $record->Description;
									if (isset($pointlist[$smallimageurlindex]->value) or is_null($pointlist[$smallimageurlindex]->value))
										$pointlist[$smallimageurlindex]->value = $smallfilespec;
								}
							}
						}
					}
				}
				break;
			case Point::DATASOURCE_CALCULATE_RECORDSET_DELIMITER:
				$layouttypeindex = $retrieve->layouttypeindex;
				$layouttype = (!empty($this->_frame->pointlist[$layouttypeindex]->value))?$this->_frame->pointlist[$layouttypeindex]->value:NULL;
				switch ($layouttype) {
					case 'mediumtiles':
					case 'largetiles':
					case 'gallery':
						$value = TRUE;
						break;
					default:
						$value = FALSE;
				}
				break;
			case Point::DATASOURCE_ISFAVOURITE:
				$user = User::instance();
				if ($user->authenticated()) {
					$articleid = $pointlist[$retrieve->articleidref]->value;
					if (!is_null($articleid)) {
						$record = DB::select('ID')
							->from('Articles_Favourites')
							->where('UserID','=',$user->userid())
							->where('ArticleID','=',$articleid)
							->as_object()
							->prepare()
							->execute()
							->current();
						$value = (bool) $record;
					} else {
						$value = FALSE;
					}
				} else {
					$value = FALSE;
				}
				break;
			case Point::DATASOURCE_ISCOMMENTS:
				$articleid = $pointlist[$retrieve->articleidref]->value;
				if (!is_null($articleid)) {
					$tablename = 'Nodes_View';
					$query = DB::select('ID')
						->from($tablename)
						->where('CommentArticleID','=',$articleid);
					Node::set_query_mixed_constraints($this->datamodel,$query,$tablename);
					$operation = $query
						->as_object()
						->prepare()
						->execute();
					$record = $operation->current();

/*					$record = DB::select('ID')
						->from('Articles_Extensions')
						->where('CommentArticleID','=',$articleid)
						->as_object()
						->prepare()
						->execute()
						->current();
*/
					$value = (bool) $record;
				} else {
					$value = FALSE;
				}
				break;
			case Point::DATASOURCE_ISATTACHMENTS:
				$articleid = $pointlist[$retrieve->articleidref]->value;
				if (!is_null($articleid)) {
					$tablename = 'Articles_Attachments_View';
					$query = DB::select('ID')
						->from($tablename)
						->where('ParentArticleID','=',$articleid);
					Node::set_query_mixed_constraints($this->datamodel,$query,$tablename);
					$record = $query
						->as_object()
						->prepare()
						->execute()
						->current();

/*					$record = DB::select('ID')
						->from('Articles_Attachments')
						->where('ArticleID','=',$articleid)
						->as_object()
						->prepare()
						->execute()
						->current();
*/
					$value = (bool) $record;
				} else {
					$value = FALSE;
				}
				break;
			case Point::DATASOURCE_ISNARRATIVE:
				$value = (isset($pointlist[$retrieve->valueref]) and (!$pointlist[$retrieve->valueref]->value));
				break;
			case Point::DATASOURCE_ISANCILLARYDATA:
				$blockrefs = $retrieve->blockrefs;
				$value = FALSE;
				foreach ($blockrefs as $blockref) {
					$value = ($value or (isset($pointlist[$blockref]) and $pointlist[$blockref]->value));
				}
				break;
			case Point::DATASOURCE_CONTENTCOUNT:
				$idindex = $retrieve->idindex;
				$constraints = $this->datamodel->datapool->constraints;
				$objecttype = (!empty($constraints['objecttype']))?$constraints['objecttype']:NULL;
				if (!$objecttype)
					$value = FALSE;
				else {
					$id = $pointlist[$idindex]->value;
					switch ($objecttype) {
						case 'file':
						case 'inbox':
						case 'folder':
						case 'directory':
							if (is_null($id))
								$value = 0;
							else {
								$tablename = 'Nodes_Instances_View';
								$query = DB::select(DB::expr('count(*) as Count'))
								->from($tablename)
								->where('FolderID','=',$id);
								Node::set_query_mixed_constraints($this->datamodel,$query,$tablename);
								$operation = $query->as_object()
								->prepare()
								->execute();
								$record = $operation->current();
								$value = $record->Count;
							}
							break;
						case 'blog':
							if (is_null($id))
								$value = 0;
							else {
								$tablename = 'Nodes_View';
								$query = DB::select(DB::expr('count(*) as Count'))
								->from($tablename)
								->where('ParentID','=',$id);
								Node::set_query_mixed_constraints($this->datamodel,$query,$tablename);
								$operation = $query->as_object()
								->prepare()
								->execute();
								$record = $operation->current();
								$value = $record->Count;
							}
							break;
						case 'page':
							if (is_null($id))
								$value = 0;
							else {
								$tablename = 'Nodes_View';
								$query = DB::select(DB::expr('count(*) as Count'))
								->from($tablename)
								->where('PageParentID','=',$id);
								Node::set_query_mixed_constraints($this->datamodel,$query,$tablename);
								$operation = $query->as_object()
								->prepare()
								->execute();
								$record = $operation->current();
								$value = $record->Count;
							}
							break;
						default:
							$value = FALSE;
					}
					(empty($value)) and ($value = FALSE);
				}
				break;
		}
		if (is_null($value)) {
			if (property_exists($metapoint,'default')) {
				$value = $metapoint->default;
			}
		}
		if (!is_null($value)) {
			if (property_exists($retrieve,'filter')) {
				$filter = (object) $retrieve->filter;
				switch ($filter->method) {
					case Point::FILTER_COMPARE: {
						$arguments = (object) $filter->arguments;
						$value = ($value === $arguments->compare);
						break;
					}
					case Point::FILTER_LOOKUP: {
						$arguments = (object) $filter->arguments;
						$record = DB::select($arguments->index,$arguments->value)
							->from($arguments->name)
							->where($arguments->index,'=',$value)
							->as_object()
							->execute($this->_database)->current();
						if ($record) {
							$field = $arguments->value;
							$value = $record->$field;
						} else {
							$value = NULL;
						}
						break;
					}
					case Point::FILTER_MARKUP: {
						$arguments = (object) $filter->arguments;
						$value = $this->_get_filtered_markup($arguments,$value);
						break;
					}
				}
			}
			$datatype = $metapoint->datatype;
			switch ($datatype) {
				case Point::DATATYPE_DATETIME:
				case Point::DATATYPE_DATE:
					$value = Database::instance($this->_database)->php_datetime($value);
					break;
				case Point::DATATYPE_BOOLEAN:
					$value = (bool) $value;
					break;
			}
		}
		return $value;
	}
	protected function _find_smaller_version($filespec,$version)
	{
		$pathinfo = pathinfo($filespec);
		do {
			switch ($version) {
				case 'largest':
					$version = 'larger';
					break;
				case 'larger':
					$version = 'large';
					break;
				case 'large':
					$version = 'medium';
					break;
				case 'medium':
					$version = 'small';
					break;
				case 'small':
					$version = 'smaller';
					break;
				case 'smaller':
					$version = 'smallest';
					break;
				default:
					$version = NULL;
			}
			if (!empty($version)) {
				$filespec = $pathinfo['dirname'].'/'.$pathinfo['filename'].'.'.$version.'.'.$pathinfo['extension'];
			} else {
				$filespec = NULL;
			}
		} while (!empty($filespec) and !is_file($filespec));
		return $filespec;
	}
	protected function _get_filtered_markup($arguments,$markup)
	{
		$parser = (object) $arguments->parser;
		$method = $parser->method;
		$name = $parser->recordindex;
		$field = $parser->field;
		switch ($method) {
			case Point::DATASOURCE_RECORD: {
				$markuptypeid = $this->datamodel->datapool->records[$name]->$field;
				break;
			}
			case Point::DATASOURCE_ACTIVERECORD: {
				$markuptypeid = $this->datamodel->datapool->activerecords[$name]->$field;
				break;
			}
		}
		if ($markuptypeid == 0) { // default == SimpleWiki
			$simplewiki = $this->datamodel->controller->simplewiki_instance();
			$markup = $simplewiki->get_html($markup);
		} else { // lookup alternate parser
		}
		return $markup;
	}
	#-----------------------[ Blockselect Retrieve ]-------------------------------
	protected function _retrieve_data_select($pointname,$metapoint,&$pointlist)
	{
		// if the select structure exists with options, we're done.
		if (isset($pointlist[$pointname]) and 
			(!empty($pointlist[$pointname]->pointlist))) return;
		// retrieve options list
		$options = (object) $metapoint->options;
		$optionsretrieve = (object) $options->retrieve;
		$datapool = $this->datamodel->datapool;
		
		$method = $optionsretrieve->method;
		switch ($method) {
			case Point::DATASOURCE_VALUE:
				$optionslist = $optionsretrieve->value;
				break;
			case Point::DATASOURCE_VARIABLE:
				$optionslist = $datapool->variables[$optionsretrieve->name];
				break;
			case Point::DATASOURCE_LOOKUP_TABLE:
				$query = DB::select($optionsretrieve->index,$optionsretrieve->value)
					->from($optionsretrieve->name);
				if (property_exists($optionsretrieve,'orderby'))
					$orderby = $optionsretrieve->orderby;
				else
					$orderby = array(array($optionsretrieve->value));
				foreach ($orderby as $order) {
					$fld = $order[0];
					if (isset($order[1])) $dir = $order[1];
					else $dir = NULL;
					$query->order_by($fld,$dir);
				}
				if (property_exists($optionsretrieve,'where')) {
					$wherelist = $optionsretrieve->where;
					foreach ($wherelist as $wherestatement) {
						list($col,$op,$val) = $wherestatement;
						$query->where($col,$op,$val);
					}
				}
				$optionslist = $query
					->prepare($this->_database)
					->execute()
					->as_array($optionsretrieve->index,$optionsretrieve->value);
		}
		// create block object, with default value
		if (!isset($pointlist[$pointname])) {
			$selection_pointlist = array(); // throwaway
			$selection = $this->_get_datavalue($metapoint,$selection_pointlist);
			$pointlist[$pointname] = Point::blockselect($metapoint,$selection);
		}
		$block = $pointlist[$pointname];
		$selection = $block->value;
		// create options list
		$selection_var_name = $options->selectionvariable;
		$datapool->variables[$selection_var_name] = $selection; // referenced by selection blocks for select test
		// get sublist
		$block->pointlist = array();
		$blockpointlist = &$block->pointlist;
		$metadata = $metapoint->metadata;
		// ** try less indirection here by making datapool array indexes constants - HB
		$datapoolarray = (object) $options->datapoolarray;
		if (!empty($optionslist)) {
			foreach ($optionslist as $index => $value) {
				// set data for iteration lookup
				$datapool->arrays[$datapoolarray->name] = array($datapoolarray->index => $index,$datapoolarray->value => $value);
				// collect single block
				$subblock = Point::block($metadata);
				$subblock->pointlist = array();
				$subpointlist = &$subblock->pointlist;
				if ($metadata)
					$this->_retrieve_data($metadata,$subpointlist);
				// collect blocks
				if (!empty($subpointlist)) $blockpointlist[] = $subblock;
			}
		}
		// capture blocks
		unset($datapool->arrays[$datapoolarray->name]);
		unset($datapool->variables[$selection_var_name]);
	}
	#-----------------------[ Blockpoint Retrieve ]-------------------------------
	protected function _retrieve_data_block($pointname,$metapoint,&$pointlist)
	{
		if (!isset($pointlist[$pointname])) 
			$pointlist[$pointname] = Point::block($metapoint);
		$block = $pointlist[$pointname];

		$retrieve = property_exists($metapoint,'retrieve')? $metapoint->retrieve: TRUE;
		
		if (is_bool($retrieve)) { // ** DEPRECATED
			if ($retrieve) {
				$metadata = $metapoint->metadata;
				$blockpointlist = &$block->pointlist;
				if ($metadata)
					$this->_retrieve_data($metadata,$blockpointlist);
			}
		} else { // *** possibly look for constraint
		}
	}
	#-----------------------[ Blocklist Retrieve ]-------------------------------
	protected function _retrieve_data_list($pointname,$metapoint,&$pointlist)
	{
		$retrieve = (object) $metapoint->retrieve;
		$method = $retrieve->method;
		switch ($method) {
			case Point::DATASOURCE_RECORDSET: {
				$this->_retrieve_data_list_recordset($pointname,$metapoint,$pointlist);
				break;
			}
			case Point::DATASOURCE_TABLEACTIVERECORD: {
				$this->_retrieve_data_list_tableactiverecord($pointname,$metapoint,$pointlist);
				break;
			}
		}
	}
	protected function _retrieve_data_list_recordset($pointname,$metapoint,&$pointlist)
	{
		$retrieve = (object) $metapoint->retrieve;
		$datapool = $this->datamodel->datapool;
		$cache = $this->datamodel->cache;
		$recordcaching = $this->datamodel->recordcaching();
		// new blocklist items
		$recordset = NULL;
		$recordvar = NULL;
		// the recordset of Point::DATASOURCE_RECORDSET:
		if ($recordcaching) {
			if (isset($cache->recordsets[$retrieve->tableindex])) {
				$recordset = $cache->recordsets[$retrieve->tableindex];
				$recordvar = $retrieve->recordvariable;
			}
		} else {
			if (isset($datapool->recordsets[$retrieve->tableindex])) {
				$recordset = $datapool->recordsets[$retrieve->tableindex];
				$recordvar = $retrieve->recordvariable;
			}
		}
				
		$count = -1; // for generating and tracking session references
		# refresh recycled items from input ($_POST)
		if (isset($pointlist[$pointname])) { // refresh existing blocklist items (from input)
			$listitems = $pointlist[$pointname]->pointlist;
			// recordset entries are removed as they are processed; any remaining are processed below
			$this->_retrieve_list_refresh_from_recordset($metapoint,$listitems,$recordset,$count);
		}
		# create new items for output
		if ($recordset and $recordvar) { // new blocklist items
			// collect new itemss...
			$collection = array();
			$this->_retrieve_list_newitems_from_recordset($metapoint,$collection,$recordset,$count);
			// merge with existing items
			if (isset($pointlist[$pointname])) {
				$blocklist = $pointlist[$pointname];
				$oldcollection = (array) $blocklist->pointlist;
				$blocklist->pointlist = array_merge($oldcollection,$collection);
			// .. or create new list item
			} else {
				$blocklist = $pointlist[$pointname] = Point::blocklist($metapoint);
				$blocklist->pointlist = $collection;
			}
		}
	}
	protected function _retrieve_list_refresh_from_recordset($listmetadata,$listitems,&$recordset,&$count)
	{
		$datapool = $this->datamodel->datapool;
		$cache = $this->datamodel->cache;
		$recordcaching = $this->datamodel->recordcaching();
		$listitem_metadata = $listmetadata->metadata;
		$retrieve = (object) $listmetadata->retrieve;
		$recordvar = $retrieve->recordvariable;
		foreach ($listitems as $listitem) { // existing blocklist items
			$count++;
			$listitem->index = $count;
			$this->_retrieve_data_list_setvariables($retrieve,$count);
			$subpointlist = &$listitem->pointlist;
			if (empty($subpointlist) and 
				($listitem->recordstatus == 'posted')) { // refresh all data including posted record data
				
				if (isset($recordset[$listitem->reference])) {
					if ($recordcaching) {
						$cache->records[$recordvar] = $recordset[$listitem->reference];
					} else {
						$datapool->records[$recordvar] = $recordset[$listitem->reference];
					}
				} else { // error?
					if ($recordcaching) {
						$cache->records[$recordvar] = new StdClass;
					} else {
						$datapool->records[$recordvar] = new StdClass;
					}
				}
				$this->_retrieve_data($listitem_metadata,$subpointlist);
				if (property_exists($retrieve,'recordstatusindex')) {
					$subpointlist[$retrieve->recordstatusindex]->value = $listitem->recordstatus;
				}
				unset($recordset[$listitem->reference]);
			} else { // retrieve ancillary data
				$this->_retrieve_data($listitem_metadata,$subpointlist);
			}
		}
		unset($subpointlist);
	}
	protected function _retrieve_list_newitems_from_recordset($listmetadata,&$collection,&$recordset,&$count)
	{
		$datapool = $this->datamodel->datapool;
		$cache = $this->datamodel->cache;
		$recordcaching = $this->datamodel->recordcaching();
		$listitem_metadata = $listmetadata->metadata;
		$retrieve = (object) $listmetadata->retrieve;
		$recordvar = $retrieve->recordvariable;
		foreach ($recordset as $record) {
			$count++;
			$this->_retrieve_data_list_setvariables($retrieve,$count);
			$subpointlist = array();
			if ($recordcaching) {
				$cache->records[$recordvar] = $record;
			} else {
				$datapool->records[$recordvar] = $record;
			}
			$this->_retrieve_data($listitem_metadata,$subpointlist);
			$listitem = $collection[] = Point::listitem($listitem_metadata,$subpointlist);
			$listitem->index = $count;
		}
	}
	// ** this is way too convoluted
	// referenceprefix not required. 
	// Instead of arrays of fields, s/b arrays of records:
	// s/b records[sessionrecordid][fieldname] not records[fieldname][sessionidrecordid]
	protected function _retrieve_data_list_tableactiverecord($pointname,$metapoint,&$pointlist)
	{
		$retrieve = (object) $metapoint->retrieve;
		$datapool = $this->datamodel->datapool;
		$table = NULL;
		if (isset($datapool->tables[$retrieve->tableindex])) {
			$table = $datapool->tables[$retrieve->tableindex];
		}
		// no table found, cannot continue
		if (!$table) {
			echo 'TABLE ' . $retrieve->tableindex . ' NOT FOUND IN DATAPOOL TABLES.';
			return;
		}
		if (property_exists($retrieve,'primarykeyfieldname'))
			$primarykeyfieldname = $retrieve->primarykeyfieldname;
		$metadata = $metapoint->metadata;
		// added blocklist items
		$newrecords = NULL;
		if (property_exists($retrieve,'newrecordsvariable')) {
			if (isset($datapool->variables[$retrieve->newrecordsvariable]))
				$newrecords = $datapool->variables[$retrieve->newrecordsvariable];
		}
		// three processes: refresh, create, new
		$count = -1;
		# refresh recycled items from input ($_POST)
		if (isset($pointlist[$pointname])) { // existing blocklist (from input)
			$listitems = $pointlist[$pointname]->pointlist;
			$refreshed_primarykeys = $this->_retrieve_list_refresh_from_table($metapoint,$listitems,$table,$count);
			// exclude already processed records from next step
			if (!empty($refreshed_primarykeys)) {
				$table->reset()->query()->reset()->where($primarykeyfieldname,'NOT IN',$refreshed_primarykeys);
			} 
		}
		# create listitem from table record
		$activerecord = $table->activerecord();
		if (property_exists($retrieve,'orderby')) {
			$orderby = $retrieve->orderby;
			$query = $table->query();
			foreach ($orderby as $order) {
				$fld = $order[0];
				if (isset($order[1])) $dir = $order[1];
				else $dir = NULL;
				$query->order_by($fld,$dir);
			}
		}
		if (property_exists($retrieve,'limit')) {
			$limit = $retrieve->limit;
			$query = $table->query();
			$query->limit($limit);
		}
		if (property_exists($retrieve,'callback')) { // to compensate for lack of sticky query values for dependent tables
			$callback = (object) $retrieve->callback;
			$method = $callback->method;
			$tableindex = $callback->tableindex;
			call_user_func($method,$this->datamodel,$tableindex);
		}
		$table->reset();
//		echo $table->query()->sql() . '</br />';
		$collection = array();
		while ($activerecord->fetch()) {
			$count++;
			$this->_retrieve_data_list_setvariables($retrieve,$count);
			$subpointlist = array();
			$this->_retrieve_data($metadata,$subpointlist);
			$listitem = $collection[] = Point::listitem($metadata,$subpointlist);
			$listitem->index = $count;
		}
		# new (added) items for output
		if ($newrecords) { // new blocklist items
			$table->reset()->query()->reset()->where($primarykeyfieldname,'=',NULL); // purposely fail to find record
			for ($newrecord = 0; $newrecord < $newrecords; $newrecord++) {
				$count++;
				$this->_retrieve_data_list_setvariables($retrieve,$count);
				$subpointlist = array();
				$this->_retrieve_data($metadata,$subpointlist);
				$listitem = $collection[] = Point::listitem($metadata,$subpointlist);
				$listitem->index = $count;
			}
		}
		if (!empty($collection)) {
			if (isset($pointlist[$pointname])) {
				$blocklist = $pointlist[$pointname];
				$oldcollection = (array) $blocklist->pointlist;
				$blocklist->pointlist = array_merge($oldcollection,$collection);
			} else {
				$blocklist = $pointlist[$pointname] = Point::blocklist($metapoint);
				$blocklist->pointlist = $collection;
			}
		}
	}
	protected function _retrieve_list_refresh_from_table($listmetadata,$listitems,$table,&$count)
	{
		$refreshedprimarykeys = array();
		$listitem_metadata = $listmetadata->metadata;
		$retrieve = (object) $listmetadata->retrieve;
		$primarykeyblockindex = $retrieve->primarykeyblockindex;
		$primarykeyfieldname = $retrieve->primarykeyfieldname;
		foreach ($listitems as $listitem) { // existing blocklist items
			$count++;
			$listitem->index = $count;
			$this->_retrieve_data_list_setvariables($retrieve,$count);
			$subpointlist = &$listitem->pointlist;
			if (empty($subpointlist) and 
				($listitem->recordstatus == 'posted')) { // refresh posted record
				
				$refreshedprimarykeys[] = $primarykeyvalue = $listitem->primarykeyvalue;
				$table->reset()->query()->reset()->where($primarykeyfieldname,'=',$primarykeyvalue);
				$this->_retrieve_data($listitem_metadata,$subpointlist);
				if (property_exists($retrieve,'recordstatusindex')) {
					$subpointlist[$retrieve->recordstatusindex]->value = $listitem->recordstatus;
				}
			} else {
				$primarykeyvalue = $refreshedprimarykeys[] = $subpointlist[$primarykeyblockindex]->value;
				$table->reset()->query()->reset()->where($primarykeyfieldname,'=',$primarykeyvalue);
				$this->_retrieve_data($listitem_metadata,$subpointlist);
			}
		}
		unset($subpointlist);
		return $refreshedprimarykeys;
	}
	protected function _retrieve_data_list_setvariables($retrieve,$count)
	{
		$datapool = $this->datamodel->datapool;
		if (property_exists($retrieve,'indexvariable')) {
			$indexvariable = $retrieve->indexvariable;
			$datapool->variables[$indexvariable] = $count;
		}
		if (property_exists($retrieve,'referencevariable')) {
			$referencevariable = $retrieve->referencevariable;
			$referenceprefix = $retrieve->referenceprefix;
			$datapool->variables[$referencevariable] = $referenceprefix . $count;
		}
		if (property_exists($retrieve,'zebrastripes')) {
			$zebrastripes = (object) $retrieve->zebrastripes;
			$zebravalues = $zebrastripes->zebravalues;
			$zebravariable = $zebrastripes->zebravariable;
			$oddeven = $count % 2;
			$datapool->variables[$zebravariable] = $zebravalues[$oddeven];
		}
		if (property_exists($retrieve,'newrowvariable')) {
			$newrowvariable = $retrieve->newrowvariable;
			if (property_exists($retrieve,'layouttypeindex') and $count > 0) {
				$mod = NULL;
				$layouttype = $this->_frame->pointlist[$retrieve->layouttypeindex]->value;
				$mod = NULL;
				switch ($layouttype) {
					case 'gallery':
						$mod = 2;
						break;
					case 'largetiles':
						$mod = 3;
						break;
					case 'mediumtiles':
						$mod = 4;
						break;
				}
				if (!is_null($mod)) {
					$modulo = $count % $mod;
					if ($modulo == 0) {
						$datapool->variables[$newrowvariable] = TRUE;
					} else {
						$datapool->variables[$newrowvariable] = FALSE;
					}
				}
			}
		}
	}
	#========================= DATAFRAME OUTPUT ==================================
	protected function _output_data($pointlist,&$datadictionary)
	{
		foreach ($pointlist as $pointname => $datapoint) {
			$metapoint = $datapoint->metapoint;
			if (property_exists($metapoint,'output')) {
				if ($datapoint instanceof Data) {
					$this->_output_data_point($datapoint, $datadictionary);
				} elseif ($datapoint instanceof Blockselect) {
					$this->_output_data_select($datapoint, $datadictionary);
				} elseif ($datapoint instanceof Blocklist) {
					$this->_output_data_list($datapoint, $datadictionary);
				} elseif ($datapoint instanceof Block) {
					$this->_output_data_block($datapoint, $datadictionary);
				}
			}
		}
	}
	protected function _output_data_point($datapoint, &$datadictionary)
	{
		$metapoint = $datapoint->metapoint;
		$output = (object) $metapoint->output;
		$value = $datapoint->value;
		if (property_exists($output,'method')) {
			$method = $output->method;
			switch ($method) {
				case Point::DATASOURCE_CONDITIONAL: {
					$compareindex = $datadictionary[$output->blockindex];
					$compare = $output->compare;
					if (isset($compare[$compareindex])) {
						$value = $compare[$compareindex];
					} else {
						$value = NULL;
					}
					break;
				}
			}
		}
		if (!($datapoint->errorstate or $datapoint->recoverstate)) {
			if (!is_null($value)) {
				$datatype = $metapoint->datatype;
				switch ($datatype) {
					case Point::DATATYPE_DATE:
						$value = date($this->_dateformat,$value);
						break;
					case Point::DATATYPE_DATETIME:
						$value = date($this->_datetimeformat,$value);
						break;
					case Point::DATATYPE_BOOLEAN:
						$value = (bool) $value;
						break;
				}
			}
		}
		if (property_exists($output,'filter')) {
			$filter = (object) $output->filter;
			switch ($filter->method) {
				case Point::FILTER_CHECKED: // ** DEPRECATED
					$value = (bool) $value;
					if ($value) {
						$value = 'checked="checked"';
					} else { // don't set false instance
						$value = NULL;
					}
					break;
				case Point::FILTER_DISABLED: // ** DEPRECATED
					$value = (bool) $value;
					if ($value) {
						$value = 'disabled="disabled"';
					} else { // don't set false instance
						$value = NULL;
					}
					break;
				case Point::FILTER_SELECTED: // ** DEPRECATED
					$value = (bool) $value;
					if ($value) {
						$value = 'selected="selected"';
					} else { // don't set false instance
						$value = NULL;
					}
					break;
			}
		} 
		// ** trying to catch a subtle bug 
		try {
			if (!property_exists($output,'index')) {
				echo 'missing output index';
				var_dump($output);exit;
			}
			$datadictionary[$output->index] = $value;
		} catch (Exception $e) {
			var_dump($e->getMessage());
			var_dump($output);
		}
	}
	protected function _output_data_select($datapoint, &$datadictionary)
	{
		$pointlist = $datapoint->pointlist;
		if (!empty($pointlist)) {
			$subdictionary = array();
			$output = (object) $datapoint->metapoint->output;
			foreach ($pointlist as $subdatapoint) {
				$pointdictionary = array();
				$pointlist = $subdatapoint->pointlist;
				// obtain list item
				$this->_output_data($pointlist, &$pointdictionary);
				// accumulate list
				if (!empty($pointdictionary)) 
					$subdictionary[] = $pointdictionary;
			}
			if (!empty($subdictionary))
				$datadictionary[$output->index] = $subdictionary;
		}
	}
	protected function _output_data_block($datapoint, &$datadictionary)
	{
		$pointlist = $datapoint->pointlist;
		if (!empty($pointlist)) {
			$subdictionary = array();
			$this->_output_data($pointlist, $subdictionary);
			$output = (object) $datapoint->metapoint->output;
			if (!empty($subdictionary))
				$datadictionary[$output->index] = (object) $subdictionary;
		}
	}
	protected function _output_data_list($datapoint, &$datadictionary)
	{
		$pointlist = $datapoint->pointlist; // an array of objects
		if (!empty($pointlist)) {
			$collection = array();
			foreach ($pointlist as $listitem) { // each array object
				$subdictionary = array();
				$this->_output_data($listitem->pointlist, $subdictionary);
				$collection[] = $subdictionary;
			}
			$output = (object) $datapoint->metapoint->output;
			if (!empty($collection))
				$datadictionary[$output->index] = $collection;
		}
	}
	#========================= DATAFRAME INPUT ==================================
	protected function _input_data($metadatalist,&$pointlist)
	{
		foreach ($metadatalist as $pointname => $metapoint) {
			if (property_exists($metapoint,'constraint')) {
				$constraint = $metapoint->constraint;
				if (!$this->datamodel->constraints($constraint)) {
					continue;
				}
			}
			if (property_exists($metapoint,'notconstraint')) {
				$constraint = $metapoint->notconstraint;
				if ($this->datamodel->constraints($constraint)) {
					continue;
				}
			}
			if (property_exists($metapoint,'token')) {
				$accessdata = (object) $metapoint->access;
				if (isset($accessdata->token)) {
					$token = $accessdata->token;
					$requires = $accessdata->requires;
					$access = Access::instance();
					if (!$access->permission($token,$requires,$this->datamodel->ownerid,$this->datamodel->articleid)) {
						continue;
					}
				} elseif (isset($accessdata->permission)) {
					$permission = 
						Node::get_permission($this->datamodel,$accessdata->permission);
					if ($permission===FALSE)
						continue;
				}
			}
			if (property_exists($metapoint,'input')) {
				if ($metapoint instanceof Meta) {
					$this->_input_data_point($pointname,$metapoint,$pointlist);
				} elseif ($metapoint instanceof Blockselect) {
					$this->_input_data_blockselect($pointname,$metapoint,$pointlist);
				} elseif ($metapoint instanceof Blocklist) {
					$this->_input_data_list($pointname,$metapoint,$pointlist);
				} elseif ($metapoint instanceof Block) {
					$this->_input_data_block($pointname,$metapoint,$pointlist);
				}
			}
		}
	}
	protected function _input_data_point($pointname,$metapoint,&$pointlist)
	{
		$input = (object) $metapoint->input;
		if (property_exists($input,'constraint')) { // specialized input constraint
			$constraint = $input->constraint;
			if (!$this->datamodel->constraints($constraint)) {
				return;
			}
		}
		if (property_exists($input,'notconstraint')) { // specialized input constraint
			$constraint = $input->notconstraint;
			if ($this->datamodel->constraints($constraint)) {
				return;
			}
		}
		$method = $input->method;
		$datatype = $metapoint->datatype;
		$datapool = $this->datamodel->datapool;
		switch ($method) {
			case Point::DATASOURCE_ARRAY: {
				$name = $input->name;
				$array = $datapool->arrays[$name];
				if (property_exists($input,'index')) {
					$index = $input->index;
					if (($name == self::POST_ARRAY) and ($datatype == Point::DATATYPE_BOOLEAN)) {
						$value = (boolean) (isset($array[$index]) and !empty($array[$index]));
					} elseif (isset($array[$index])) {
						$value = $array[$index];
					} else {
						$value = NULL;
					}
				} elseif (property_exists($input,'path')) {
					$path = $input->path;
					$referencevariable = $input->referencevariable;
					foreach ($path as $index) {
						if (isset($array[$index])) {
							$array = $array[$index];
							$success = TRUE;
						} else {
							$success = FALSE;
							break;
						}
					}
					if (!$success) {
						if (($name == self::POST_ARRAY) and ($datatype == Point::DATATYPE_BOOLEAN)) {
							$value = FALSE;
						} else {
							$value = NULL;
						}
					} else {
						$index = $datapool->variables[$referencevariable];
						if (($name == self::POST_ARRAY) and ($datatype == Point::DATATYPE_BOOLEAN)) {
							$value = (boolean) (isset($array[$index]) and !empty($array[$index]));
						} elseif (isset($array[$index])) {
							$value = $array[$index];
						} else {
							$value = NULL;
						}
					}
				} else {
					$value = NULL;
				}
				break;
			}
			case Point::DATASOURCE_VARIABLE: {
				$value = $datapool->variables[$input->name];
				break;
			}
			case Point::DATASOURCE_SELF: {
				$value = property_exists($input,'self')?$input->self:NULL;
				break;
			}
		}
		if ($value === '') $value = NULL;
/*		if (is_null($value) and property_exists($metapoint,'default')) {
			$value = $metapoint->default;
		} */
		if (!is_null($value)) {
			$datatype = $metapoint->datatype;
			switch ($datatype) {
				case Point::DATATYPE_STRING:
					$value = trim($value);
			}
		}
		$point = $pointlist[$pointname] = Point::data($value,$metapoint);
		if (!is_null($this->_input_default_recover_state))
			$point->recoverstate = $this->_input_default_recover_state;
	}
	protected function _input_data_blockselect($pointname,$metapoint,&$pointlist)
	{
		$input = (object) $metapoint->input;
		if (property_exists($input,'constraint')) { // specialized input constraint
			$constraint = $input->constraint;
			if (!$this->datamodel->constraints($constraint)) {
				return;
			}
		}
		if (property_exists($input,'notconstraint')) { // specialized input constraint
			$constraint = $input->notconstraint;
			if ($this->datamodel->constraints($constraint)) {
				return;
			}
		}
		$method = $input->method;
		$datatype = $metapoint->datatype;
		$datapool = $this->datamodel->datapool;
		switch ($method) {
			case Point::DATASOURCE_ARRAY: {
				$name = $input->name;
				$array = $datapool->arrays[$name];
				$index = $input->index;
				if (($name == self::POST_ARRAY) and ($datatype == Point::DATATYPE_BOOLEAN))
					$value = (boolean) (isset($array[$index]) and !empty($array[$index]));
				elseif (isset($array[$index]))
					$value = $array[$index];
				else $value = NULL;
				break;
			}
		}
		if (is_null($value) and property_exists($metapoint,'default')) {
			$value = $metapoint->default;
		}
		if (!is_null($value)) {
			$datatype = $metapoint->datatype;
			switch ($datatype) {
				case Point::DATATYPE_STRING:
					$value = trim($value);
			}
		}
		$point = $pointlist[$pointname] = Point::blockselect($metapoint,$value);
		if (!is_null($this->_input_default_recover_state))
			$point->recoverstate = $this->_input_default_recover_state;
	}
	// possibly no datasource, just a list directly from post
	protected function _input_data_block($pointname,$metapoint,&$pointlist)
	{
		$input = (object) $metapoint->input;
		if (property_exists($input,'constraint')) {
			$constraint = $input->constraint;
			if (!$this->datamodel->constraints($constraint)) {
				return;
			}
		}
		if (property_exists($input,'notconstraint')) { // specialized input constraint
			$constraint = $input->notconstraint;
			if ($this->datamodel->constraints($constraint)) {
				return;
			}
		}
		$block = $pointlist[$pointname] = Point::block($metapoint);
		
		$metadatalist = $metapoint->metadata;
		$blockpointlist = &$block->pointlist;
		$this->_input_data($metadatalist,$blockpointlist);
	}
	// could be datasource, for subarray
	protected function _input_data_list($pointname,$metapoint,&$pointlist)
	{
		$input = (object) $metapoint->input;
		if (property_exists($input,'constraint')) {
			$constraint = $input->constraint;
			if (!$this->datamodel->constraints($constraint)) {
				return;
			}
		}
		if (property_exists($input,'notconstraint')) { // specialized input constraint
			$constraint = $input->notconstraint;
			if ($this->datamodel->constraints($constraint)) {
				return;
			}
		}
		$datapool = $this->datamodel->datapool;
		$blocklist = $pointlist[$pointname] = Point::blocklist($metapoint);
		$metadata = $metapoint->metadata; // for individual list items
		$blocklist_pointlist = &$blocklist->pointlist;
		if (empty($this->datamodel->datapool->arrays[$input->name][$input->index])) { // may be first item
			$this->datamodel->datapool->arrays[$input->name][$input->index] = array();
		}
		$postlist = $this->datamodel->datapool->arrays[$input->name][$input->index];
		$referencevariable = $input->referencevariable;
		$recordstatusindex = $input->recordstatusindex;
		if (empty($postlist[$recordstatusindex]))
			$listproxy = array(); // new dataset
		else
			$listproxy = array_keys($postlist[$recordstatusindex]);
		$collection = array();
		$count = -1;
		foreach ($listproxy as $reference) {
			$count++;
			$subpointlist = array();
			$datapool->variables[$referencevariable] = $reference;
			$this->_input_data($metadata,$subpointlist);
			$listitem = $collection[$reference] = Point::listitem($metadata,$subpointlist);
			$listitem->reference = $reference;
			$listitem->index = $count;
			$listitem->recordstatus = $subpointlist[$recordstatusindex]->value;
		}
		$blocklist->pointlist = $collection;
	}
	#======================= DATAFRAME VALIDATE =================================
	protected function _validate_data($pointlist,$errorindex = NULL)
	{
		$validation = new Validation(array());
		$dictionary = array();
		foreach ($pointlist as $pointname => $datapoint) {
			if ($datapoint instanceof Data) {
				$this->_validate_data_point($pointname,$datapoint,$validation,$dictionary);
			} elseif ($datapoint instanceof Blockselect) {
				$this->_validate_data_blockselect($pointname,$datapoint,$validation,$dictionary);
			} elseif ($datapoint instanceof Blocklist) {
				$this->_validate_data_list($datapoint);
			} elseif ($datapoint instanceof Block) {
				$this->_validate_data_block($datapoint);
			}
		}
		return $this->_run_validation($validation,$dictionary,$pointlist,$errorindex);
	}
	protected function _validate_data_point($pointname,$datapoint,$validation,&$dictionary)
	{
		if (property_exists($datapoint->metapoint,'rules')) {
			$rules = $datapoint->metapoint->rules;
			$validation->rules($pointname,$rules);
		}
		$dictionary[$pointname] = $datapoint->value;
	}
	protected function _validate_data_blockselect($pointname,$datapoint,$validation,&$dictionary)
	{
		if (property_exists($datapoint->metapoint,'rules')) {
			$rules = $datapoint->metapoint->rules;
			$validation->rules($pointname,$rules);
		}
		$dictionary[$pointname] = $datapoint->value;
	}
	protected function _validate_data_block($datapoint)
	{
		$result = $this->_validate_data($datapoint->pointlist);
		if (!$result) $datapoint->errorstate = Point::ERRORSTATE_ERROR;
	}
	protected function _validate_data_list($datapoint)
	{
		$pointlist = $datapoint->pointlist;
		$count = 0;
		foreach ($pointlist as $listitem) {
			$count++;
			if (($listitem->recordstatus == 'changed') or ($listitem->recordstatus == 'error')) {
				$result = $this->_validate_data($listitem->pointlist,$count);
				if (!$result) {
					$listitem->errorstate = Point::ERRORSTATE_ERROR;
					$listitem->recordstatus = 'error';
					$listitem->pointlist[$datapoint->metapoint->recordstatusindex]->value = 'error';
					$listitem->pointlist[$datapoint->metapoint->errorindexindex]->value = $listitem->index + 1;
				}
			}
		}
	}
	protected function _run_validation($validation,$dictionary,$pointlist,$errorindex = NULL) 
	{
		$validation = $validation->copy($dictionary);
		if (!$validation->check()) {
			$errors = $validation->errors($this->_errormessageindex); // looks up messages
			$errorlist = array();
			foreach ($errors as $index => $error) {
				if ($errorindex) $error .= ' ['.$errorindex.']';
				$errorlist[] = array('error' => $error);
				$pointlist[$index]->errorstate = Point::ERRORSTATE_ERROR;
			}
			$errorvar = (array) $this->datamodel->datapool->variables[$this->_errordictionaryindex]->errors;
			$errorvar = array_merge($errorvar,$errorlist);
			$this->datamodel->datapool->variables[$this->_errordictionaryindex]->errors = $errorvar;
			$this->datamodel->errorstate = Point::ERRORSTATE_ERROR;
			return FALSE;
		} else {
			return TRUE;
		}
	}
	#===================== DATAFRAME FINISHINPUT ===============================
	protected function _finishinput_data($pointlist)
	{
		foreach ($pointlist as $datapoint) {
			if ($datapoint instanceof Data) {
				$this->_finishinput_data_point($datapoint);
			} elseif ($datapoint instanceof Blockselect) {
				$this->_finishinput_data_select($datapoint);
			} elseif ($datapoint instanceof Blocklist) {
				$this->_finishinput_data_list($datapoint);
			} elseif ($datapoint instanceof Block) {
				$this->_finishinput_data_block($datapoint);
			}
		}
	}
	protected function _finishinput_data_point($datapoint)
	{
		$value = $datapoint->value;
		$metapoint = $datapoint->metapoint;
		if (!($datapoint->errorstate or $datapoint->recoverstate)) {
			if (!is_null($value)) {
				$datatype = $metapoint->datatype;
				switch ($datatype) {
					case Point::DATATYPE_DATETIME:
					case Point::DATATYPE_DATE:
						$value = strtotime($value);
						if (!$value) $value = NULL;
						break;
					case Point::DATATYPE_BOOLEAN:
						$value = (bool) $value;
						break;
				}
				$datapoint->value = $value;
			}
		}
	}
	protected function _finishinput_data_select($datapoint)
	{
		$value = $datapoint->value;
		$metapoint = $datapoint->metapoint;
		if (!($datapoint->errorstate or $datapoint->recoverstate)) {
			if (!is_null($value)) {
				$datatype = $metapoint->datatype;
				switch ($datatype) {
					case Point::DATATYPE_DATETIME:
					case Point::DATATYPE_DATE:
						$value = strtotime($value);
						if (!$value) $value = NULL;
						break;
					case Point::DATATYPE_BOOLEAN:
						$value = (bool) $value;
						break;
				}
				$datapoint->value = $value;
			}
		}
	}
	protected function _finishinput_data_block($datapoint)
	{
		$pointlist = $datapoint->pointlist;
		if ($pointlist) $this->_finishinput_data($pointlist);
	}
	protected function _finishinput_data_list($datapoint)
	{
		$datalist = $datapoint->pointlist;
		foreach ($datalist as $listitem) {
			$this->_finishinput_data($listitem->pointlist);
		}
	}
	#===================== DATAFRAME RECOVERINPUT ===============================
	protected function _recoverinput_data($metadatalist,&$pointlist)
	{
		foreach ($metadatalist as $pointname => $metapoint) {
			if (property_exists($metapoint,'constraint')) {
				$constraint = $metapoint->constraint;
				if (!$this->datamodel->constraints($constraint)) {
					continue;
				}
			}
			if (property_exists($metapoint,'notconstraint')) {
				$constraint = $metapoint->notconstraint;
				if ($this->datamodel->constraints($constraint)) {
					continue;
				}
			}
			if (property_exists($metapoint,'token')) {
				$accessdata = (object) $metapoint->access;
				if (isset($accessdata->token)) {
					$token = $accessdata->name;
					$requires = $accessdata->requires;
					$access = Access::instance();
					if (!$access->permission($token,$requires,$this->datamodel->ownerid,$this->datamodel->articleid)) {
						continue;
					}
				} elseif (isset($accessdata->permission)) {
					$permission = 
						Node::get_permission($this->datamodel,$accessdata->permission);
					if ($permission===FALSE)
						continue;
				}
			}
			if ($metapoint instanceof Meta) {
				if (property_exists($metapoint,'recover')) { // to be created
						$this->_recoverinput_data_point($pointname,$metapoint,$pointlist);
				}
			} elseif ($metapoint instanceof Blockselect) {
				$this->_recoverinput_data_select($pointname,$metapoint,$pointlist);
			} elseif ($metapoint instanceof Blocklist) {
				$this->_recoverinput_data_list($pointname,$metapoint,$pointlist);
			} elseif ($metapoint instanceof Block) {
				$this->_recoverinput_data_block($pointname,$metapoint,$pointlist);
			}
		}
	}
	protected function _recoverinput_data_point($pointname,$metapoint,&$pointlist)
	{
		if (property_exists($metapoint,'recover'))
		{
			$recover = (object) $metapoint->recover;
			$method = $recover->method;
			switch ($method) {
				case Point::DATASOURCE_ARRAY: {
					$name = $recover->name;
					$index = $recover->index;
					$value = $this->datamodel->datapool->arrays[$name][$index];
					break;
				}
			}
			if (!is_null($value)) {
				if (property_exists($recover,'filter')) {
					$filter = (object) $recover->filter;
					switch ($filter->method) {
						case Point::FILTER_COMPARE: {
							$arguments = (object) $filter->arguments;
							$value = ($value === $arguments->compare);
							break;
						}
						case Point::FILTER_LOOKUP: {
							$arguments = (object) $filter->arguments;
							$record = DB::select($arguments->index,$arguments->value)
								->from($arguments->name)
								->where($arguments->index,'=',$value)
								->as_object()
								->execute($this->_database)->current();
							if ($record) {
								$field = $arguments->value;
								$value = $record->$field;
							}
							break;
						}
					}
				}
			}
			if (isset($pointlist[$pointname])) {
				$pointlist[$pointname]->value = $value;
			} else {
				$pointlist[$pointname] = Point::data($value,$metapoint);
			}
			$point = $pointlist[$pointname];
			$point->recoverstate = Point::RECOVERSTATE_RECOVER; // suppress re-format for output
		}
	}
	protected function _recoverinput_data_select($pointname,$metapoint,&$pointlist)
	{
		if //((!isset($pointlist[$pointname])) and 
			(property_exists($metapoint,'recover'))//)
		{
			$recover = (object) $metapoint->recover;
			$method = $recover->method;
			switch ($method) {
				case Point::DATASOURCE_ARRAY: {
					$name = $recover->name;
					$index = $recover->index;
					$value = $this->datamodel->datapool->arrays[$name][$index];
					break;
				}
			}
			if (!isset($pointlist[$pointname]))
				$pointlist[$pointname] = Point::blockselect($metapoint,$value);
			else
				$pointlist[$pointname]->value = $value;
			$point = $pointlist[$pointname];
			$point->recoverstate = Point::RECOVERSTATE_RECOVER; // suppress re-format for output
		}
	}
	protected function _recoverinput_data_block($pointname,$metapoint,&$pointlist)
	{
		if (!isset($pointlist[$pointname])) 
			return;
		else
			$block = $pointlist[$pointname];
		
		$metadatalist = $metapoint->metadata;
		$blockpointlist = &$block->pointlist;
		$this->_recoverinput_data($metadatalist,$blockpointlist);
	}
	protected function _recoverinput_data_list($pointname,$metapoint,&$pointlist)
	{
		if (!isset($pointlist[$pointname])) 
			return;
		else
			$blocklist = $pointlist[$pointname];
		$metadatalist = $metapoint->metadata;
		$blockitemlist = &$blocklist->pointlist;
		foreach ($blockitemlist as $listitem) {
			$sublist = &$listitem->pointlist;
			$this->_recoverinput_data($metadatalist,$sublist);
		}
	}
	#========================= DATAFRAME STORE ==================================
	protected function _store_data($pointlist)
	{
		foreach ($pointlist as $datapoint) {
			$metapoint = $datapoint->metapoint;
			if ($datapoint instanceof Data) {
				if (property_exists($metapoint,'store')) {
					$this->_store_data_point($datapoint);
				}
			} elseif ($datapoint instanceof Blockselect) {
				if (property_exists($metapoint,'store')) {
					$this->_store_data_point($datapoint); // same as datapoint
				}
			} elseif ($datapoint instanceof Blocklist) {
				$this->_store_data_list($datapoint);
			} elseif ($datapoint instanceof Block) {
				$this->_store_data_block($datapoint);
			}
		}
	}
	protected function _store_data_point($datapoint)
	{
		$metapoint = $datapoint->metapoint;
		$store = (object) $metapoint->store;
		$method = $store->method;
		$datamodel = $this->datamodel;
		$datatype = $metapoint->datatype;
		$value = $datapoint->value;
		if (!is_null($value)) {
			if (property_exists($store,'filter')) {
				$filter = (object) $store->filter;
				switch ($filter->method) {
					case Point::FILTER_LOOKUP: {
						$arguments = (object) $filter->arguments;
						$record = DB::select($arguments->index,$arguments->value)
							->from($arguments->name)
							->where($arguments->index,'=',$value)
							->as_object()
							->execute($this->_database)->current();
						if ($record) {
							$field = $arguments->value;
							$value = $record->$field;
						} else {
							$value = NULL;
						}
						break;
					}
				}
			}
		}
		if (!is_null($value)) {
			switch ($datatype) {
				case Point::DATATYPE_DATETIME:
				case Point::DATATYPE_DATE:
					$value = Database::instance($this->_database)->sql_datetime($value);
					break;
				case Point::DATATYPE_BOOLEAN:
					$value = (bool) $value;
					break;
			}
		}
		if (property_exists($store,'filter')) {
			$filter = (object) $store->filter;
			switch ($filter->method) {
				case Point::FILTER_HASHPASSWORD: {
					$value = User::instance()->hash_password($value);
					break;
				}
				case Point::FILTER_PURIFYTRUSTED: {
					// csstidy to allow styles to be extracted - purifier instantiates even if not used
					require_once MODPATH.'purifier/vendor/csstidy/class.csstidy.php'; 
					// purifier replaces & with &amp; (entity), but google embedded charts require unaltered & in json data
					// .. so hide and then restore amp chars and amp entities
					$value = strtr($value,array('&amp;'=>'__AMP__ENTITY__','&'=>'__AMP__CHAR__')); // hide amp entity and char from purify
					$value = Security::xss_clean($value);
					$value = strtr($value,array('__AMP__ENTITY__'=>'&amp;','__AMP__CHAR__'=>'&')); // restore amp entity and char
					$purifier = Security::htmlpurifier();
					// restore style blocks which purifier strips out
					$styles = $purifier->context->get('StyleBlocks');
					if (!empty($styles)) {
						$styletext = '<style>';
						foreach ($styles as $style)
							$styletext .= $style;
						$styletext .= '</style>';
						$value = $styletext . $value;
					}
					break;
				}
			}
		}
		switch ($method) {
			case Point::DATATARGET_RECORD: {
				$name = $store->recordindex;
				$field = $store->field;
				if ($datamodel->recordcaching()) {
					$record = $datamodel->cache->records[$name];
				} else {
					$record = $datamodel->datapool->records[$name];
				}
				$record->$field = $value;
				break;
			}
			case Point::DATATARGET_RECORD_JSON: {
				$name = $store->recordindex;
				$field = $store->field;
				$json = $store->json;
				if ($datamodel->recordcaching()) {
					$record = $datamodel->cache->records[$name];
				} else {
					$record = $datamodel->datapool->records[$name];
				}
				if (!$record->$field) $record->$field = new StdClass;
				$record->$field->$json = $value;
				break;
			}
		}
	}
	protected function _store_data_block($datapoint)
	{
		$pointlist = $datapoint->pointlist;
		if (!empty($pointlist)) {
			$this->_store_data($pointlist);
		}
	}
	// ** should be clarified for alternate datatargets
	protected function _store_data_list($datapoint)
	{
		$metapoint = $datapoint->metapoint;
		$pointlist = $datapoint->pointlist;
		$store = (object) $metapoint->store;
		$method = $store->method;
		$recordvariable = $store->recordvariable;
		switch ($method) {
			case Point::DATATARGET_RECORDSET: {
				if ($this->datamodel->recordcaching()) {
					$datastore = $this->datamodel->cache;
				} else {
					$datastore = $this->datamodel->datapool;
				}
				$datastore->recordsets[$store->tableindex] = array();
				$storelist = &$datastore->recordsets[$store->tableindex];
				break;
			}
		}
		foreach ($pointlist as $reference => $listitem) {
			if ($listitem->recordstatus == 'changed') {
				$newrecord = $datastore->records[$store->recordvariable] = new StdClass;
				$this->_store_data($listitem->pointlist);
				$storelist[$listitem->reference] = $newrecord;
			}
		}
		unset($datastore->records[$store->recordvariable]);
	}
	#========================================================================
	#-------------------------[ CONTAINTERFRAME **deferred - may not be implemented** ]----------------------------
	#========================================================================
	protected function _process_containerframe()
	{
		$frame = $this->_frame;
		$metadata = $frame->metadata;
		$this->_input_container($metadata,$frame);
	}
	protected function _input_container($metadata,&$pointlist)
	{
/*		foreach ($metadata as $pointname => $metapoint) {
			if (property_exists($metapoint,'input')) {
				$this->_input_container_point($pointname,$metapoint,$pointlist);
			}
		}
*/
	}
	# CONTAINTERFRAME INPUT DATA
	protected function _input_container_point($pointname,$metadata,&$pointlist)
	{
	}
}
