<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster;

use StdClass;
use Kohana, DB, Services;
use User;
use Muster\Node\Type;
	
defined('SYSPATH') or die('No direct script access.');

/*
The mandate of this module is to take node parameters as input as
return dictionary data as output.
A node can be:
	- folder
	- document
	- blog
	- blogentry
	- cluster
	- page
	
Notes:
	- authorid => Ownerid in url query
	
	object types:
	document
	entry
	page
	blog
	folder
	image
	xdoc
	link
	contact
*/
abstract class Node {

	const FLOW_CONTINUE = TRUE;
	const FLOW_REDIRECT = FALSE;
	
	static $node_instances;

	protected $_controller;
	protected $_user;
	protected $_extensionflag = FALSE;
	public $_permissions;
	# =====================================================================
	# ----------------------------- STATIC --------------------------------
	# =====================================================================
	# utilities
	// ** obsolete?
	public static function set_query_constraints($datamodel,$query,$tablename)
	{
		$node = self::$node_instances[spl_object_hash($datamodel)];
		$node->_set_query_constraints($query,$tablename); // allows for override
	}
	// recordset callback from gateway
	public static function set_query_mixed_constraints($datamodel,$query,$tablename)
	{
		$node = self::$node_instances[spl_object_hash($datamodel)];
		$node->_set_query_mixed_constraints($query,$tablename); // allows for override
	}
	// uses current instance permissions
	public static function set_table_constraints($datamodel,$tableindex)
	{
		$node = self::$node_instances[spl_object_hash($datamodel)];
		$table = $datamodel->datapool->tables[$tableindex];
		$query = $table->query();
		$tablename = $table->tablename();
		$node->_set_query_constraints($query,$tablename); // allows for override
	}
	// uses universal mixed permissions
	public static function set_table_mixed_constraints($datamodel,$tableindex)
	{
		$node = self::$node_instances[spl_object_hash($datamodel)];
		$table = $datamodel->datapool->tables[$tableindex];
		$query = $table->query();
		$tablename = $table->tablename();
		$node->_set_query_mixed_constraints($query,$tablename); // allows for override
	}
	public static function get_permissions($datamodel)
	{
		$node = self::$node_instances[spl_object_hash($datamodel)];
		return $node->_permissions;
	}
	public static function get_permission($datamodel,$permission)
	{
		$permissions = self::get_permissions($datamodel);
		return isset($permissions->$permission)?$permissions->$permission:NULL;
	}
	public static function register_node($node)
	{
		self::$node_instances[spl_object_hash($node->datamodel())] = $node;
	}
	# =====================================================================
	# ----------------------------- INSTANCE ------------------------------
	# =====================================================================
	// constructor
	public function __construct($controller)
	{
		$this->_controller = $controller;
		$this->_user = User::instance();
	}
	// access
	public function datamodel()
	{
		return $this->_datamodel;
	}
	
	public function context()
	{
		return $this->_context;
	}
	// operation
	public function extend_context($sub_context)
	{
		$sub = array($sub_context);
		$context = $this->_context;
		foreach ($sub_context as $property => $value) {
			$context->$property = $value;
		}
	}

	# ==========================[ UTILITIES ]===========================================
	// can over_ride in sub_class...
	// this is called from gateway through set_folder_query static above
	// the purpose is to restore basic rights filtering when master record changes
	// THIS METHOD ASSUMES THAT ALL QUERY TABLES HAVE TABLE.ID=ARTICLEID
	public function _set_query_constraints($query,$tablename)
	{
//		$permissions = $this->_permissions;
//		$this->_set_recordset_constraints($query,$tablename,$permissions);
		$this->_set_recordset_constraints($query,$tablename);
	}
	public function _set_query_mixed_constraints($query,$tablename)
	{
		$this->_set_recordset_constraints($query,$tablename);
	}
	protected function _set_recordset_constraints($query,$tablename)
	{
		$user = User::instance();
		$query->where_open() // must be readable
			->where(NULL,'EXISTS',
				DB::select('RecordType')->from('temp_permissions_content')
					->where('RecordType','=',DB::expr('`'.$tablename.'`.`ObjectType`'))
					->where('Read','IS',DB::expr('TRUE')
				)
			);
		$query->where_close(); 

		if (!$user->authenticated()) {
			$query->where_open()
				->where('Privileged','IS',DB::expr('FALSE'))
			->where_close();
		}
		$query->where_open();
			$query->where_open(); // unpublished
				$query->where_open()
					->where('Approval','=','A')
					->or_where('Approval','=','P')
				->where_close();
				$query->or_where_open()
					->where(NULL,'EXISTS',
						DB::select('RecordType')->from('temp_permissions_unpublished')
							->where('RecordType','=',DB::expr('`'.$tablename.'`.`ObjectType`'))
							->where('Read','IS',DB::expr('TRUE'))
					)
				->or_where_close();
			$query->where_close();

			$query->where_open(); // hidden
				$query->where_open()
					->where('Hide','IS',DB::expr('FALSE'))
				->where_close();
				$query->or_where_open()
					->where(NULL,'EXISTS',
						DB::select('RecordType')->from('temp_permissions_hidden')
							->where('RecordType','=',DB::expr('`'.$tablename.'`.`ObjectType`'))
							->where('Read','IS',DB::expr('TRUE'))
					)
				->or_where_close();
			$query->where_close();
			
			$query->where_open(); // trash
				$query->where_open()
					->where('Trash','IS',DB::expr('FALSE'))
				->where_close();
				$query->or_where_open()
					->where(NULL,'EXISTS',
						DB::select('RecordType')->from('temp_permissions_trash')
							->where('RecordType','=',DB::expr('`'.$tablename.'`.`ObjectType`'))
							->where('Read','IS',DB::expr('TRUE'))
					)
				->or_where_close();
			$query->where_close();
				
			if ($user->authenticated()) { // ... but allow visibility of owned or shared records
				$query->or_where_open()->where('AuthorID','=',$user->userid());
					$query->or_where_open()
						->or_where(DB::expr($user->userid()),'IN',
						DB::expr('(Select `Articles_Authors`.`UserID` from `Articles_Authors` where `Articles_Authors`.`ArticleID` = `'
						. $tablename . '`.`ID`)'))
					->or_where_close();
				$query->or_where_close();
			}
		$query->where_close();
//		var_dump($query->sql());
	}
	protected function _set_datapool_error($parmdata,$messageblockindex = 'messagesection')
	{
		extract($parmdata);
		if (empty($datapool)) return;
		if (empty($error_info)) return;
		(!empty($errorfile)) or ($errorfile = NULL); // ** set to configure default value?
		if (is_string($error_info))
		{
			$error_info = array($messageindex,'internal',$error_info);
		} else {
			if (!isset($error_info[2])) {
				if ((int) $error_info[0] == 0) 
					$error_info[2] = 'Bogus message';
				else
					$error_info[2] = 'Unknown message';
			}
		}
		$index = $error_info[0];
		$error_message = $errorfile?Kohana::message($errorfile,$index):NULL;
		if (!$error_message) $error_message = $error_info[2];
		$this->_set_datapool_message($datapool,$error_message,'errors',$messageblockindex);
	}
	protected function _set_datapool_message($datapool,$message,$messageset,$messageblock = 'messagesection')
	{
		if (empty($datapool)) return;
		switch ($messageset) {
			case 'messages':
				$messageindex = 'message';
				break;
			case 'warnings':
				$messageindex = 'warning';
				break;
			case 'errors':
				$messageindex = 'error';
				break;
			default:
				return;
		}
		empty($datapool->variables[$messageblock]) and ($datapool->variables[$messageblock] = new StdClass);
		$block = $datapool->variables[$messageblock];
		(!property_exists($block,$messageset)) and ($block->$messageset = array());
		$set = &$block->$messageset;
		$set[] = array($messageindex => $message);
	}
	protected function _get_datapool_messageblock($datapool,$messageblock = 'messagesection')
	{
		if (empty($datapool)) return; // error
		if (empty($datapool->variables[$messageblock])) return new StdClass;
		return $datapool->variables[$messageblock];
	}
	// @returns void
	protected function _record_report_dberror($datamodel,$error_info,$messageblock = 'messagesection',$messagefile = 'article')
	{
		if (empty($error_info)) return;
		if (is_string($error_info))
		{
			$error_info = array('error','internal',$error_info);
		} else {
			if (!isset($error_info[2])) {
				if ((int) $error_info[0] == 0) 
					$error_info[2] = 'Bogus error';
				else
					$error_info[2] = 'Unknown error';
			}
		}
		$index = $error_info[0];
		$error_message = Kohana::message($messagefile,$index);
		if (!$error_message) $error_message = $error_info[2];
		$this->_set_datapool_message($datamodel->datapool,$error_message,'errors',$messageblock);
	}
	// @returns constant Node::FLOW_CONTINUE
	protected function _record_report_access_error($datamodel,$messageblock = 'messagesection',$messagefile = 'article')
	{
		$command = $this->_parameters->command;
		if (!$command) return;
		$error_message = Kohana::message($messagefile,$command . 'error');
		if (!$error_message) $error_message = $command . ' is unavailable';
		$this->_set_datapool_message($datamodel->datapool,$error_message,'errors',$messageblock);
		return Node::FLOW_CONTINUE; // recycle process
	}
	// ** this function name is deprecated **
	public function _record_set_folder_query($query,$tablename) // apply permissions
	{
		$this->_set_query_constraints($query,$tablename);
	}
	// ** this function name is deprecated
	public static function set_folder_query($datamodel,$tableindex)
	{
		self::set_table_constraints($datamodel,$tableindex);
	}
}