<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster;

use Kohana, Database, Kohana_Exception;
use User, Services;
use 
	Muster\Tablemodel\Tabledef;

defined('SYSPATH') or die('No direct script access.');
/***********************************
Development Notes:
- get default audit fields from config file
- test and review all error messages
- verify messages are in valid file location and file format
**************************************/

/*
 file: tablemodel.php
 package: Tablemodel
 category: ORM (Object Relational Manager)
 author: Henrik Bechmann, bechmann.ca
 copyright:  (c) 2009-2010 Henrik Bechmann, Toronto Canada. All rights reserved
 licence: BSD Licence, see licence.txt
 date: August 2010
 description: master class; 
	- contains model and instances of forward and backward linked heirarchical table data
	- based on schema graph
	- dynamic cascading synchronized foreign key updates of model data
GENERAL API
	->__construct($parameters)
	# builder methods
	::factory($modelclass,$parameters = array()) static factory class, builder method
	->generate($mastertable = NULL)
	# accessors
	->db()
	->userid($userid = NULL) // get or set
	->table($tablename = NULL, $instance = 0) // master or selected
	->query($tablename = NULL, $instance = 0) // master or selected
	->activerecord($tablename = NULL, $instance = 0) // master or selected
	->recordset($tablename = NULL, $instance = 0) // master or selected
	# advanced (debug)
	->debug_level()
	->debug_display_table_tree()
	->debug_display_observers_list() 
	----------------------------------------------
*/
class Tablemodel
{
	// relation types
	const ONE_TO_ONE = 1; // 1:1
	const ONE_TO_MANY = 2; // 1:M
	const MANY_TO_ONE = 3; // M:1
	// index types
	const UNIQUE = 1;
	const INDEX = 2;

	protected $_db; // shared database connection
	protected $_master_tabledef_name; // name of current master table (root of hierarchy)
	protected $_root_table; // current root table by reference
	protected $_tabledefs = array() ; // associative array of table names and tabledef objects
	protected $_userid; // userid to stamp on audit fields (CreateUserID, EditUserID)
	protected $_config;

	protected $_debug_level; // from tablemodel config
	
/*	public static function factory($modelclass,$parameters = array())
	{
		$class = 'Model_' . $modelclass;
		return new $class($parameters);
	} */
	// definition = $parameters is required at creation time
	public function __construct($parameters, $parmadjustments = NULL)
	{
		if (is_string($parameters))
			$parameters = self::get_parameters($parameters);
		if ($parmadjustments)
			$parameters = $parmadjustments + $parameters;
		$config = $this->_config = Kohana::$config->load('tablemodel');
		$this->_debug_level = $config->debug_level;
		// check parameters
		$parameters = (object) $parameters;
		$this->_validate_properties($parameters);
		$this->_assert_properties($parameters);
		//set properties
		$this->_db = Database::instance($parameters->database);
		$mastertable = $this->_master_tabledef_name = $parameters->mastertable;
		$this->_userid = User::instance()->userid(); // $parameters->userid;
		if (is_null($this->_userid)) $this->_userid = 0;
		// create schema: tabledefs without buffers
		$tabledefinitions = (object) $parameters->tables;
		$this->_create_model($tabledefinitions);
		if ($mastertable) $this->generate();
	}
	
	public static function get_parameters($filename)
	{
		$file = Kohana::find_file('schema',$filename);
		if ($file)
			return Kohana::load($file);
		else
			return array();
	}
	
	protected function _validate_properties($parameters)
	{
		if ($this->_debug_level > 0) 
		{
			// check for required parameters
			$result = Services::verify_properties
				(
					$parameters,
					array
					(
						'mastertable',
						'tables'
					)
				);
			if ($result !== TRUE)
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','tablemodel.missing_parameters'),
							array(':parameters' => implode(',',$result))
					);
			}
		}
	}

	protected function _assert_properties($parameters)
	{
		$user = User::instance();
		if ($user->authenticated())
			$userid = $user->userid();
		else
			$userid = 0;
		Services::assert_properties
			(
				$parameters,
				array
				(
					'database' => Database::$default,
					'userid' => $userid,
				)
			);
	}		
	// creates table, index, and Relation definitions, but not data buffers
	protected function _create_model($tabledefinitions) 
	{
		// create table definition objects.
		// Relation definitions are created next step when all tabledefs and indexes are available for cross referencing
		$relations_required = (boolean) (count($tabledefinitions) > 1);
		//create tabledefs
		foreach ($tabledefinitions as $tabledef => $definition)
		{
			$definition = (object) $definition;
			$this->_validate_tabledef_properties($tabledef, $definition, $relations_required);
			$this->_assert_tabledef_properties($tabledef,$definition);
			// add created table object, with index definitions, to list of table objects
			$this->_tabledefs[$tabledef] = new Tabledef($definition);
		}
		//iterate through table dictionary to create Relation definitions, 
		//	ready for reciprocal Relation cross-referencing next step -
		// 	this requires access to other table defs, so run after table creation above
		$tabledefs = $this->_tabledefs;
		foreach ($tabledefs as $tabledef) 
		{
			$tabledef->create_relation_definitions();
		}
		//iterate through table dictionary to set reciprocal Relations, 
		//	now that all Relation definitions are available - makes the schema a graph
		foreach ($tabledefs as $tabledef) 
		{
			$tabledef->set_reciprocal_relationdefs();
		}
	}

	protected function _validate_tabledef_properties($tabledef, $definition, $relations_required)
	{
		if ($this->_debug_level > 0) 
		{
			// check for required parameters
			$required_parameters = array
				(
					'primary_index',
					'indexes'
				);
			//Relations are required where there is more than one table
			if ($relations_required)
			{
				$required_parameters[] = 'relations';
			}
			$result = Services::verify_properties
				(
					$definition,
					$required_parameters
				);
			if (!($result === TRUE))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','tablemodel.missing_tabledef_parameters'),
							array(':parameters' => implode(',',$result),':table' => $tabledef)
					);
			}
		}
	}

	protected function _assert_tabledef_properties($tabledef,$definition)
	{
		Services::assert_properties
			(
				$definition,
				array
				(
					'tablemodel' => $this,
					'name' => $tabledef,
					// schema name, or tablename if schema name is alias
					'tablename' => property_exists($definition,'tablename')
						? $definition->tablename
						: $tabledef,
					'validation' => NULL,
					'audit' => NULL, // will use configuration
					'concurrency' => NULL, // will use configuration
					'perpage' => NULL, // will use default if available
					'orderby' => NULL, // will use alternate-index if available
				)
			);
		return $definition;
	}
	
	// call to change root table - will automatically destroy previous settings
	protected function create_buffers($master_table_name)
	{
		$this->_validate_master_tablename($master_table_name);
		// previous tree needs to be cleaned out first
		if ($this->_root_table)
		{
			$this->destroy_buffers();
			$this->_root_table = NULL;
		}
		// create data tree
		$root_table = $this->_root_table = 
			$this->_tabledefs[$master_table_name]->create_table();
		// relation tables in hierarchy, set recursively
		$root_table->set_children();
	}
	// used internally for cleanup
	public function destroy_buffers()
	{
		foreach ($this->_tabledefs as $tabledef)
		{
			$tabledef->destroy_tables();
		}
	}
	
	protected function _validate_master_tablename($master_table_name)
	{
		if ($this->_debug_level > 0) 
		{
			if (!isset($this->_tabledefs[$master_table_name]))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','tablemodel.invalid_master_table'),
							array(':table' => $master_table_name)
					);
			}
		}
	}
	
	public function tabledefs()
	{
		return $this->_tabledefs;
	}
	// get master or selected tabledef
	public function tabledef($tablename = NULL)
	{
		if (!($tablename === NULL))
		{
			return isset($this->_tabledefs[$tablename])
				? $this->_tabledefs[$tablename]
				: NULL;
		}
		else
		{
			return $this->_root_table->tabledef();
		}
	}
	
	public function config()
	{
		return $this->_config;
	}
	
	// [--------------------------- Table Model GENERAL API ---------------------------]
	#builder methods
	public function generate($mastertable = NULL)
	{
		if (!is_null($mastertable))
			$this->_master_tabledef_name = $mastertable;
		$this->create_buffers($this->_master_tabledef_name);
		return $this;
	}
	# acessor methods
	public function master_tablename()
	{
		return $this->_master_tabledef_name;
	}
	public function db()
	{
		return $this->_db;
	}
	// get or set user id for audit trail
	public function userid($userid = NULL)
	{
		if (!is_null($userid))
		{
			$this->_userid = $userid;
		}
		return $this->_userid;
	}
	// get master or selected table
	public function table($tablename = NULL, $n = 0)
	{
		if (!($tablename === NULL))
		{
			if (isset($this->_tabledefs[$tablename]))
			{
				$tables = $this->_tabledefs[$tablename]->tables();
				return $tables[$n];
			}
			else
			{
				return NULL;
			}
		}
		else
		{
			return $this->_root_table;
		}
	}
	
	// get master or selected query
	public function query($tablename = NULL, $n = 0)
	{
		if (!($tablename === NULL))
		{
			if (isset($this->_tabledefs[$tablename]))
			{
				$tables = $this->_tabledefs[$tablename]->tables();
				return $tables[$n]->query();
			}
			else
			{
				return NULL;
			}
		}
		else
		{
			return $this->_root_table->query();
		}
	}
	// get master or selected activerecord
	public function activerecord($tablename = NULL, $n = 0)
	{
		if (!($tablename === NULL))
		{
			if (isset($this->_tabledefs[$tablename]))
			{
				$tables = $this->_tabledefs[$tablename]->tables();
				return $tables[$n]->activerecord();
			}
			else
			{
				return NULL;
			}
		}
		else
		{
			return $this->_root_table->activerecord();
		}
	}
	// get master or selected recordset
	public function recordset($tablename = NULL, $n = 0)
	{
		if (!($tablename === NULL))
		{
			if (isset($this->_tabledefs[$tablename]))
			{
				$recordsets = $this->_tabledefs[$tablename]->recordsets();
				return $recordsets[$n]->recordset();
			}
			else
			{
				return NULL;
			}
		}
		else
		{
			return $this->_root_table->recordset();
		}
	}


	# Debug
	public function debug_level()
	{
		return $this->_debug_level;
	}
	// internally called
	public static function debug_observer_list($observers)
	{
		$array = array();
		foreach($observers as $observer)
		{
			if ($observer instanceOf Tablemodel_Observed)
			{
				$content = $observer->debug_observer_list();
			}
			else
			{
				$content = '(not observer interface)';
			}
			$array[$observer->observer_name() . ' (' . get_class($observer) . ')'] = $content;
		}
		if (!$array)
		{
			$array = '(leaf)';
		}
		return $array;
	}
	// api call for debug
	public function debug_display_table_tree()
	{
		$array[$this->_root_table->observer_name()] = 
			$this->_root_table->debug_child_name_list();
		Services::debug($array);
	}
	// api call for debug
	public function debug_display_observers_tree() 
	{
		$observed = $this->_root_table;
		$array[$observed->observer_name() . ' ('. get_class($observed).')']=$observed->debug_observer_list();
		Services::debug($array);
	}	
} // End Tablemodel
