<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster\Tablemodel;

use StdClass;
use DB;
use Tablemodel, Services;

defined('SYSPATH') or die('No direct script access.');
/*
 file: activerecord.php
 package: Tablemodel
 category: ORM (Object Relational Manager)
 date: August 2010
 description: table cursor, and used to insert records.
 class: Activerecord (current record of Table)
 observes its Table
 observed by its Indexes
API:
	# builder methods
	->insert($data = FALSE) // returns $this
	->update($data) // returns $this; or assign values to record column names: $activerecord->ColumnName = $value;
	->load() // runs current, but returns activerecord
	# operations
	->save($validation = NULL) // returns TRUE or FALSE
	->delete() // returns TRUE or FALSE; TRUE if no record existed
	# data access
	->fetch() // returns record object or NULL; fetches next (or first) record. typical use: while ($record = $activerecord->fetch())
	->current() // returns record object or NULL; gets current record (including first record)
	->working_record() // record for insert, update, or delete
	# accessors
	->validation($validation = NULL) // get or set
	->table() 
	->operation()
	->BOF() // TRUE or FALSE // deprecated
	->EOF() // TRUE or FALSE // deprecated
	# posting side effects
	->error_info() // array ($sqlerror, $driver_error, $message)
	->insert_id() // id or NULL
*/
class Activerecord implements Observer, Observed // observes underlying table, notifies its Indexes
{
	protected $_is_dirty = FALSE;
	protected $_is_insert = FALSE;
	// Indexes synchronized to Activerecord
	// references
	protected $_observers = array(); // Indexes
	protected $_table; // object reference
	protected $_columndefs; // filter assignments in __set
	protected $_joincolumns; // disallow assignments in __set
	protected $_result; // query result object
	protected $_operation; // insert, delete, update result object
	protected $_record; // raw record object
	protected $_pending_post; // new column values for posting, whether existing or inserted record
	protected $_working_record; // working copy of pending_post for save()
	protected $_validation; // kohana Validate object, from table object
//	protected $_last_post; // last posted pending post
	protected $_record_indexes = array(); // associative array of index names and index activerecords
	protected $_relations = array(); // associative array of relation names and relation activerecords; notified by activerecord indexes
	protected $_insert_id = NULL; // for new records
	protected $_error_info; // standard pdo error structure
	protected $_controls; // audit and concurrency fields
	protected $_query; // TableModel_Query
	protected $_parameters; // insert parameters, saved for debugging
	protected $_failed_insert; // flags condition of being unable to synchronize parent
	
	public function __construct($definition)
	{
		$table = $this->_table = $definition->table;
		$this->_columndefs = $table->tabledef()->columnspecs();
		$this->_joincolumns = $table->tabledef()->join_columns();
		$this->_table->attach_observer($this); // activerecord observes table (for reset)
		$this->_validation = $table->tabledef()->validation(); // use tabledef's validation object
		//create and store references to Indexes
		$indexdefs = $this->_table->tabledef()->indexdefs();
		foreach ($indexdefs as $indexdef)
		{
			$parameters = new StdClass;
			$parameters->activerecord = $this;
			$parameters->indexdef = $indexdef;
			$this->_record_indexes[$indexdef->name()] =	new Index($parameters);
		}
		// create relations and if necessary remote tables
		$relationdefs = $this->_table->tabledef()->relationdefs();
		foreach ($relationdefs as $relationdef)
		{
			$remote_tabledef = $relationdef->remote_indexdef()->tabledef();
			$remote_tabledef_name = $remote_tabledef->tablealias();
			$is_ancestor = $table->is_ancestor($remote_tabledef_name);
			if (!$is_ancestor)
			{
				if ($this->allow_expansion($relationdef))
				{
					// create remote (ie child) table
					$remote_table = $remote_tabledef->create_table($table,$relationdef);
					
					// create remote Foreignkey
					$parameters = new StdClass;
					$parameters->indexdef = $relationdef->remote_indexdef();
					$parameters->table = $remote_table;
					$remote_table_foreignkey = new Foreignkey($parameters);
					
					// create local Recordrelation
					$parameters = new StdClass;
					$parameters->relationdef = $relationdef;
					$parameters->recordindex = $this->_record_indexes[$relationdef->indexdef()->name()];
					$parameters->foreignkey = $remote_table_foreignkey;
					$relation = $this->_relations[$relationdef->name()] = new Relation($parameters);
					
					// link relation to foreign key
					$remote_table_foreignkey->set_relation($relation);
				}
			}
		}
	}
	// constrains redundant expansion of children in tree
	protected function allow_expansion($relationdef)
	{
		if (!$this->_table->parent()) 
			return TRUE;
		elseif ($this->_table->parent_relation_type() === Tablemodel::ONE_TO_MANY) 
			return TRUE;
		elseif ($relationdef->relation_type() === Tablemodel::ONE_TO_MANY) 
			return !($relationdef->terminate_expansion());
		else
			return TRUE;
	}
	
	public function destroy()
	{
		foreach ($this->_relations as $recordrelation)
		{
			$recordrelation->destroy();
		}
		$this->_relations = NULL;
		foreach ($this->_record_indexes as $recordindex)
		{
			$recordindex->destroy();
		}
		$this->_record_indexes = NULL;
		
		$this->_observers = NULL;
		$this->_table = NULL;
		$this->_result = NULL;
		$this->_operation = NULL;
		$this->_validation = NULL;
		$this->_query = NULL;
	}
	
	// observer methods...
	public function update_observer(Observed $table) // Observer method
	{
		$this->_record = NULL;
		$this->_result = $table->result();
		$this->_operation = NULL;
		$this->_is_dirty = FALSE;
		$this->_is_insert = FALSE;
		$this->_pending_post = NULL;
		$this->_operation = NULL;
		$this->notify_observers();
	}
	public function attach_observer(Observer $recordindex) // Observed method
	{
		if (! in_array($recordindex, $this->_observers))
		{
			$this->_observers[] = $recordindex;
		}
	}
	public function detach_observer(Observer $recordindex) // Observed method
	{
		$pos = array_search($recordindex,$this->_observers);
		if (!($pos === FALSE))
		{
			array_splice($this->_observers,$pos,1);
		}
	}
	public function notify_observers() // Observed method; notify recordindexes
	{
		foreach ($this->_observers as $observer)
		{
			$observer->update_observer($this);
		}
	}
	public function observer_name() // TableModelObserver method
	{
		return $this->_table->tabledef()->tablealias();
	}
	public function debug_observer_list() // TableModelObserved method
	{
		return TableModel::debug_observer_list($this->_observers);
	}
	// open calls to object properties of column names
	public function __get($column_name)
	{
		// if in insert mode, look in pending post record, and return
		if ($this->_is_insert)
		{
			if (property_exists($this->_pending_post,$column_name))
				return $this->_pending_post->$column_name;
			else
				return NULL;
		}
		// if not insert, look in existing record
		if ($this->_record and property_exists($this->_record,$column_name))
		{
			if ($this->_pending_post and property_exists($this->_pending_post,$column_name))
			{
				return $this->_pending_post->$column_name;
			}
			else
			{
				return $this->_record->$column_name;
			}
		}
		// if neither is found, return null
		else
		{
			return NULL;
		}
	}
	// get original value; new values are stored in pending_post
	public function get_old($column_name)
	{
		if ($this->_record and property_exists($this->_record,$column_name))
			return $this->_record->$column_name;
		else
			return NULL;
	}
	// set new properties if exist
	public function __set($column_name,$column_value)
	{
		if ((!$this->_record) and (!$this->_is_insert))
		{
			$this->current();
			if (!$this->_record) return;
		}
		
		$columndefs = $this->_columndefs;
		$joincolumns = $this->_joincolumns;
		if ($joincolumns and in_array($column_name,$joincolumns))
			return;
		if (array_key_exists($column_name,$columndefs))
		{
			$this->_pending_post or ($this->_pending_post = new StdClass);
			$this->_pending_post->$column_name = $column_value;
			$this->_is_dirty or ($this->_is_dirty = TRUE);
		}
	}
	public function is_insert()
	{
		return $this->_is_insert;
	}
	public function is_dirty()
	{
		return $this->_is_dirty;
	}
	# components
	public function tabledef() 
	{
		return $this->_table->tabledef();
	}
	public function relations()
	{
		return $this->_relations;
	}
	public function indexes()
	{
		return $this->_record_indexes;
	}
	public function result()
	{
		return $this->_result;
	}
	public function operation()
	{
		return $this->_operation;
	}
	public function pending_post()
	{
		return $this->_pending_post;
	}

	public function set($data)
	{
		if (($this->has_record()) or $this->is_insert())
		{
			foreach ($data as $field => $value)
			{
				$this->$field = $value;
			}
		}
		return $this;
	}
	// prepares for new record
	public function add() 
	{
		$this->_failed_insert = FALSE;
		$this->_error_info = NULL;
		$this->_is_dirty = FALSE;
		$this->_pending_post = NULL;
		// assert readiness
		$parent = $this->_table->parent();
		$parent and $parent->synchronize();
		if ((!$parent) or $parent->has_result())
		{
			return $this->_is_insert = TRUE;
		}
		else // no result with syncrhonize, which constitutes a failure
		{
			$this->_failed_insert = TRUE;
			return FALSE;
		}
	}
	public function has_record()
	{
		return (bool) $this->_record;
	}
	
	// [--------------------------- ACTIVERECORD GENERAL API ---------------------------]
	# builder methods
	public function insert($postingdata = FALSE)
	{
		if (!$this->add())
		{
			$this->_error_info = array('x',0,'Activerecord posting exception: Add failed in insert');
		} else {
			if ($postingdata) $this->set($postingdata);
		}
		return $this;
	}
	public function update($data)
	{
		if ($this->has_record())
		{
			$this->set($data);
		}
		return $this;
	}
	public function save($validation = NULL)
	{
		if (!is_null($validation))
			$this->_validation = $validation;
		$this->_insert_id = NULL;
		if ($this->_failed_insert) // _error_info is set, _is_dirty is not
			$this->_failed_insert = FALSE;
		else
			$this->_error_info = NULL;
		if ($this->_is_dirty) // not true with failed insert
		{
			$this->_working_record = clone $this->_pending_post; // working copy
			if ($this->save_validate()) // try to apply validate before save
			{
				if ($this->_is_insert)
				{
					$this->save_insert();
				}
				else // update
				{
					$this->save_update();
				}
			}
			$this->_working_record = NULL;
		}
		return (bool) !$this->_error_info;
	}
	protected function save_validate()
	{
		// apply kohana validation
		$validation = $this->_validation;
		if ($validation)
		{
			$record = (array) $this->_working_record;
			$validation = $validation->copy($record);
			if (!$validation->check())
			{
				$errors = $validation->errors();
				$this->_error_info = array('x',0,$errors);
			}
//			else // filters have been removed, so not required
//				$this->_working_record = (object) array_merge((array)$this->_working_record, $validation->as_array()); 
		}
		return (bool) !$this->_error_info;
	}
	protected function save_insert()
	{
		$controls = $this->get_controls();
		$table = $this->table();
		$tabledef = $this->_table->tabledef();
		$userid = $tabledef->model()->userid();
		$record = $this->_working_record;
		# impose foreignkey
		if ($this->_table->foreignkey())
		{
			foreach ($this->_table->foreignkey()->column_values() as $column_name => $column_value)
			{
				$record->$column_name = $column_value;
			}
		}
		# apply audit fields
		extract($controls);
		$CreateUserID and ($record->$CreateUserID = $userid);
		$CreateDate and ($record->$CreateDate = Services::get_sql_datetime(time())); // should come from pdo
		$EditUserID and ($record->$EditUserID = $userid);
		$EditDate and ($record->$EditDate = Services::get_sql_datetime(time()));
		
		# apply concurrency fields
		$RecordVersion and ($record->$RecordVersion = 0);
		$TimeStamp and ($record->$TimeStamp = Services::get_sql_datetime(time()));
		
		# prepare for post
		$record = (array) $record;
		$fields = array_keys($record);
		$parameters = $this->_parameters = array_values($record);
		$values = array_fill(0,count($parameters),'?');
		$tablename = $this->table()->tabledef()->insert_table(); // tablename();
		
		try
		{
			# post
			$query = $this->_query = DB::insert($tablename,$fields);
			$operation = $this->_operation = $query
				->values($values)
				->prepare($this->_table->db());
			$success = $operation->execute($parameters)->success();
				
			# no synchronize - the record disappears - user has to requery to see result
			
			# capture result
			if (!$success)
			{
				$this->_error_info = $operation->error_info();
			}
			else
			{
				$id = $this->_insert_id = $operation->insert_id();
				$keyname = $tabledef->primary_key_column_name();
				# cancel insert
				$this->_working_record->$keyname = $id;
				$this->_is_insert = FALSE;
				$this->_is_dirty = FALSE;
				$this->_pending_post = new StdClass;
			}
		} 
		catch (Exception $e) 
		{
			$this->_error_info = $this->_table->db()->statement_error();
		}
		return (bool) !$this->_error_info;
	}
	protected function save_update()
	{
		$controls = $this->get_controls();
		$table = $this->table();
		$tabledef = $this->_table->tabledef();
		$userid = $tabledef->model()->userid();
		$primary_key_column_name = $tabledef->primary_key_column_name();
		$record = $this->_working_record;
		$is_operation = FALSE; // bit of a hack
		try
		{
			# lock table
			if ($table->auto_lock()) {
				$this->table()->lock_table();
			}
				
			# get previous record
			$copyrecord = NULL;
			$result = DB::select()
				->from($this->table()->tabledef()->tablename())
				->where($primary_key_column_name,'=',$this->$primary_key_column_name)
				->as_object()
				->execute();
			$result->success() and ($copyrecord = $result->current());
				
			# check for record deleted condition
			if (!$copyrecord)
			{
				$message = 'Another User has deleted the record - cancelling update (' 
					. $this->table()->tabledef()->tablename() 
					. ' ' . $this->$primary_key_column_name . ')';
				$this->_error_info = array('deleted','0',$message);
			}
			else
			{
				# check and increment concurrency controls
				extract($controls);
				if ($RecordVersion)
				{
					if (!property_exists($record,$RecordVersion))
						$record->$RecordVersion = $copyrecord->$RecordVersion;
					if ($copyrecord->$RecordVersion !== $record->$RecordVersion)
					{
						$message = 	'Another User has changed the record - (version) cancelling update (' 
							. $this->table()->tabledef()->tablename() 
							. ' ' . @$copyrecord->$primary_key_column_name . ')';
						$this->_error_info = array('conflict','0',$message);
					}
					else
					{
						$record->$RecordVersion = ++$record->$RecordVersion;
					}
				}
				if ($TimeStamp)
				{
					if (!property_exists($record,$TimeStamp))
						$record->$TimeStamp = $copyrecord->$TimeStamp;
					if ($copyrecord->$TimeStamp !== $record->$TimeStamp)
					{
						$message = 'Another User has changed the record (timestamp) - cancelling update (' 
							. $this->table()->tabledef()->tablename() 
							. ' ' . @$copyrecord->$primary_key_column_name . ')';
						$this->_errorinfo = array('x','0',$message);
					}
					else
					{
						$record->$TimeStamp = Services::get_sql_datetime(time());
					}
				}
				if (empty($this->_error_info)) {
					# apply audit data
					$EditUserID and ($record->$EditUserID = $userid);
					$EditDate and ($record->$EditDate = Services::get_sql_datetime(time()));
				
					# protect previous create audit stamp; only existing properties are posted
					if ($CreateDate) unset($record->$CreateDate);
					if ($CreateUserID) unset($record->$CreateUserID);
				
					# prepare for post
					$record = (array) $record;
//					var_dump($record);
					$fields = array_keys($record);
					$parameters = array_values($record);
					$values = array_fill(0,count($parameters),'?');
					$fields = array_combine($fields,$values);
				
					# post record
					$query = $this->_query = DB::update($this->_table->tabledef()->tablename());
					$is_operation = TRUE;
					$operation = $this->_operation = $query
						->set($fields)
						->where($primary_key_column_name, '=',$this->$primary_key_column_name)
						->prepare($this->_table->db());
//					var_dump($operation->sql());
					$operation->execute($parameters);
					$success = $operation->success();
					$is_operation = FALSE;
					# collect error information
					if (!$success)
						$this->_error_info = $operation->error_info();
					else
					{
						# get refreshed version of the record
						$query = clone $this->_table->query();
						$query->where($primary_key_column_name, '=',$this->$primary_key_column_name);
						$result = $query->execute();
						if (!is_object($result)) {
							$db = $this->_table->db();
							echo 'failed to refresh record';
							var_dump($query->sql());
							var_dump($db->statement_error());
							exit;
						}
						$this->_record = $result->current();
						unset($query);
					}
				}
			}
				
		} catch (Exception $e) {
			if ($is_operation) // bit of a hack
				$this->_error_info = $this->_table->db()->operation_error();
			else
				$this->_error_info = array('tmx-ar',$e->getCode(),$e->GetMessage());
		} // finally...
		if ($table->auto_lock())
			$this->table()->unlock_table();
		if (!$this->_error_info)
		{
			$this->_is_dirty = FALSE;
			$this->_pending_post = new StdClass;
			$this->notify_observers();
		}
		return (bool) !$this->_error_info;
	}
	protected function get_controls()
	{
		if (empty($this->_controls))
		{
			$tabledef = $this->_table->tabledef();
			$audit = $tabledef->audit();
			$concurrency = $tabledef->concurrency();
			$config = $tabledef->model()->config();
			// assert data:
			($audit === FALSE) or ($audit) or ($audit = $config->table['audit']);
			($concurrency === FALSE) or ($concurrency) or ($concurrency = $config->table['concurrency']);
			// localize vars:
			if ($audit) extract($audit);
			if ($concurrency) extract($concurrency);
			// assert vars:
			isset($CreateUserID) or ($CreateUserID = NULL);
			isset($CreateDate) or ($CreateDate = NULL);
			isset($EditUserID) or ($EditUserID = NULL);
			isset($EditDate) or ($EditDate = NULL);
			isset($RecordVersion) or ($RecordVersion = NULL);
			isset($TimeStamp) or ($TimeStamp = NULL);
			// save vars:
			$this->_controls = compact(
					'CreateUserID','CreateDate',
					'EditUserID','EditDate',
					'RecordVersion','TimeStamp'
				);
		}
		return $this->_controls;
	}
	// substitute of delete table assumes match of primary key fields
	public function delete() 
	{
		$this->_error_info = NULL;
		$this->_insert_id = NULL;
		if ($this->has_record())
		{
			try
			{
				$this->_insert_id = NULL;
				$tablename = $this->table()->tabledef()->delete_table();
				$query = $this->_query = DB::delete($tablename);
				// *** use primary_key_name and value
				$columns = $this->_record_indexes[
						$this->table()
							->tabledef()->primary_index_name()]
							->column_values();
				foreach ($columns as $column => $value);
					$query->where($column,"=",$value);
				$operation = $this->_operation = $query->prepare($this->_table->db())->execute();
				$retval = $operation->success();
				$this->_insert_id = $operation->insert_id();
				if (!$retval)
					$this->_error_info = $operation->error_info();
				else
				{
					$this->_working_record = $this->_record;
					$this->_record = NULL;
					$this->_is_dirty = FALSE;
					$this->_is_insert = FALSE;
					$this->notify_observers();
					
					$this->fetch();
				}
				
			} catch (Exception $e) {
				$this->_error_info = $this->_table->db()->error_info(); // *** wrong error batch ***
			}
			return (bool) !$this->_error_info;
		}
		else
		{
			return TRUE;
		}
	}
	public function working_record()
	{
		return $this->_working_record;
	}
	
	# data acess
	public function fetch() // next, including first
	{
		// moving records implies abandoning changes in progress...
		if ($this->_is_dirty)
		{
			$this->_pending_post = new StdClass;
			$this->_is_dirty = FALSE;
		}
		$this->_is_insert = FALSE;
		
		$table = $this->_table;
		$parent = $table->parent();
		// assertions:
		$parent and $parent->synchronize();
		$table->has_result() or $table->get_result();
		// query result
		if ($result = $table->result())
		{
			$this->_record = $result->fetch();
			$this->notify_observers();
		}
		return $this->_record;
	}
	public function current() // not including pending changes
	{
		$table = $this->_table;
		$parent = $table->parent();
		// assertions:
		$parent and $parent->synchronize();
		$table->has_result() or $table->get_result();
		// query result
		if (!$this->_record) 
		{
			if ($result = $this->_result)
			{
				$this->_record = $result->current();
				$this->notify_observers();
			}
		}
		return $this->_record;
	}
	public function has_current()
	{
		return (boolean) $this->_record;
	}
	public function load()
	{
		$this->current();
		return $this;
	}
	# accessors
	public function validation($validation = NULL)
	{
		if (!is_null($validation))
			$this->_validation = $validation;
		return $this->_validation;
	}
	public function table() 
	{
		return $this->_table;
	}
	public function BOF()
	{
		$result = $this->_result;
		if (!$result) return TRUE;
		return $result->BOF();
	}
	public function EOF()
	{
		$result = $this->_result;
		if (!$result) return TRUE;
		return $result->EOF();
	}
	# posting side effects
	public function error_info()
	{
		return $this->_error_info;
	}
	public function insert_id()
	{
		return $this->_insert_id;
	}
	public function query() // for debugging
	{
		return $this->_query;
	}
	public function parameters() // for debugging
	{
		return $this->_parameters; 
	}
}
