<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster\Tablemodel;

use StdClass;
use Database, Database_Query_Builder_Select, Kohana_Exception;

defined('SYSPATH') or die('No direct script access.');
/*
protect against empty defaults

respond when prepare or execute silently come back FALSE for failure
*/
/*
 file: query.php
 package: Tablemodel
 category: ORM (Object Relational Manager)
 date: August 2010
 description:
 class: Querydef
 Querydef objects can be cloned
 this is the query definition for table
 Querydef delegates most operations to Database_Query_Builder_Select

GENERAL API
	# query builder methods:
	the following methods are passed on to Kohana Database_Query_Builder_Select
	->where($column, $op, $value)
	->and_where($column, $op, $value)
	->or_where($column, $op, $value)
	->where_open()
	->and_where_open()
	->or_where_open()
	->where_close()
	->and_where_close()
	->or_where_close()
	->having($column, $op, $value = NULL)
	->and_having($column, $op, $value = NULL)
	->or_having($column, $op, $value = NULL
	->having_open()
	->and_having_open()
	->or_having_open()
	->having_close()
	->and_having_close()
	->or_having_close()
	->select($columns = NULL)
	->select_array(array $columns)
	->distinct($value)
	->group_by($columns)
	->order_by($column, $direction = NULL)
	->offset($number)
	->limit($number)
	(other kohana query builder methods, like join() are filtered out)
	# additional builder methods:
	->requery()
	->config($config = NULL)
	# operation
	->prepare($bindings = NULL,$binding_datatypes = array())
	->execute($parameters = NULL)
	->reset() // clear all but constraint settings
	# accessor methods
	->result()
	->sql($is_prepared = FALSE)
	->table() returns table (NULL for clone query
*/
class Query
{
	public $constraints;

	protected $_table; // associated table; null if query is cloned
	protected $_db; // the database
	protected $_query; // Kohana query object (Muster extension)
	protected $_result; // the result of the select query
	
	protected $_sql; // the sql string
	protected $_query_methods; // legitimate query methods to pass to Kohana query object
	protected $_conditions = array(); // collect conditions for later bulk passing to Kohana query (assert_conditions)
	protected $_called_conditions = array(); // conditional list of conditions being collected for passing to Kohana
	protected $_parameters; // PDO statement parameters
	protected $_bindings; // PDO statement bindings
	protected $_bindings_datatypes; // for pdo statements
	protected $_config; // config extended kohana query
	protected $_defaults; // defaults (perpage and orderby) from table
	protected $_has_conditions = FALSE;

	
	public function __construct($specification)
	{
		$where_methods = array(
			'where','and_where','or_where',
			'where_open','and_where_open','or_where_open',
			'where_close','and_where_close','or_where_close');
		$having_methods =  array(
			'having','and_having','or_having',
			'having_open','and_having_open','or_having_open',
			'having_close','and_having_close','or_having_close');
		$other_methods = array(
			'select','select_array',
			'distinct','group_by','order_by',
			'offset','limit');
		$this->_query_methods = array_merge($where_methods,$having_methods,$other_methods);
		$this->constraints = new StdClass;
		$this->_query = new Database_Query_Builder_Select();
		if (!isset($specification->db))
			$this->_db = Database::instance();
		else
			$this->_db = $specification->db;
		$table = $this->_table = $specification->table;
		$this->_defaults = $table->defaults();
	}
	// save arguments for assert_conditions().
	public function __call($method,$args)
	{
		if (in_array($method,$this->_query_methods))
		{
			$this->_conditions[] = array($method,$args);
			if (($method == 'select') or ($method == 'select_array'))
			{
				if (!empty($args))
					$this->_called_conditions[$method] = TRUE;
			}
			else
				$this->_called_conditions[$method] = TRUE; // to test for default requirements
			$this->_has_conditions = TRUE;
			return $this;
		}
		else
		{
			throw new Kohana_Exception('Invalid call to tablemodel_query: ' . $method);
		}
	}
	
	public function __clone()
	{
		$this->_table = NULL;
		$this->_query = clone $this->_query;
		$this->_result = NULL;
//		if ($this->_result)
//			$this->_result = clone $this->_result;
	}
	
	public function destroy()
	{
		$this->_db = NULL;
		$this->_table = NULL;
		$this->_query = NULL;
		$this->_result = NULL;
	}
	
	public function sql($is_prepared = FALSE)
	{
		if ($result = $this->_result)
		{
			$sql = $result->sql();
		}
		else 
		{
			$this->_query->reset();
			$this->assert_conditions();
			$this->assert_defaults();
			$this->assert_constraints($is_prepared);
			$sql = $this->_query->compile($this->_db);
			$this->_query->reset();
		}
		return $sql;
	}
	public function reset()
	{
		$this->_query->reset();
		$this->_result = NULL;
		$this->_config = NULL;
		$this->_bindings = NULL;
		$this->_bindings_datatypes = NULL;
		$this->_parameters = NULL;
		$this->_called_conditions = array();
		$this->_conditions = array();
		$this->_has_conditions = FALSE;
		
		return $this;
	}
	public function clear_select()
	{
		$conditions = $this->_conditions;
		foreach ($conditions as $index => $condition)
		{
			if (($condition[0] == 'select') or ($condition[0] == 'select_array'))
				unset($conditions[$index]);
		}
		$this->_conditions = array_values($conditions);
		$this->_query->clear_select();
		return $this;
	}
	public function clear_result()
	{
		$this->_result = NULL;
		return $this;
	}

	public function table()
	{
		return $this->_table;
	}
	public function result()
	{
		return $this->_result;
	}
	public function has_conditions()
	{
		return $this->_has_conditions;
	}
	public function requery()
	{
		if ($this->_table)
			$this->_table->requery();
		elseif ($this->_result)
			$this->_result->requery();
		return $this;
	}
	public function query()
	{
		return $this->_query;
	}
	public function config($config = NULL)
	{
		$this->_config = $config;
		return $this;
	}
	protected function assert_conditions()
	{
		$conditions = $this->_conditions;
		$query = $this->_query;
		foreach ($conditions as $condition)
		{
			list($method,$args) = $condition;
			call_user_func_array(array($query,$method),$args);
		}
	}
	protected function assert_defaults()
	{
		$defaults = $this->_defaults;
		$query = $this->_query;
		if (empty($this->_called_conditions['limit']))
			$query->limit($defaults->perpage);
		if (empty($this->_called_conditions['order_by']) and $defaults->orderby)
		{
			$order_by = $defaults->orderby;
			foreach ($order_by as $column => $direction)
			{
				$query->order_by($column,$direction);
			}
		}
	}
	protected function assert_constraints($is_prepared)
	{
		$query = $this->_query;
		$query->from($this->constraints->tablename);
		if (isset($this->constraints->foreignkey))
		{
			$foreignkeys = $this->constraints->foreignkey;
			foreach ($foreignkeys as $key => $value) // only one
			{
				if ($is_prepared)
					$query->where($key,'=',$this->_db->escape($value));
				else
					$query->where($key,'=',$value); // escaped by $this->_query
			}
		}
		if ((isset($this->_called_conditions['select']) or isset($this->_called_conditions['select'])) 
			and isset($this->constraints->indexcolumns))
		{
			$query->select_array($this->constraints->indexcolumns);
		}
	}
	public function prepare($bindings = NULL,$binding_datatypes = array())
	{
		if (is_null($bindings)) $bindings = $this->_bindings;
		if (is_null($binding_datatypes)) $binding_datatypes = $this->_binding_datatypes;
		$this->_bindings = $bindings;
		$this->_binding_datatypes = $binding_datatypes;
		$this->_query->reset();
		$this->assert_conditions();
		$this->assert_defaults();
		$this->assert_constraints(TRUE);
		return $this->_result = $this->_query->config($this->_config)->prepare($this->_db,$bindings,$binding_datatypes);
	}
	public function execute($parameters = NULL)
	{
		if (is_null($this->_result)) // first run, not prepared
		{
			$this->_query->reset();
			$this->assert_conditions();
			$this->assert_defaults();
			$this->assert_constraints(FALSE);
			if ($this->_table)
				$this->_table->reset();
			$result = $this->_result = 
				$this->_query->as_object()
					->statement_config($this->_config)
					->execute($this->_db);
			return $result;
		}
		// with existing result...
		if (is_null($parameters))
		{
			$parameters = $this->_parameters;
		}
		else
		{
			$this->_parameters = $parameters;
		}
		if (is_null($parameters) and is_null($this->_bindings)) // no need for prepare, re-compile and rerun
		{
			$this->_query->reset();
			$this->assert_conditions();
			$this->assert_defaults();
			$this->assert_constraints(FALSE);
			if ($this->_table)
				$this->_table->reset();
			return $this->_result = $this->_query->as_object()->statement_config($this->_config)->execute($this->_db); // not prepared
		}
		else // parameters or bindings indicate prepared statement
		{
			return $this->_result->execute($parameters); // previously prepared
		}
	}
}
