<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster\Tablemodel;

use Kohana, Kohana_Exception;
use Tablemodel;

defined('SYSPATH') or die('No direct script access.');
/**
 file: relationdef.php
 package: Tablemodel
 category: ORM (Object Relational Manager)
 date: August 2010
 class: RelationDef
 relation metadata
 
* class: RecordRelationdef
*/
class Relationdef
{
	protected $_tabledef; // table object reference
	protected $_name;
	protected $_relation_type; // TableModel:: ONE_TO_ONE | TableModel::ONE_TO_MANY | TableModel::MANY_TO_ONE
	protected $_indexdef; // object reference - index definition
	protected $_remote_indexdef; // object reference - index definition
	protected $_reciprocal_relationdef; // object reference - relation definition
	protected $_terminate_expansion; // if parent is 1:1 or M:1 then if true do not expand tree further
	
	public function __construct($parameters)
	{
		// set properties
		$this->_tabledef = $parameters->tabledef;
		$this->_name = $parameters->name;
		$this->_relation_type = $parameters->relation_type;
		$terminate_expansion = $this->_terminate_expansion = $parameters->terminate_expansion;
		if (is_null($terminate_expansion))
			$this->_terminate_expansion = $this->_tabledef->model()->config()->relation['terminate_expansion'];
		
		// check indexdefs
		$index_name = $parameters->index;
		$remote_table_name = $parameters->remote_table;
		$remote_index_name = $parameters->remote_index;
		
		$tabledefs = $this->_tabledef->model()->tabledefs();
		$indexdefs = $this->_tabledef->indexdefs();
		$remote_indexdefs = $tabledefs[$remote_table_name]->indexdefs();
		$this->_validate_indexdef(
			$indexdefs,$index_name,
			$tabledefs,$remote_table_name,
			$remote_indexdefs,$remote_index_name
		);
		// set relation resources
		$this->_indexdef = $indexdefs[$index_name];
		$this->_remote_indexdef = $remote_indexdefs[$remote_index_name];
		$this->_validate_index_typematches();
	}

	protected function _validate_indexdef(
		$indexdefs,$index_name,
		$tabledefs,$remote_table_name,
		$remote_indexdefs,$remote_index_name)
	{
		$debug_level = $this->_tabledef->model()->debug_level();
		if ($debug_level > 0) 
		{
			// check for existence of named index
			if (!isset($indexdefs[$index_name]))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.missing_indexdef'),
							array(':index' => $index_name,
							':relation' => $this->_name,
							':table' => $this->_tabledef->tablealias())
					);
			}
			// check for existence of named remote table
			if (!isset($tabledefs[$remote_table_name]))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.missing_tabledef'),
							array(':remotetable' => $remote_table_name,
							':relation' => $this->_name,
							':table' => $this->_tabledef->tablealias())
					);
			}
			// check for existence of remote named index
			if (!isset($remote_indexdefs[$remote_index_name]))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.missing_remote_indexdef'),
							array(':remotetable' => $remote_table_name,
							':remoteindex' => $remote_index_name,
							':relation' => $this->_name,
							':table' => $this->_tabledef->tablealias())
					);
			}
		}
	}
	
	protected function _validate_index_typematches()
	{
		$debug_level = $this->_tabledef->model()->debug_level();
		if ($debug_level > 0) 
		{
			// check for correct matching of relation type and relation index type
			$relation_type = $this->_relation_type;
			$index_type = $this->_indexdef->index_type();
			if (! (
				(($relation_type == TableModel::ONE_TO_MANY) and ($index_type == TableModel::UNIQUE))
				or
				(($relation_type == TableModel::ONE_TO_ONE) and ($index_type == TableModel::UNIQUE))
				or
				(($relation_type == TableModel::MANY_TO_ONE) and ($index_type == TableModel::INDEX))
				))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.mismatched_indextype'),
						array(':relation' => $this->_name,
						':table' => $this->_tabledef->tablealias())
					);
			}
		}
	}
	
	// each relation definition must have exactly one reciprocal relation
	public function find_reciprocal_relationdef() 
	{
		//already set by the reciprocal relation
		if ($this->_reciprocal_relationdef) 
		{
			return;
		}
		// find reciprocal relation
		// values to find...
		$index_name = $this->_indexdef->name();
		$table_name = $this->_tabledef->tablealias();
		$remote_table = $this->_remote_indexdef->tabledef();
		$reciprocal_relationdefs = array();
		foreach ($remote_table->relationdefs() as $remote_relationdef)
		{
			if (($remote_relationdef->remote_indexdef()->name() == $index_name) and 
				($remote_relationdef->remote_indexdef()->tabledef()->tablealias() == $table_name))
			{
				$reciprocal_relationdefs[] = $remote_relationdef;
			}
		}
		// check validity of reciprocal relation
		$this->_validate_reciprocal_relation($reciprocal_relationdefs,$table_name);
		// set reciprocal relation, both ways to save time
		$this->_reciprocal_relationdef = $reciprocal_relationdefs[0]; // validation proved that there is only one.
		$this->_reciprocal_relationdef->set_reciprocal_relationdef($this);
		// check for matching relation types
		$this->_validate_reciprocal_matches($table_name);
	}
	
	protected function _validate_reciprocal_relation($reciprocal_relationdefs,$table_name)
	{
		$debug_level = $this->_tabledef->model()->debug_level();
		if ($debug_level > 0) 
		{
			if (count($reciprocal_relationdefs) == 0)
			{
				// check that reciprocal relation exists
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.mismatched_relations'),
							array(':relation' => $this->_name,
							':table' => $table_name)
					);
			}
			// check that only one reciprocal relation exists
			if (count($reciprocal_relationdefs) > 1)
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.excess_relations'),
							array(':relation' => $this->_name,
							':table' => $table_name)
					);
			}
		}
	}
	
	protected function _validate_reciprocal_matches($table_name)
	{
		$debug_level = $this->_tabledef->model()->debug_level();
		if ($debug_level > 0) 
		{
			// check that local and remote relation types match
			$relation_type = $this->_relation_type;
			$reciprocal_relation_type = $this->_reciprocal_relationdef->relation_type();
			if (! (
				(($relation_type == TableModel::ONE_TO_MANY) and 
					($reciprocal_relation_type == TableModel::MANY_TO_ONE))
				or 
				(($reciprocal_relation_type == TableModel::ONE_TO_MANY) and 
					($relation_type == TableModel::MANY_TO_ONE))
				or
				(($relation_type == TableModel::ONE_TO_ONE) and 
					($reciprocal_relation_type == TableModel::ONE_TO_ONE))
				))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.mismatched_relationtypes'),
							array(':reciprocalrelation' => $this->_reciprocal_relationdef->name(),
							':relation' => $this->_name,
							':table' => $table_name)
					);
			}
			$indexdef = $this->_indexdef;
			$reciprocal_indexdef = $this->_reciprocal_relationdef->indexdef();
			// check that local and reciprocal relations have the same number of columns
			$column_names = $indexdef->column_names();
			$reciprocal_column_names = $reciprocal_indexdef->column_names();
			if (count($column_names) != count($reciprocal_column_names))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','relationdef.mismatched_columncount'),
							array(':reciprocalrelation' => $this->_reciprocal_relationdef->name(),
							':relation' => $this->_name,
							':table' => $table_name)
					);
			}
		}
	}
	
	//property accessors...
	public function name()
	{
		return $this->_name;
	}
	public function relation_type()
	{
		return $this->_relation_type;
	}
	public function indexdef()
	{
		return $this->_indexdef;
	}
	public function remote_indexdef()
	{
		return $this->_remote_indexdef;
	}
	public function reciprocal_relationdef()
	{
		return $this->_reciprocal_relationdef;
	}
	public function set_reciprocal_relationdef($value)
	{
		return $this->_reciprocal_relationdef = $value;
	}
	public function model()
	{
		return $this->_tabledef->model();
	}
	public function terminate_expansion($value = NULL)
	{
		if (!empty($value))
			$this->_terminate_expansion = $value;
		return $this->_terminate_expansion;
	}
}
