<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster\Tablemodel;

use StdClass;
use Database,DB;
use Tablemodel;

defined('SYSPATH') or die('No direct script access.');
/*
 file: table.php
 package: Tablemodel
 category: ORM (Object Relational Manager)
 date: August 2010
 description:
 class table
 observes its Foreignkey
 observed by its Activerecord
 instantiated by tabledef
 is a view defined by the query and accessed through the result object 
	from underlying database driver (which contains lazy list of records)
 linked to activerecord that wraps current table result record
GENERAL API:
	->__construct($specification)
	# operations (builder methods)
	->requery() // close and re-execute the statement (like rewind, but with database refresh)
	->execute($parameters = NULL); // parameters for pdo statement
	->reset() // anticipate call using new query
	->load() // from established query def
	->preload_parent_record() // establish parent current record 
		(this avoids notify observer which clears query, when current activerecrd->current() is called)
	# accessors
	->name()
	->tablename()
	->db()
	->query()
	->activerecord($tablename = NULL)
	->recordset($tablename = NULL)
	->record_count()
	->defaults($defaultname = NULL,$value = NULL); //'perpage' or 'orderby'
	# utilities...
	->unlimited_count($query = NULL) // how many records are available from defined query
	->calculate_record_offset(array $searchvalues, array $offsetcondition,$searchlimit = 200)
	# advanced
	->set_autolock($lock = TRUE)
	->lock_table() // locks the current table (for write) - used automatically by activerecord update
	->unlock_table() // unlocks table - used automatically by activerecord update
	->lock_branch()
	->unlock_tables() // same as singular
	->parent()
	->child($tablename)
	->children()
*/
class Table implements Observer, Observed // observes its Foreignkey, notifies its Activerecord
{
	// to provide defaults for pagination...
	protected $_defaults;
	// properties
	protected $_auto_lock = TRUE; // lock tables when posting updates and checking recordversion number
	// components
	protected $_tabledef; // table object reference
	protected $_db;
	protected $_observers = array(); // Activerecord
	protected $_children; // child tables
	protected $_parent; // parent table
	protected $_parent_relation_type; // for this instance
	protected $_result; //holds object reference to Database_Result object (Select query)
	protected $_activerecord; // object reference to the current record wrapper
	protected $_recordset; // object containing working array of record objects placed in this wrapper from result
	protected $_foreignkey; // controlling constraint index object
	protected $_query; // object reference to sql source object

	# lifecycle
	public function __construct($specification)
	{
		$tabledef = $this->_tabledef = $specification->tabledef;
		if ($specification->parent)
		{
			$this->_parent = $specification->parent;
			$this->_parent_relation_type = $specification->relationdef->relation_type();
		}
		$defaults = $this->_defaults = new StdClass;
		$perpage = $defaults->perpage = $tabledef->perpage();
		$orderby = $defaults->orderby = $tabledef->orderby();
		# assertions 
		$perpage or
			($defaults->perpage = $tabledef->model()->config()->table['perpage']);
		if (!($orderby or ($orderby === FALSE)))
		{
			$columns = $tabledef->alternate_index_column_names();
			if (!empty($columns)) {
				$orderby = array();
				foreach ($columns as $column)
					$orderby[$column] = NULL;
				$defaults->orderby = $orderby;
			}
		}
		$this->_db = $tabledef->model()->db();
		$this->create_components();
	}
	// internal
	protected function create_components()
	{
		$parameters = new StdClass;
		$parameters->table = $this;
		$parameters->db = $this->_db;
		// create query
		$query = $this->_query = new Query($parameters);
		$query->constraints->tablename = $this->tabledef()->tablename();
		$query->constraints->tableindex = $this->tabledef()->name();
		// create Activerecord, which creates Indexes, Relations and remote tables
		$this->_activerecord = new Activerecord($parameters);
		# collect indexcolumns for query - auto inclusion in select statements
		$indexcolumns = array();
		($indexes = $this->_activerecord->indexes()) or ($indexes = array());
		foreach ($indexes as $index)
			$indexcolumns = array_merge($indexcolumns,$index->indexdef()->column_names());
		empty($indexcolumns) or ($query->constraints->indexcolumns = array_unique($indexcolumns));
		// recordset is created on demand, not here
	}
	
	// internal
	public function destroy()
	{
		// remove references for garbage collection
		$this->_query->destroy();
		$this->_activerecord->destroy();
		$this->_recordset and $this->_recordset->destroy();
		
		$this->_activerecord = NULL;
		$this->_recordset = NULL;
		$this->_tabledef = NULL;
		$this->_db = NULL;
		$this->_observers = NULL;
		$this->_children = NULL;
		$this->_parent = NULL;
		$this->_result = NULL;
		$this->_foreignkey = NULL;
		$this->_query = NULL;
	}
	# observers
	// observer methods, internal
	public function update_observer(Observed $foreignkey) // Observer method
	{
		$this->_query->constraints->foreignkey = $foreignkey->column_values();
//		$this->_query->clear_result();
		$this->_query->reset();
		$this->reset();
	}
	public function attach_observer(Observer $activerecord) // Observed method
	{
		if (! in_array($activerecord, $this->_observers))
		{
			$this->_observers[] = $activerecord;
		}
	}
	public function detach_observer(Observer $activerecord) // Observed method
	{
		$pos = array_search($activerecord,$this->_observers);
		if (!($pos === FALSE))
		{
			array_splice($this->_observers,$pos,1);
		}
	}
	public function notify_observers() // Observed method; notify Activerecords
	{
		foreach ($this->_observers as $observer)
		{
			$observer->update_observer($this);
		}
	}
	public function observer_name() // TableModelObserver method
	{
		if ($this->_foreignkey)
		{
			$relationtype = $this->_foreignkey->recordrelation()->relationdef()->relation_type();
			switch ($relationtype)
			{
				case TableModel::ONE_TO_MANY: $type = '1:M'; break;
				case TableModel::MANY_TO_ONE: $type = 'M:1'; break;
				case TableModel::ONE_TO_ONE: $type = '1:1'; break;
			}
			return '{' . $type . '} ' . $this->_tabledef->tablealias();
		}
		else
		{
			return $this->_tabledef->tablealias();
		}
	}
	# operations, internal
	public function set_children() 
	{
		$relations = $this->_activerecord->relations();
		$children = array();
		foreach ($relations as $relation)
		{
			$child = $relation->foreignkey()->table();
			$child->set_children();
			$children[$child->name()] = $child;
		}
		$this->_children = $children;
	}
	
	public function synchronize()
	{
		if ($this->_activerecord->has_record())
			return TRUE;
		elseif ($this->has_result())
			return (bool) $this->_activerecord->current();
		else // no current result
		{
			$parent = $this->_parent;
			$parent and $parent->synchronize();
			$this->get_result();
			if ($this->has_result())
				return (bool) $this->_activerecord->current();
			else
				return FALSE;
		}
	}

	public function set_foreignkey($value)
	{
		return $this->_foreignkey = $value;
	}
	// triggered from synchronize, internal
	public function get_result($parameters = NULL)
	{
		$parent = $this->_parent;
		if ((!$parent) or $parent->activerecord()->has_record())
		{
			$query = $this->_query;
			$result = $this->_result = $query->result();
			if (!$result) {
				$this->_result = $query->execute($parameters);
			}
		}
		else
		{
			$this->_result = NULL;
		}
		$this->notify_observers();
		return $this;
	}
	# accessors, internal
	public function tabledef() 
	{
		return $this->_tabledef;
	}
	public function is_ancestor($tabledef_name)
	{
		$is_ancestor = FALSE;
		$parent = $this;
		do 
		{
			$ancestor = $parent->parent();
			if ($ancestor)
			{
				if ($ancestor->name() == $tabledef_name)
				{
					$is_ancestor = TRUE;
					break;
				}
			}
		} while ($parent = $parent->parent());
		return $is_ancestor;
	}
	public function parent_relation_type()
	{
		return $this->_parent_relation_type;
	}
	public function foreignkey()
	{
		return $this->_foreignkey;
	}
	public function has_result()
	{
		return (bool) $this->_result;
	}
	// get or set
	public function auto_lock($value = NULL)
	{
		if (!is_null($value))
			$this->_auto_lock = $value;
		return (bool) $this->_auto_lock;
	}
	public function as_array($keyfield = NULL,$valuefield = NULL) // called from recordset
	{
		$parent = $this->_parent;
		// assertions:
		$parent and $parent->synchronize();
		$this->has_result() or $this->get_result();
		if (!$this->_result) return array();
		
		return $this->_result->as_array($keyfield,$valuefield);
	}
	public function has_records()
	{
		return (bool) ($this->record_count() > 0);
	}
	
	// [--------------------------- table GENERAL API ---------------------------]
	# operations
	// prepare for subquery and subsequent $this->activerecord->current().
	// if parent has not yet been record-initialized, the current() would
	// trigger a record fetch in parent, which would trigger notification
	// which would clear (reset) the subquery!!
	// This is a subtle unwanted side effect of lazy loading combined with
	// observer behaviour.
	public function preset()
	{
		$this->preload_parent_record();
		$this->reset()->query()->reset();
		return $this;
	}
	public function preload_parent_record()
	{
		$parent = $this->_parent;
		$parent and $parent->activerecord()->load();
		return $this;
	}
	public function result()
	{
		return $this->_result;
	}
	public function requery()
	{
		if ($this->_result) 
		{
			$this->_result->requery();
			$this->notify_observers();
		}
		return $this;
	}
	public function execute($parameters = NULL) // pdo statement parameters
	{
		if (!$this->_result)
		{
			$this->get_result($parameters);
		}
		else
		{
			$this->_result->execute($parameters);
			$this->notify_observers();
		}
		return $this;
	}
	public function reset()
	{
		$this->_result = NULL;
		$this->_query->clear_result();
		$this->notify_observers();
		return $this;
	}
	public function load()
	{
		$this->_result or $this->get_result();
//		var_dump($this->_result);
		return $this;
	}
	# accessors
	public function name() // schema table name
	{
		return $this->_tabledef->tablealias();
	}
	public function tablename() // database table name
	{
		return $this->_tabledef->tablename();
	}
	public function db() // $db = NULL)
	{
//		if (!is_null($db))
//			$this->_db = $db;
		return $this->_db;
	}
	public function  set_query($query) // query restored by recordset
	{
		$this->_query = $query;
	}
	public function  query($tablename = NULL) // query restored by recordset
	{
		if (!($tablename === NULL))
		{
			return isset($this->_children[$tablename])
				? $this->_children[$tablename]->query()
				: NULL;
		}
		else
		{
			return $this->_query;
		}
	}
	public function activerecord($tablename = NULL)
	{
		if (!($tablename === NULL))
		{
			return isset($this->_children[$tablename])
				? $this->_children[$tablename]->activerecord()
				: NULL;
		}
		else
		{
			return $this->_activerecord;
		}
	}

	public function recordset($tablename = NULL)
	{
		if (!($tablename === NULL))
		{
			return isset($this->_children[$tablename])
				? $this->_children[$tablename]->recordset()
				: NULL;
		}
		else
		{
			if (!$this->_recordset)
			{
				$parameters = new StdClass;
				$parameters->table = $this;
				$parameters->db = $this->_db;
				// create recordset
				$this->_recordset = New Recordset($parameters);
			}
			return $this->_recordset;
		}
	}
	public function record_count()
	{
		if ($result = $this->_result)
			return $result->row_count();
		else
			return NULL; // 0;
	}
	public function defaults($defaultindex = NULL, $value = NULL)
	{
		if (is_null($defaultindex)) return $this->_defaults;
		else 
		{
			if (!is_null($value))
			{
				if (property_exists($this->_defaults,$defaultindex))
					$this->_defaults->$defaultindex = $value;
			}
			if (property_exists($this->_defaults,$defaultindex))
				return $this->_defaults->$defaultindex;
			else
				return NULL;
		}
	}

	# utilities
	// ** clear_select does not clear out constrained select which in this context leaves unnecessary index select components
	// ** consider clear_all_select();
	public function unlimited_count($query = NULL)
	{
		if (is_null($query))
			$query = clone $this->query();
		$field = DB::expr('count(*) as count');
		$query
			->clear_select()
			->select($field)
			->limit(1)
			->offset(0);
//		var_dump($query->sql()); exit;
		$retval = $query->execute();
		if (!$retval or !$retval->success()) {
			echo 'SQL ERROR <br />';
			var_dump($query->sql());
			exit;
		}
		$result = $query->result();
		$record = $result->current();
		return $record->count;
	}
	public function calculate_record_offset(array $searchvalues, array $offsetcondition,$searchlimit = 200)
	{
		$queryrecordcount = $this->unlimited_count();
		$query = clone $this->query();
		list($column,$op,$value) = $offsetcondition;
		$result = $query
			->where($column,$op,$value)
			->limit($searchlimit)
			->execute();
		$partitionrecordcount = $this->unlimited_count($query);
		$searchcount = 0;
		$found = FALSE;
		while ($record = $result->current())
		{
			$searchcount++;
			foreach ($searchvalues as $key => $value)
			{
				$found = ($record->$key === $value);
				if (!$found) break;
			}
			if ($found) break;
			if ($searchcount > $searchlimit) break;
			$result->next();
		}
		if (!$found) 
			$searchcount = 0;
		else
			$searchcount--;
		return (integer) ($queryrecordcount - $partitionrecordcount + $searchcount);
	}

	# advanced
	public function set_autolock($lock = TRUE)
	{
		$this->_auto_lock = $lock;
	}
	public function lock_table()
	{
		return $this->_db->lock_tables(array($this->tabledef()->tablename()));
	}
	public function unlock_table()
	{
		return $this->_db->unlock_tables();
	}
	public function unlock_tables()
	{
		return $this->_db->unlock_tables();
	}
	public function lock_branch()
	{
		$tablelist = $this->get_descendants();
		return $this->_db->lock_tables($tablelist);
	}
	protected function get_descendants()
	{
		$descendants = array();
		$queue = array($this);
		while ($child = reset($queue))
		{
			$descendants[] = $child->tabledef()->tablename();
			unset($queue[0]);
			$queue = array_merge($queue,array_values($child->children()));
		}
		return array_values(array_unique($descendants));
	}
	public function parent()
	{
		return $this->_parent;
	}
	public function child($tablename)
	{
		return isset($this->_children[$tablename])
			? $this->_children[$tablename]
			: NULL;
	}
	public function children()
	{
			return $this->_children;
	}
	# debug methods, internal
	public function debug_observer_list() // Observed method
	{
		return TableModel::debug_observer_list($this->_observers);
	}
	public function debug_child_name_list() 
	{
		$array = array();
		foreach ($this->_children as $child)
		{
			$child->get_result();
			$array[$child->observer_name()] = $child->debug_child_name_list();
		}
		if (!$array) 
		{
			return '(leaf)';
		}
		else 
		{
			return $array;
		}
	}
}
