<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster\Tablemodel;

use StdClass;
use Kohana, Kohana_Exception;
use Services;
use Tablemodel;

defined('SYSPATH') or die('No direct script access.');
/**
 file: tabledef.php
 package: Tablemodel
 category: ORM (Object Relational Manager)
 date: August 2010
* class: Tabledef
* table metadata
* contains relations to tables
* contains relations to index and relation definition objects (not the activerecords)
* contains back relation to tablemodel
API
	Tabledef:
		->tablealias() // schema table name
		->name() // same as tablealias()
		->tablename() // database table name
		->table($n=0)
		->activerecord($n=0)
		->recordset($n=0)
		->query(($n=0)
		->column_names()
		->alternate_index_column_names()
		->primary_key_column_name()
		->model()
*/
class Tabledef
{
	protected $_table_definition; // definition used to create the table object, 
		// temporarily held for relation definitions call  ->create_relation_definitions()

	protected $_tablename; //database table name
	protected $_tablemodel; // tablemodel object
	protected $_name; // model table name
	protected $_columnspecs; //associative array of column names and attributes
	protected $_column_names; // lazy loading; derived from keys of columnspecs on demand
	protected $_join_columns; // list array of join columns which cannot be updated
	protected $_base_table; // table to use to insert or delete
	protected $_primary_index_name; // name of one of the indexdefs
	protected $_alternate_index_name; // name of one of the indexdefs
	protected $_indexdefs = array(); // associative array of indexdef objects
	protected $_relationdefs = array(); // associative array of relationdef objects
	protected $_tables = array(); // list array; references to tables created with this def
	protected $_validation; // Kohana validation object
	protected $_audit; // audit columns: CreateUserID, CreateDate, EditUserID, EditDate; or FALSE
	protected $_concurrency; // concurrency columns: RecordVersion, TimeStamp; or FALSE
	protected $_perpage; // default perpage; or NULL [*** s/b FALSE?***]
	protected $_orderby; // default orderby; or NULL [*** s/b FALSE?***]

	public function __construct($parameters) 
	{
		// set properties - primary key property is set later after index object creation
		$this->_table_definition = $parameters;
		$this->_tablemodel = $parameters->tablemodel;
		$this->_name = $parameters->name;
		$this->_tablename = $parameters->tablename;
		// load column names and attributes for error checking, and for insert checking
		$this->_columnspecs = $this->_tablemodel->db()->list_columns($this->_tablename);
		$this->_validation = $parameters->validation;
		$this->_audit = $parameters->audit;
		$this->_concurrency = $parameters->concurrency;
		$this->_perpage = $parameters->perpage;
		$this->_orderby = $parameters->orderby;
		
		//create indexdefs
		$indexdefs = $parameters->indexes;
		foreach ($indexdefs as $indexdef => $definition) 
		{
			$definition = (object) $definition;
			$this->_validate_index_properties($indexdef,$definition);
			$this->_assert_index_properties($indexdef, $definition);
			$this->_indexdefs[$indexdef] = new Indexdef($definition);
		}
		$this->_validate_keys($parameters); // primary and alternate
		// set primary key property
		$this->_primary_index_name = $parameters->primary_index;
		// set alternate key property
		if (isset($parameters->alternate_index))
			$this->_alternate_index_name = $parameters->alternate_index;
		if (isset($parameters->joincolumns)) // columns that cannot be edited
			$this->_join_columns = $parameters->joincolumns;
		if (isset($parameters->basetable)) // direct deletion or insertion not allowed through view
			$this->_base_table = $parameters->basetable;
	}
	
	protected function _validate_index_properties($indexdef, $definition)
	{
		if ($this->_tablemodel->debug_level() > 0) 
		{
			// check for required parameters
			$result = Services::verify_properties
				(
					$definition,
					array
					(
						'columns',
						'index_type'
					)
				);
			if (!($result === TRUE))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','tabledef.missing_index_parameters'),
							array(':parameters' => implode(',',$result), 
							':index'=>$indexdef, 
							':tabledef' => $this->_name )
					);

			}
		}
	}
	
	protected function _assert_index_properties($indexdef, $definition)
	{
		Services::assert_properties
			(
				$definition,
				array
				(
					'name' => $indexdef,
					'tabledef' => $this
				)
			);
	}
	
	protected function _validate_keys($parameters)
	{
		if ($this->_tablemodel->debug_level() > 0) 
		{
			// check that named primary key index exists
			if (! isset($this->_indexdefs[$parameters->primary_index])) 
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','tabledef.missing_indexdef_parameters'),
							array(':parameter' => $parameters->primary_index,
							':table' => $this->_name)
					);

			}
			// check that named primary key is UNIQUE
			if ($this->_indexdefs[$parameters->primary_index]->index_type() != TableModel::UNIQUE) 
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','tabledef.mismatched_indextype_parameter'),
							array(':parameter'=>$parameters->primary_index,
							':table'=>$this->_name)
					);

			}
			if (isset($parameters->alternate_index))
			{
				if (! isset($this->_indexdefs[$parameters->alternate_index])) 
				{

					throw new Kohana_Exception
						(
							Kohana::message('tablemodel','tabledef.missing_alternate_parameters'),
								array(':parameter'=>$parameters->alternate_index,
								':table'=>$this->_name)
						);

				}	
				// check that named alternate key is UNIQUE

				if ($this->_indexdefs[$parameters->alternate_index]->index_type() != TableModel::UNIQUE) 
				{
					throw new Kohana_Exception
						(
						Kohana::message('tablemodel','tabledef.mismatched_alternatetype_parameter'),
								array(':parameter'=>$parameters->primary_index,
								':table'=>$this->_name)
						);
				}

			}
		}
	}
	
	// create relation definitions for the table (called from tablemodel, after all tables have been created from tabledef)
	public function create_relation_definitions() 
	{
		$tabledefinition = $this->_table_definition;
		unset($this->_table_definition); // no longer needed
		if (!property_exists($tabledefinition,'relations')) return; // no relations to define
		$relationdefinitions = $tabledefinition->relations;
		foreach ($relationdefinitions as $relation_name => $definition)
		{
			$definition = (object) $definition;
			$this->_validate_relation_properties($relation_name,$definition);
			$this->_assert_relation_properties($definition,$relation_name);
			$this->_relationdefs[$relation_name] = new Relationdef($definition);
		}
	}

	protected function _validate_relation_properties($relation_name,$definition)
	{
		if ($this->_tablemodel->debug_level() > 0) 
		{
			// check for parameters
			$result = Services::verify_properties
				(
					$definition,
					array
					(
						'index',
						'remote_table',
						'remote_index',
						'relation_type'
					)
				);
			if (!($result === TRUE))
			{
				throw new Kohana_Exception
					(
						Kohana::message('tablemodel','tabledef.missing_relationdef_parameters'),
							array(':parameters'=>implode(',',$result),
							':relation'=> $relation_name,
							':tabledef'=>$this->_name)
					);

			}
		}
	}
	protected function _assert_relation_properties($definition,$relation_name)
	{
		Services::assert_properties
			(
				$definition,
				array
				(
					'tabledef' => $this,
					'name' => $relation_name,
					'terminate_expansion' => NULL,
				)
			);
	}
	// set reciprocal relations for the table
	public function set_reciprocal_relationdefs()
	{
		foreach ($this->_relationdefs as $relation)
		{
			$relation->find_reciprocal_relationdef();
		}
	}
	//table, activerecord, and recordindexes
	// table constructor registers table with host tabledef
	// called from tablemodel
	public function create_table($parent = NULL, $relationdef = NULL)
	{
		$parameters = new StdClass;
		$parameters->tabledef = $this;
		$parameters->parent = $parent;
		$parameters->relationdef = $relationdef;
		$table = $this->_tables[] = new Table($parameters);
//		$table->create_components(); // called from table constructor
		return $table;
	}
	
	public function destroy_tables()
	{
		$tables = $this->_tables;
		foreach ($tables as $table)
		{
			$table->destroy();
		}
		$this->_tables = array();
	}
	
	// property accessors...
	public function tables()
	{
		return $this->_tables;
	}
	public function relationdefs()
	{
		return $this->_relationdefs;
	}
	public function indexdefs() 
	{
		return $this->_indexdefs;
	}
	public function alternate_index_name()
	{
		return $this->_alternate_index_name;
	}
	public function primary_index_name()
	{
		return $this->_primary_index_name;
	}
	public function columnspecs()
	{
		return $this->_columnspecs;
	}
	public function join_columns()
	{
		return $this->_join_columns;
	}
	public function base_table()
	{
		if ($this->_base_table)
			return $this->_base_table;
		else
			return $this->_tablename;
	}
	public function delete_table()
	{
		return $this->base_table();
	}
	public function insert_table()
	{
		return $this->base_table();
	}
	public function model()
	{
		return $this->_tablemodel;
	}
	public function audit()
	{
		return $this->_audit;
	}
	public function concurrency()
	{
		return $this->_concurrency;
	}
	public function perpage()
	{
		return $this->_perpage;
	}
	public function orderby()
	{
		return $this->_orderby;
	}
	// [--------------------------- TABLEDEF GENERAL API ---------------------------]
	public function tablealias() // model table name
	{
		return $this->_name;
	}
	public function name()
	{
		return $this->_name;
	}
	public function tablename() // database table name
	{
		return $this->_tablename;
	}
	public function primary_key_column_name() // must be single field
	{
		$column_names = $this->_indexdefs[$this->_primary_index_name]->column_names();
		if (count($column_names) <> 1)
			return NULL;
		else
			return $column_names[0];
	}
	public function alternate_index_column_names()
	{
		if (!isset($this->_alternate_index_name)) return NULL;
		return $this->_indexdefs[$this->_alternate_index_name]->column_names();
	}
	public function table($n=0)
	{
		return isset($this->_tables[$n])
			? $this->_tables[$n]
			: NULL;
	}
	public function activerecord($n=0)
	{
		return isset($this->_tables[$n])
			? $this->_tables[$n]->activerecord()
			: NULL;
	}
	public function query($n=0)
	{
		return isset($this->_tables[$n])
			? $this->_tables[$n]->query()
			: NULL;
	}
	public function recordset($n=0)
	{
		return isset($this->_tables[$n])
			? $this->_tables[$n]->recordset()
			: NULL;
	}
	public function column_names()
	{
		if (!$this->_column_names)
			$this->_column_names = array_keys($this->_columnspecs);
		return $this->_column_names;
	}
	public function validation($validation = NULL)
	{
		if (!is_null($validation))
			$this->_validation = $validation;
		return $this->_validation;
	}
}
