<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster;

use StdClass;
use Database, DB;

defined('SYSPATH') or die('No direct script access.');
/* 
	An implementation of Modified Preorder Tree Traversal (MPTT)
	This is a forest of trees. 
	LeftEdge and RightEdge are always interpreted in conjuntion with RootID.
	RootID is carried in all nodes.
	
	$parameters should be in the form:
	->db = <db object>
	->tablename = <string>
	->basetablename = <string> (for insert/delete);
	->columns = <object>
	columms object (all strings representing column names):
		->primarykeycolumn
		->leftedgecolumn
		->rightedgecolumn
		->parentcolumn
		->levelcolumn
		->defaultordercolumn
		->alphadirectioncolumn
		->datedirectioncolumn
		->childcountcolumn
		->alphabeticordercolumn
		->dateordercolumn
		->ismanualordercolumn
		->manualordercolumn
		->rootidcolumn
*/

// ** add validation methods
// ** rationalize exception handling with operation failure
class TreeManager {

	static $bindertree;
	static $foldertree;

	protected $_db;
	protected $_tablename;
	protected $_basetablename;
	protected $_columns;
	protected $_default_order = 'a';
	protected $_default_direction = 'a';
	protected $_operation; // last operation
	protected $_error_info = array(); // last error from last operation, possibly internal
	
	static function binder_tree()
	{
		if (!is_null(self::$bindertree)) return self::$bindertree;
		
		$parameters = new StdClass;
		$parameters->db = Database::instance('default');
		$parameters->tablename = 'Binders_Tree_View';
		$parameters->basetablename = 'Binders_Tree';
		$columns = $parameters->columns = new StdClass;
		{
			$columns->primarykeycolumn = 'NodeID';
			$columns->leftedgecolumn = 'LeftEdge';
			$columns->rightedgecolumn = 'RightEdge';
			$columns->parentcolumn = 'ParentID';
			$columns->levelcolumn = 'Level';
			$columns->defaultordercolumn = 'DefaultOrder';
			$columns->alphadirectioncolumn = 'AlphaDirection';
			$columns->datedirectioncolumn = 'DateDirection';
			$columns->childcountcolumn = 'ChildCount';
			$columns->alphabeticordercolumn = 'Title';
			$columns->dateordercolumn = 'Date';
			$columns->ismanualordercolumn = 'isOrdered';
			$columns->manualordercolumn = 'SiblingOrder';
			$columns->rootidcolumn = 'RootID';
		}
		
		$tree = self::$bindertree = new TreeManager($parameters);
		return $tree;
	}
	public function __construct($parameters)
	{
		$parameters = (object) $parameters;
		$columns = $this->columns = new StdClass;
		$this->_db = $parameters->db;
		$this->_tablename = $parameters->tablename;
		$this->_basetablename = $parameters->basetablename;;
		$this->_columns = $parameters->columns;
	}
	public function operation() {
		return $this->_operation;
	}
	public function error_info() {
		return $this->_error_info;
	}
	public function db() 
	{
		return $this->_db;
	}
	// order = 'a','d' (alphabetic, date); direction = 'a','d' (ASC, DESC)
	// for root which has no parent to specify sort parameters
	public function default_order($order,$direction='a')
	{
		$this->_default_order = $order;
		$this->_default_direction = $direction;
	}
	# locking
	public function lock()
	{
		return $this->_db->lock_tables(array($this->_tree_tablename));
	}
	public function unlock()
	{
		return $this->_db->unlock_tables();
	}
	# initialize
	// all roots are initialized
	public function initialize() 
	{
		$cursor = $this->get_roots();
		while ($record = $cursor->fetch()) {
			$result = $this->initialize_tree($record);
			if (!$result) return FALSE;
		}
		return TRUE;
	}
	// $node is a single root of a tree, no parentid
	public function initialize_tree($rootnode)
	{
		$level = 1;
		$edgeid = 0;
		$columns = $this->_columns;
		$rootid = $rootnode->{$columns->primarykeycolumn};
		$leftedge = ++$edgeid;
		$childcount = $this->_init_children($rootid,$rootnode->{$columns->primarykeycolumn},&$edgeid,$level);
		if ($childcount === FALSE) return FALSE;
		$result = $this->_set(
			array(
				$columns->leftedgecolumn => $leftedge,
				$columns->rightedgecolumn => ++$edgeid,
				$columns->levelcolumn => $level,
				$columns->childcountcolumn => $childcount,
				$columns->rootidcolumn => $rootid,
			),$columns->primarykeycolumn,'=',$rootnode->{$columns->primarykeycolumn});
		return $result;
	}
	protected function _init_children($rootid,$nodeid,&$edgeid,$level)
	{
		$columns = $this->_columns;
		$cursor = $this->get_children($nodeid); // sorted
		$recordcount = 0;
		$level++;
		while ($record = $cursor->fetch()) {
			$recordcount++;
			$leftedge = ++$edgeid;
			$childcount = $this->_init_children($rootid,$record->{$columns->primarykeycolumn},&$edgeid,$level);
			if ($childcount === FALSE) return FALSE;
			$result = $this->_set(
				array(
					$columns->leftedgecolumn => $leftedge,
					$columns->rightedgecolumn => ++$edgeid,
					$columns->levelcolumn => $level,
					$columns->childcountcolumn => $childcount,
					$columns->rootidcolumn => $rootid,
				),$columns->primarykeycolumn,'=',$record->{$columns->primarykeycolumn});
			if (!$result) return FALSE; 
		}
		return $recordcount;
	}
	# add
	public function add_root($node)
	{
		$record = $this->_insert($node);
		if ($record) {
			$result = $this->initialize_tree($record); // technical: self left and right edge, level, childcount
		}
		if ($record and $result)
			return $record;
		else
			return FALSE;
	}
	// with specified parent and sortfield data; use add_root for no parent
	// no children
	public function insert_node($node)
	{
		$columns = $this->_columns;
		if (is_null($node->{$columns->parentcolumn}))
			return $this->add_root($node);
		# insert the record
		$record = $this->_insert($node);
		if (!$record) return FALSE;
		# locate the record
		list($offset,$level) = $this->_get_offset_and_level($record);
		if ($offset === FALSE) return FALSE;
		# increment values to the right to make room
		$parentnode = $this->get_node($node->{$columns->parentcolumn});
		if (!$parentnode) return FALSE;
		$rootid = $parentnode->{$columns->rootidcolumn};
		$result = $this->_increment($rootid,$offset,2);
		if (!$result) return FALSE;
		# set the left and right edges of the new node
		$result = $this->_set(
			array($columns->leftedgecolumn => $offset+1,
				$columns->rightedgecolumn => $offset+2,
				$columns->levelcolumn => $level,
				$columns->rootidcolumn => $rootid,
			),
			$columns->primarykeycolumn,'=',$record->{$columns->primarykeycolumn});
		if (!$result) return FALSE;
		# increment the childcount of the parent
		$result = $this->_set(
			array($columns->childcountcolumn => $parentnode->{$columns->childcountcolumn} +1),
			$columns->primarykeycolumn,'=',$parentnode->{$columns->primarykeycolumn});
		if (!$result) return FALSE;
		return TRUE;
	}
	// does a sequential search of siblings (parent children)
	// assumes relatively few children...
	protected function _get_offset_and_level($record)
	{
		$columns = $this->_columns;
		$primarykey = $columns->primarykeycolumn;
		$nodeid = $record->$primarykey;
		$found = FALSE;
		$previousrecord = NULL;
		$cursor = $this->get_children($record->{$columns->parentcolumn}); // sorts appropriately
		while ($childrecord = $cursor->fetch()) {
			if ($childrecord->$primarykey == $nodeid) {
				$found = TRUE;
				break;
			}
			$previousrecord = $childrecord;
		}
		if (!$found) return array(FALSE,FALSE);
		# locate reference record
		if (!is_null($previousrecord)) {
			$offset = $previousrecord->{$columns->rightedgecolumn};
			$level = $previousrecord->{$columns->levelcolumn};
		} else {
			$nextrecord = $cursor->fetch();
			if (!is_null($nextrecord)) {
				$offset = $nextrecord->{$columns->leftedgecolumn} - 1;
				$level = $nextrecord->{$columns->levelcolumn};
			} else {
				$parentrecord = $this->get_node($record->{$columns->parentcolumn});
				if (!is_null($parentrecord)) {
					$offset = $parentrecord->{$columns->leftedgecolumn};
					$level = $parentrecord->{$columns->levelcolumn} + 1;
				} else {
					return array(FALSE,FALSE);
				}
			}
		}
		return array($offset,$level);
	}
	// save change, resort children if necessary, change sortorder if necessary
	public function update_node($node,$referencevalues)
	{
		$columns = $this->_columns;
		// get oldrecord for comparison
		$oldrecord = $this->get_node($node->{$columns->primarykeycolumn});
		if (!$oldrecord) return FALSE;
		$result = $this->_set((array) $node,$columns->primarykeycolumn,'=',$node->{$columns->primarykeycolumn});
		if (!$result) {
			$this->_error_info = array('update node','treemanager','update data failure',$this->_error_info);
			return FALSE;
		}
		else $newrecord = $this->get_node($node->{$columns->primarykeycolumn});
		
		if ($oldrecord->{$columns->parentcolumn} != $newrecord->{$columns->parentcolumn}) {
			$result = $this->_changed_parent($newrecord,$oldrecord->{$columns->parentcolumn}); // 1
			if (!$result) {
				$this->_error_info = array('update node','treemanager','change parent failure',$this->_error_info);
				return FALSE;
			}
			else $newrecord = $this->get_node($node->{$columns->primarykeycolumn});
		}
		// re-sort children
		if (($oldrecord->{$columns->defaultordercolumn} !== 
				$newrecord->{$columns->defaultordercolumn}) or
			(($oldrecord->{$columns->defaultordercolumn} == 'a') and 
			($oldrecord->{$columns->alphadirectioncolumn} !== 
				$newrecord->{$columns->alphadirectioncolumn})) or
			(($oldrecord->{$columns->defaultordercolumn} == 'd') and 
			($oldrecord->{$columns->datedirectioncolumn} !== 
				$newrecord->{$columns->datedirectioncolumn})) or
			($oldrecord->{$columns->ismanualordercolumn} !==
				($newrecord->{$columns->ismanualordercolumn}))) {
			$result = $this->_resort_children($newrecord); // 2
			if (!$result) {
				$this->_error_info = array('changesort','treemanager','change sort failure',$this->_error_info);
				return FALSE;
			}
		}
		// change in sibling order
		$parentid = $newrecord->{$columns->parentcolumn};
		if (!is_null($parentid)) {
			// ** check parent for manual order before checking change of manual order column
			if (($referencevalues->{$columns->manualordercolumn} !== 
					$newrecord->{$columns->manualordercolumn}) or
				($referencevalues->{$columns->alphabeticordercolumn} !=
					$newrecord->{$columns->alphabeticordercolumn}) or
				($referencevalues->{$columns->dateordercolumn} !=
					$newrecord->{$columns->dateordercolumn})) {
				$result = $this->_change_sort_position($newrecord); // 3
				if (!$result) {
					$this->_error_info = array('changesort','treemanager','change sort failure',$this->_error_info);
					return FALSE;
				}
			}
		}
		return TRUE;
	}
	# modify parent
	/*
	- record already has the new parent saved to the record
	- the purpose of this method is to update the presort data
	- branch has been pre-sorted; new parentid has been set by client
	- more or less resets the tree, 
	  backward from old position, forward from new position
	*/
	protected function _changed_parent($record,$oldparentid)
	{
		// collect data
		$columns = $this->_columns;
		if (!is_null($oldparentid)) {
			$oldparentnode = $this->get_node($oldparentid);
			if (!$oldparentnode) {
				$this->_error_info = array('changed parent','treemanager','get old parent node faulure: ' . $oldparentid);
				return FALSE;
			}
		}
		$leftedge = $record->{$columns->leftedgecolumn};
		$rightedge = $record->{$columns->rightedgecolumn};
		$oldlevel = $record->{$columns->levelcolumn};
		
		$newparentid = $record->{$columns->parentcolumn};
		if (!is_null($newparentid)) {
			// 1 - edge values made negative to isolate them
			$isolatevalue = $rightedge + 1;
			$result = $this->_isolate_branch($record,$isolatevalue); 
			if (!$result) {
				$this->_error_info = array('changed parent','treemanager','isolate branch failure');
				return FALSE;
			}
		}
		
		$span = (($rightedge - $leftedge)+ 1);
		$oldrootid = $record->{$columns->rootidcolumn};
		if (!is_null($oldparentid)) {
			// 2 - close gap for old position, and decrement childcount
			$result = $this->_increment($oldrootid,$rightedge,-$span);
			if (!$result) {
				$this->_error_info = array('changed parent','treemanager','close gap failure');
				return FALSE;
			}
			$result = $this->_set(array($columns->childcountcolumn => $oldparentnode->{$columns->childcountcolumn} - 1),
				$columns->primarykeycolumn,'=',$oldparentnode->{$columns->primarykeycolumn});
			if (!$result) {
				$this->_error_info = array('changed parent','treemanager','decrement childcount failure');
				return FALSE;
			}
		}
		
		// 3 - find new references
		if (is_null($newparentid)) {
			return $this->initialize_tree($record);
		} else {
			$parentnode = $this->get_node($newparentid);
			if (!$parentnode) {
				$this->_error_info = array('changed parent','treemanager','get parent node failure: '. $newparentid);
				return FALSE;
			}
			$newrootid = $parentnode->{$columns->rootidcolumn};
			list($newoffset,$newlevel) = $this->_get_offset_and_level($record);
			if ($newoffset === FALSE) {
				$this->_error_info = array('changed parent','treemanager','get offset and level failure');
				return FALSE;
			}
		}
		// 4 - open gap for new position, and increment childcount
		$result = $this->_increment($newrootid,$newoffset,$span);
		if (!$result) {
			$this->_error_info = array('changed parent','treemanager','open gap failure');
			return FALSE;
		}
		if ($parentnode) {
			$result = $this->_set(array($columns->childcountcolumn => $parentnode->{$columns->childcountcolumn} + 1),
				$columns->primarykeycolumn,'=',$parentnode->{$columns->primarykeycolumn});
			if (!$result) {
				$this->_error_info = array('changed parent','treemanager','increment childcount failure');
				return FALSE;
			}
		}
		
		// 5 - integrate new position
		$increment = $newoffset - ($leftedge - $isolatevalue) +1; // on isolated value
		$levelincrement = $newlevel - $oldlevel;
		$result = $this->_un_isolate_branch(
			$oldrootid,$leftedge-$isolatevalue,$rightedge-$isolatevalue,
			$newrootid,$increment,$levelincrement);
		if (!$result) {
			$this->_error_info = array('changed parent','treemanager','integrate new position');
			return FALSE;
		}
		return $result;
	}
	# modify order
	// no change in number of descendents, just a change of children order
	// brute force: this just resorts (re-initializes) all children (and descendents) of the parent
	protected function _resort_children($record)
	{
		$columns = $this->_columns;
		$edgeid = $record->{$columns->leftedgecolumn};
		$level = $record->{$columns->levelcolumn};
		$rootid = $record->{$columns->rootidcolumn};
		$recordid = $record->{$columns->primarykeycolumn};
		$result = $this->_init_children($rootid,$recordid,&$edgeid,&$level);
		if ($result === FALSE) return FALSE;
		else return TRUE;
	}
	// same as change_parent but with parent rightedge limit
	protected function _change_sort_position($record)
	{
		// collect data
		$columns = $this->_columns;
		$parentnode = $this->get_node($record->{$columns->parentcolumn});
		if (!$parentnode) {
			return TRUE;
		}
		$limit = $parentnode->{$columns->rightedgecolumn};
		$leftedge = $record->{$columns->leftedgecolumn};
		$rightedge = $record->{$columns->rightedgecolumn};
		$oldlevel = $record->{$columns->levelcolumn};
		
		// 1 - edge values made negative to isolate them
		$isolatevalue = $rightedge + 1;
		$result = $this->_isolate_branch($record,$isolatevalue); 
		if (!$result) {
			$this->_error_info = array('changesort','treemanager','isolate branch failure');
			return FALSE;
		}
		
		// 2 - close gap for old position, and decrement childcount
		$span = (($rightedge - $leftedge)+ 1);
		$oldrootid = $record->{$columns->rootidcolumn};
		$this->_increment($oldrootid,$rightedge,-$span,$limit);
		$result = $this->_set(array($columns->childcountcolumn => $parentnode->{$columns->childcountcolumn} - 1),
			$columns->primarykeycolumn,'=',$parentnode->{$columns->primarykeycolumn});
		if (!$result) {
			$this->_error_info = array('changesort','treemanager','close gap failure');
			return FALSE;
		}
		
		// 3 - find new references
		$parentid = $record->{$columns->parentcolumn};
		$parentnode = $this->get_node($parentid);
		if (!$parentnode) return FALSE;
		list($newoffset,$newlevel) = $this->_get_offset_and_level($record);
		if ($newoffset === FALSE) {
			$this->_error_info = array('changesort','treemanager','find reference failure');
			return FALSE;
		}
		$newrootid = $parentnode->{$columns->rootidcolumn};
		
		// 4 - open gap for new position, and increment childcount
		$result = $this->_increment($newrootid,$newoffset,$span,$limit);
		if (!$result) {
			$this->_error_info = array('changesort','treemanager','increment failure');
			return FALSE;
		}
		$result = $this->_set(array($columns->childcountcolumn => $parentnode->{$columns->childcountcolumn} + 1),
			$columns->primarykeycolumn,'=',$parentnode->{$columns->primarykeycolumn});
		if (!$result) {
			$this->_error_info = array('changesort','treemanager','increment childcount failure');
			return FALSE;
		}
		// 5 - integrate new position
		$increment = $newoffset - ($leftedge - $isolatevalue) +1; // on isolated value
		$levelincrement = $newlevel - $oldlevel;
		$result = $this->_un_isolate_branch(
			$oldrootid,$leftedge-$isolatevalue,$rightedge-$isolatevalue,
			$newrootid,$increment,$levelincrement);
		return $result;
	}
	# move
	// [[=========[ ** pending ]===============]]
	// null is the same as move_start
	public function move_insert($node,$rightnode=NULL)
	{
	}
	// [[=========[ ** pending ]===============]]
	// null is the same as move_end
	public function move_append($node,$leftnode=NULL)
	{
	}
	public function move_start($node)
	{
		return $this->move_insert($node);
	}
	public function move_end($node)
	{
		return $this->move_append($node);
	}
	# copy
	public function copy_branch($node)
	{
	}
	// [[=========[ ** pending ]===============]]
	# delete (cascades)
	// Node required solely for ID
	public function delete_node($node) // including children
	{
		$columns = $this->_columns;
		// delete node
		$operation = $this->_operation = DB::delete($this->_basetablename)
			->where($columns->primarykeycolumn,'=',$node->{$columns->primarykeycolumn})
			->prepare($this->_db)
			->execute();
		$result = $operation->success();
		if (!$result) {
			$this->_error_info = $operation->error_info();
			return FALSE;
		}
		// close gap of removed node for parent
		$rightedge = $node->{$columns->rightedgecolumn};
		$leftedge = $node->{$columns->leftedgecolumn};
		$increment = (($rightedge - $leftedge)+ 1);
		$rootid = $node->{$columns->rootidcolumn};
		$result = $this->_increment($rootid,$rightedge,-$increment);
		if (!$result) return FALSE;
		// decrement childcount of parent
		$parentid = $node->{$columns->parentcolumn};
		if ($parentid) {
			$parentnode = $this->get_node($parentid);
			if ($parentnode) {
				$result = $this->_set(
					array($columns->childcountcolumn => $parentnode->{$columns->childcountcolumn} - 1),
					$columns->primarykeycolumn,'=',$parentnode->{$columns->primarykeycolumn});
				if (!$result) return FALSE;
			} else {
				return FALSE;
			}
		}
		return TRUE;
	}
	# get
	// returns object
	public function get_node($id)
	{
		$query = DB::select('*')
			->from($this->_tablename)
			->as_object()
			->where($this->_columns->primarykeycolumn,'=',$id);
		$result = $query->prepare($this->_db)
			->execute();
		$record = $result->current(); // works with failure?? ***
		if (!$result->success()) $this->_error_info = $result->error_info();
		return $record;
	}
	// get tree parts, returns cursor
	public function get_roots()
	{
		$columns = $this->_columns;
		$defaultorder = $this->_default_order;
		$defaultdirection = $this->_default_direction;
		if ($defaultdirection == 'a') $direction = 'ASC';
			else $direction = 'DESC';
		if ($defaultorder == 'a') $ordercolumn = $columns->alphabeticordercolumn;
			else $ordercolumn = $columns->dateordercolumn;
		return $this->_query_children(NULL,array(array($ordercolumn,$direction)));
	}
	// returns cursor
	public function get_children($parentid) // sorted by sortfield
	{
		$columns = $this->_columns;
		if (is_null($parentid)) return NULL;
		$parentnode = $this->get_node($parentid);
		if (!$parentnode) return FALSE;
		$orderlist = array();
		$defaultorder = $parentnode->{$columns->defaultordercolumn};
		if ($defaultorder == 'a') $defaultdirection = $parentnode->{$columns->alphadirectioncolumn};
			else $defaultdirection = $parentnode->{$columns->datedirectioncolumn};
		$ismanualorder = $parentnode->{$columns->ismanualordercolumn};
		if ($ismanualorder) {
			$orderlist[] = array($columns->manualordercolumn,'ASC');
		}
		if ($defaultorder == 'a') $ordercolumn = $columns->alphabeticordercolumn;
			else $ordercolumn = $columns->dateordercolumn;
		if ($defaultdirection == 'a') $direction = 'ASC';
			else $direction = 'DESC';
		$orderlist[] = array($ordercolumn,$direction);
		return $this->_query_children($parentid,$orderlist);
	}
	// returns array of objects
	public function get_branch($id)
	{
		$node = $this->get_node($id);
		if (!$node) return FALSE;
		$columns = $this->_columns;
		$leftedge = $node->{$columns->leftedgecolumn};
		$rightedge = $node->{$columns->rightedgecolumn};
		$rootid = $node->{$columns->rootidcolumn};
		$branch = DB::select('*')
			->from($this->_tablename)
			->where($columns->leftedgecolumn,'BETWEEN',DB::expr($leftedge.' AND '.$rightedge))
			->where($columns->rootidcolumn,'=',$rootid)
			->order_by($columns->leftedgecolumn)
			->as_object()
			->prepare($this->_db)
			->execute();
		if ($branch)
			$branch = $branch->as_array();
		return $branch;
	}
/*
For get_path:
Note: probably better to do this using parentID to build path, but this is more interesting.
from: http://www.artfulsoftware.com/mysqlbook/sampler/mysqled1ch20.html listing 15
SELECT
	t2.NodeID AS Node,
	t2.Title AS Title,
	(SELECT COUNT(*)
	FROM Binders_Tree_View AS t4
	WHERE t4.leftedge BETWEEN t1.leftedge AND t1.rightedge
		AND t2.leftedge BETWEEN t4.leftedge AND t4.rightedge
   ) AS Path
FROM Binders_Tree_View AS t1
	INNER JOIN Binders_Tree_View AS t2 ON t2.leftedge BETWEEN t1.leftedge AND t1.rightedge
	INNER JOIN Binders_Tree_View AS t3 ON t3.leftedge BETWEEN t2.leftedge AND t2.rightedge
WHERE t1.nodeID=<rootid>
	AND t3.nodeID=<targetid>;

- modified for forest of trees with rootid; count removed, order_by added
Generated Example:

SELECT `t2`.`NodeID` AS `NodeID`, 
	`t2`.`Title` AS `Title` 
FROM Binders_Tree_View as t1 
	INNER JOIN Binders_Tree_View as t2 
		ON (`t2`.`leftedge` BETWEEN t1.leftedge AND t1.rightedge 
		AND `t2`.`RootID` = `t1`.`RootID`) 
	INNER JOIN Binders_Tree_View as t3 
		ON (`t3`.`leftedge` BETWEEN t2.leftedge AND t2.rightedge 
		AND `t3`.`RootID` = `t2`.`RootID`) 
WHERE `t1`.`nodeid` = '142' 
	AND `t3`.`nodeid` = 177 
ORDER BY `t2`.`LeftEdge`
*/
	public function get_path($id)
	{
		$node = $this->get_node($id); // for RootID
		if (!$node) return NULL;
		$result = DB::select(
			array('t2.NodeID','node'),array('t2.Title','title'),array('t2.isBox','isbox'))
			->from(DB::expr('Binders_Tree_View as t1'))
			->join(DB::expr('Binders_Tree_View as t2'),'INNER')
			->on('t2.leftedge','BETWEEN',DB::expr('t1.leftedge AND t1.rightedge'))
			->on('t2.RootID','=','t1.RootID')
			->join(DB::expr('Binders_Tree_View as t3'),'INNER')
			->on('t3.leftedge','BETWEEN',DB::expr('t2.leftedge AND t2.rightedge'))
			->on('t3.RootID','=','t2.RootID')
			->where('t1.nodeid','=',(int) $node->RootID)
			->where('t3.nodeid','=',$id)
			->order_by('t2.LeftEdge')
			->prepare($this->_db)
			->execute();
		$path = $result->as_array();
		foreach ($path as &$item) {
			$item['isbox'] = (bool) $item['isbox'];
		}
		return $path;
	}
	# operations
	// return a set of records as a cursor
	protected function _query_children($parentid,$orderlist = NULL)
	{
		// rationalize op and parentid
		if (is_null($parentid)) {
			$op = 'IS';
			$parentid = DB::expr('NULL');
		} else {
			$op = '=';
		}
		// set selection criteria
		$parentcolumn = $this->_columns->parentcolumn;
		$query = DB::select('*')
			->from($this->_tablename)
			->as_object()
			->where($parentcolumn,$op,$parentid);
		// order
		if ($orderlist) {
			foreach ($orderlist as $order) {
				list($ordercolumn,$orderdirection) = $order;
				$query->order_by($ordercolumn,$orderdirection);
			}
		} // end of order
		return $query->prepare($this->_db)
			->execute();
	}
	// returns saved record
	protected function _insert($node)
	{
		$nodearray = (array) $node;
		$fields = array_keys($nodearray);
		$values = array_values($nodearray);
		$operation = $this->_operation = DB::insert($this->_basetablename,$fields)
			->values($values)
			->prepare($this->_db)
			->execute();
		if ($operation->success()) {
			$id = $node->{$this->_columns->primarykeycolumn};
			$record = $this->get_node($id);
			return $record;
		} else {
			$this->_error_info = $operation->error_info();
			return FALSE;
		}
	}
	// increments branch edge values, returns boolean success
	protected function _increment($rootid,$offset,$increment,$limit = NULL)
	{
		$columns = $this->_columns;
		$leftedge = $columns->leftedgecolumn;
		$rightedge = $columns->rightedgecolumn;
		$rootidcolumn = $columns->rootidcolumn;
		
		// leftedge
		$query = DB::update($this->_tablename)
			->set(array($leftedge=>DB::expr('`'.$leftedge.'` + '.$increment)))
			->where($leftedge,'>',$offset)
			->where($rootidcolumn,'=',$rootid);
		if (!is_null($limit)) {
			$query->where($leftedge,'<=',$limit);
		}
		$operation = $this->_operation = 
			$query->prepare($this->_db)
				->execute();
		if (!$operation->success()) {
			$this->_error_info = $operation->error_info();
			return FALSE;
		}
		
		// rightedge
		$query = DB::update($this->_tablename)
			->set(array($rightedge=>DB::expr('`'.$rightedge.'` + '.$increment)))
			->where($rightedge,'>',$offset)
			->where($rootidcolumn,'=',$rootid);
		if (!is_null($limit)) {
			$query->where($rightedge,'<=',$limit);
		}
		$operation = $this->_operation = $query->prepare($this->_db)
			->execute();
			
		$result = $operation->success();
		if (!$result) $this->_error_info = $operation->error_info();
		return $result;
	}
	// returns success (true or false)
	protected function _set($assignmentarray,$wherefield,$op,$value)
	{
		$operation = $this->_operation = DB::update($this->_tablename)
		->set($assignmentarray)
		->where($wherefield,$op,$value)
		->prepare($this->_db)
		->execute();
		$result = $operation->success();
		if (!$result)
			$this->_error_info = $operation->error_info();
		return $result;
	}
	// temporarily sets leftedge and rightedge to negative numbers
	protected function _isolate_branch($node,$isolatevalue) // isolatevalue always 1 more than right edge
	{
		$columns = $this->_columns;
		$leftedge = $columns->leftedgecolumn;
		$rightedge = $columns->rightedgecolumn;
		$rootid = $columns->rootidcolumn;
		
		$operation = $this->_operation = DB::update($this->_tablename)
			->set(array($leftedge=>DB::expr('`'.$leftedge.'` - '.$isolatevalue),$rightedge=>DB::expr('`'.$rightedge.'` - '.$isolatevalue)))
			->where($leftedge,'BETWEEN',DB::expr($node->$leftedge.' AND '.$node->$rightedge))
			->where($rootid,'=',$node->$rootid)
			->prepare($this->_db)
			->execute();
		$result = $operation->success();
		if (!$result) $this->_error_info = $operation->error_info();
		return $result;
	}
	// integrates node into new position, after isolate:
	// level increment, rootid update, increment to fit into gap
	protected function _un_isolate_branch(
		$oldrootid,$isolatedleftedge,$isolatedrightedge,
		$newrootid,$increment,$levelincrement = 0)
	{
		if ($levelincrement != 0)
			$result = $this->_increment_level($oldrootid,$isolatedleftedge,$isolatedrightedge,$levelincrement);
		if ($oldrootid != $newrootid)
			$result = $this->_update_rootid($oldrootid,$isolatedleftedge,$isolatedrightedge,$newrootid);
		$result = $this->_increment($newrootid,$isolatedleftedge-1,$increment,$isolatedrightedge);
		return $result;
	}
	// update rootid of moved branch
	protected function _update_rootid($oldrootid,$leftedge,$rightedge,$newrootid)
	{
		$columns = $this->_columns;
		$leftedgecolumn = $columns->leftedgecolumn;
		$rootidcolumn = $columns->rootidcolumn;
		$operation = DB::update($this->_tablename)
			->set(array($rootidcolumn => $newrootid))
			->where($leftedgecolumn,'BETWEEN',DB::expr($leftedge.' AND '.$rightedge))
			->where($rootidcolumn,'=',$oldrootid)
			->prepare($this->_db)
			->execute();
		$result = $operation->success();
		if (!$result) {
			$this->_error_info = $operation->error_info();
			return FALSE;
		}
		return $result;
	}
	// increment level of branch nodes
	protected function _increment_level($rootid,$leftedge,$rightedge,$levelincrement)
	{
		$columns = $this->_columns;
		$leftedgecolumn = $columns->leftedgecolumn;
		$levelcolumn = $columns->levelcolumn;
		
		$operation = $this->_operation = DB::update($this->_tablename)
		->set(array($levelcolumn => DB::expr('`'.$levelcolumn.'` + '.$levelincrement)))
		->where($leftedgecolumn,'BETWEEN',DB::expr($leftedge.' AND '.$rightedge))
		->where($columns->rootidcolumn,'=',$rootid)
		->prepare($this->_db)
		->execute();
		$result = $operation->success();
		if (!$result) {
			$this->_error_info = $operation->error_info();
			return FALSE;
		}
		return $result;
	}
}
