<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Muster;

use StdClass;
use Kohana, Session, Cookie, Request, Text, DB;
use System, Services, TableModel;

defined('SYSPATH') or die('No direct script access.');

/***
DEVELOPMENT NOTES:

- support moderated subscriptions
- review use of RecordVersion in this module

***/

/***
Class: User
Description: This class is designed to hold the functionality required to enforce user authentication
Acknowledgement: originally (2008) loosely based on Authentication model created by Kohana's lead Woody Gilk (Shadowhand): Auth.php

General API:
	__construct // tries to auto_login()
static:
	UserAccount::instance()

methods:
	// operations
	login($username, $password, $remember = FALSE)
	logout($destroy = FALSE)
	
	// property accessors
	authenticated()
	table_model()
	roleid_array() // for accessfilter
	user_data()
	status_code()
	error_message($status_code = NULL)
	error_info()
	userid()
	==========
***/
class User {
	const SALT_LENGTH = 32; // encryption of passwords
	
	// constants ending in _DB have error_info set for them
	const LOGIN_SUCCESS = 0;
	const LOGIN_NO_SUCH_USERID = 1;
	const LOGIN_NO_USERID = 2;
	const LOGIN_COMPLETE_LOGIN_FAILURE = 3;
	const LOGIN_TOKEN_RECORD_NOT_FOUND = 4;
	const LOGIN_TOKEN_EXPIRED = 5;
	const LOGIN_INVALID_TOKEN = 6;
	const LOGIN_USER_NOT_REGISTERED = 7;
	const LOGIN_TOKEN_ADD_RECORD_FAILURE = 8;
	const LOGIN_TOKEN_SAVE_RECORD_FAILURE = 9;
	const LOGIN_SESSION_EXPIRED = 10;
	const LOGIN_WRONG_PASSWORD = 11;
	const LOGIN_SAVE_RECORD_FAILURE = 12;
	const LOGIN_NOT_ACTIVE = 13;
	
	protected $_user_columns = array( // subset likely required for login and general administration
		'ID','Code','Name','Email','MembershipExpires','LastLogin','Logins',
		'Expiry','Password','TimeZone','StatusID','RoleID','Permissions',
		'CreateDate','CreateUserID','EditDate','EditUserID','RecordVersion','BlogID','InboxID');
	
	// should be in i18n...
	protected $_error_messages = array
	(
		0 => 'Operation was successful',
		self::LOGIN_NO_USERID => 'No such User ID',
		self::LOGIN_COMPLETE_LOGIN_FAILURE => 'System error: Unable to complete login',
		self::LOGIN_TOKEN_RECORD_NOT_FOUND => 'System error: Token record not found',
		self::LOGIN_TOKEN_EXPIRED => 'Login Token expired',
		self::LOGIN_INVALID_TOKEN => 'Invalid Token',
		self::LOGIN_USER_NOT_REGISTERED => 'User not registered',
		self::LOGIN_TOKEN_ADD_RECORD_FAILURE => 'System error: Add Token record failed',
		self::LOGIN_TOKEN_SAVE_RECORD_FAILURE => 'System error: Save Token record failed',
		self::LOGIN_SESSION_EXPIRED => 'Session expired',
		self::LOGIN_WRONG_PASSWORD => 'Wrong password',
		self::LOGIN_SAVE_RECORD_FAILURE => 'Failure to save record',
		self::LOGIN_NOT_ACTIVE => 'User registration is not active',

	);
	protected static $_instance;
	
	protected $_session;
	protected $_authenticated = FALSE;
	protected $_cookie_lifetime; // from config, in seconds
	protected $_now; // from clock
	protected $_error_code;
	protected $_userid = NULL;
	protected $_error_info;
	protected $_system;
	protected $_table_model;
	protected $_user_status;
	protected $_db;

	public function __construct($parameters = array())
	{
		$parameters = (array) Services::assert_properties
		(
			$parameters,
			array(
				'mastertable' => 'users'
			)
		);
		$config = Kohana::$config->load('userauth');
		$this->_cookie_lifetime = $config['cookielifetime'];
		$this->_session = Session::instance();
		$this->_now = time();
		self::$_instance = $this;
		$model = $this->_table_model = new TableModel('userschema',$parameters);
		$this->_db = $model->db();
		$this->_system = System::instance();
		$this->auto_login();
	}
	
	public static function instance($parameters = array())
	{
		if (!self::$_instance)
		{
			return new User($parameters = array());
		}
		return self::$_instance;
	}
	public function model()
	{
		return $this->_table_model;
	}
	protected function auto_login()
	{
		$session = $this->_session;
		
		if (mt_rand(1, 100) === 1) // done early in process to avoid mismatch; one out of a hundred connections clear out dead tokens
		{
			// Do garbage collection of tokens
			$this->delete_expired_tokens();
		}
		// get userid
		$userid = $session->get_once('userid');
		$token = NULL;
		if (!$userid)
		{
			$token = Cookie::get('authautotoken'); // if user specified 'remember me'
			// validate token
			if ($token)
			{
				$userid = $this->get_userid_from_token($token);
				if (!$userid) 
				{
					$token = NULL;
					Cookie::delete('authautotoken');
				}
			}
		}
		// verify userid
		if ($userid)
		{
			$table = $this->_table_model->table();
			$table->reset()->query()->select_array($this->_user_columns)->where('ID','=',$userid);
			if (!$table->activerecord()->current())
			{
				$this->_error_code = self::LOGIN_NO_SUCH_USERID;
				$userid = NULL;
			}
			else
			{
				if ($token) // used to get userid
				{
					if (!$this->validate_token($token))
					{
						$userid = NULL;
					}
				}
			}
		}
		if ($userid)
		{
			$validated = $this->validate_user_status();
		}
		else
		{
			$validated = FALSE;
			$this->_error_code = self::LOGIN_NO_USERID;
		}
		if ($validated and $token) // almost authenticated, must record login as last step
		{
			$this->_userid = $userid;
			$validated = $this->update_logindata();
		}
		if ($validated)
		{
			$this->complete_login();
			$this->_error_code = self::LOGIN_SUCCESS;
		} else {
			$this->_user_status = 'Guest';
		}
	}
	
	private function update_logindata()
	{
		$cursor = $this->_table_model->activerecord();
		// Update the number of logins
		$cursor->Logins += 1;
		// Set the last login date
		$cursor->LastLogin = Services::get_sql_datetime(time());
//		$cursor->RecordVersion = $cursor->RecordVersion; // for concurrency
		// Save the user record
		$this->_error_info = NULL;
		if (!$cursor->save())
		{
			$this->_error_info = $cursor->error_info();
			$this->_error_code = self::LOGIN_COMPLETE_LOGIN_FAILURE;
			return FALSE;
		}
		return TRUE;
	}

	public function login($username, $password, $remember = FALSE)
	{
		$table = $this->_table_model->table();
		$table->reset()->query()->select_array($this->_user_columns)->where('Code','=',$username);
		$cursor = $this->_table_model->activerecord();
 		if (!$cursor->fetch()) // check for user record
		{
			$this->_error_code = self::LOGIN_USER_NOT_REGISTERED;
			return FALSE; // user is not registered
		}
		// check password
		$userpassword = $cursor->Password;
		if ($userpassword != $this->hash_password($password,$userpassword))
		{
			$this->_error_code = self::LOGIN_WRONG_PASSWORD;
			return FALSE;
		}
		// validate user status
		$validated = $this->validate_user_status();
		if (!$validated) return FALSE;
		// set or remove token as called for
		$token_cursor = $this->_table_model->activerecord('sessiontokens');
		$token_cursor->current();
		if ($remember === TRUE)
		{
			// Create a new autologin token
			if (!$token_cursor->has_record())
			{
				if (!$token_cursor->add())
				{
					$this->_error_code = self::LOGIN_TOKEN_ADD_RECORD_FAILURE;
					return FALSE;
				}
				$token_cursor->Code = $username;
				$token_cursor->UserAgent = sha1(Request::$user_agent);
				$token_cursor->Created = Services::get_sql_datetime(time());
				$token_cursor->RecordVersion = 0;
			}
			$token_cursor->Expires = Services::get_sql_datetime(time() + $this->_cookie_lifetime);
			// lock table here
			$token = $this->get_new_token();
			$token_cursor->Token = $token;
//			$token_cursor->Recordversion = $token_cursor->RecordVersion; // concurrency
			$this->_error_info = NULL;
			if (!$token_cursor->save())
			{
				$this->_error_info = $token_cursor->error_info();
				$this->_error_code = self::LOGIN_TOKEN_SAVE_RECORD_FAILURE;
				return FALSE;
			}
			// unlock table here
			// Set the autologin cookie
			Cookie::set('authautotoken', $token, $this->_cookie_lifetime);
		}
		else
		{
			if ($token_cursor->has_record())
			{
				$token_cursor->delete();
				Cookie::delete('authautotoken');
			}
		}
		// Finish the login
		if (!$this->update_logindata()) return FALSE;
		$this->complete_login();
		$this->_error_code = self::LOGIN_SUCCESS;
		return TRUE;
	}
	
	protected function validate_token($token)
	{
		if ($token)
		{
			$token_cursor = $this->_table_model->activerecord('sessiontokens');
			if (!$token_cursor->current()) // shouldn't happen by now - rare chance of admin deleting token
			{
				die('TOKEN RECORD NOT FOUND');
				$this->_error_code = self::LOGIN_TOKEN_RECORD_NOT_FOUND;
				return FALSE; //something happened to the token in the meantime, must log in again
			}
			if (Services::get_php_datetime($token_cursor->Expires) < $this->_now)
			{
				$token_cursor->delete();
				$this->_error_code = self::LOGIN_TOKEN_EXPIRED;
				return FALSE; // must login again
			}
			if ($token_cursor->Token === $token)
			{
				if ($token_cursor->UserAgent === sha1(Request::$user_agent))
				{
					// get a new unique token
					$token_cursor->Token = $this->get_new_token();
//					$token_cursor->RecordVersion = $token_cursor->RecordVersion; // concurrency
					// Set the new token
					$this->_error_info = NULL;
					if (!$token_cursor->save())
					{
						$this->_error_info = $token_cursor->error_info();
						// Automatic login was successful
						$this->_error_code = self::LOGIN_TOKEN_SAVE_RECORD_FAILURE;
						return FALSE;
					}
					else
					{
						Cookie::set('authautotoken', $token_cursor->Token, $this->_cookie_lifetime);
						$this->_error_code = self::LOGIN_SUCCESS;
						return TRUE;
					}
				}
			}
			$token_cursor->delete();
			$this->_error_code = self::LOGIN_INVALID_TOKEN;
			return FALSE;
		}
		else // no token
		{
			$this->_error_code = self::LOGIN_INVALID_TOKEN;
			return FALSE;
		}
	}
	// assumes cursor is set to current user
	protected function validate_user_status()
	{
		$cursor = $this->_table_model->activerecord()->load();
		$statusID = $cursor->StatusID;
		$status = $this->_user_status = Services::lookup('Access_Lookup_User_Status',$statusID);
		if ($status != 'Active')
		{
			$this->_error_code = self::LOGIN_NOT_ACTIVE;
			return FALSE;
		}
		return TRUE;
	}

	protected function complete_login()
	{
		$userid = $this->_table_model->activerecord()->ID;
		$this->_session->set('userid',$userid);
		$this->_userid = $userid;
		$this->model()->userid($userid);
		$this->_authenticated = TRUE;
		return TRUE;
	}

	// generates a salted password if no salt provided, or creates one from salt prefix if salt (ie stored password) provided
	// taken from http://phpsec.org/articles/2005/password-hashing.html
	public function hash_password($password, $salt = NULL)
	{
		if ($salt === NULL)
		{
			$salt = substr(md5(uniqid(rand(), true)), 0, self::SALT_LENGTH);
		}
		else
		{
			$salt = substr($salt, 0, self::SALT_LENGTH);
		}
		return $salt . sha1($salt . $password);
	}

	// infinitely small chance that same token will be found before this one is used to create a record, 
	// 	but posting would fail in that case
	// ** can lock table for guarantee
	protected function get_new_token()
	{
		while (TRUE)
		{
			// Create a random token
			$token = Text::random('alnum', 32);

			// Make sure the token does not already exist
			if (DB::select()
				->where('Token','=',$token)
				->from('Access_User_Session_Tokens')
				->execute($this->_db)
				->count() === 0)
			{
				// A unique token has been found
				return $token;
			}
		}
	}
	
	protected function delete_expired_tokens()
	{
		// Delete all expired tokens
		DB::delete('Access_User_Session_Tokens')
			->where('Expires','<',$this->_now)
			->execute($this->_db);

		return $this;
	}
	
	protected function get_userid_from_token($token)
	{
		$resultset = DB::
			select('UserID')
			->from('Access_User_Session_Tokens')
			->where('Token','=',$token)
			->as_object()
			->execute($this->_db);
		$row = $resultset->current();
		if ($row)
		{
			return $row->UserID;
		}
		else
		{
			return NULL;
		}
	}
	public function logout($destroy = FALSE)
	{
		// Delete the autologin cookie if it exists
		Cookie::delete('authautotoken');
		$this->_table_model->activerecord('sessiontokens')->load()->delete();
		$session = $this->_session;
		if ($destroy === TRUE)
		{
			// Destroy the session completely
			$session->destroy();
		}
		else
		{
			// Remove the user object from the session
			$session->delete('userid');

			// Regenerate session_id
			$session->regenerate();
		}
		$this->_userid = NULL;

		// Double check
		return TRUE;
	}
	public function error_code()
	{
		return $this->_error_code;
	}
	public function error_message($status_code = NULL)
	{
		if ($status_code === NULL)
		{
			$status_code = $this->_error_code;
		}
		return $this->_error_messages[$status_code];
	}
	public function authenticated()
	{
		return $this->_authenticated;
	}
	public function error_info()
	{
		return $this->_error_info;
	}
	public function user_data()
	{
		$system = System::instance();
		if ($this->authenticated())
		{
			$record = clone $this->_table_model->activerecord()->current();
			unset (
				$record->Password,
				$record->MembershipExpires,
				$record->Expiry
			);
			$blogid = $record->BlogID;
			if (!is_null($blogid)) {
				$blogcode = Services::dblookup('Articles_Extensions','BlogCode','ArticleID',$blogid);
				$record->BlogCode = $blogcode;
			} else {
				$record->BlogCode = NULL;
			}
			if ($record->Permissions) {
				$record->Permissions = new JsonField(json_decode($record->Permissions));
			}
			if (!$system->MultipleGroups) {
				$roleid = $record->RoleID;
				if (is_null($roleid)) {
					$record->RoleID = $system->DefaultRoleID;
					$record->GroupID = $system->DefaultGroupID;
				} else {
					$record->GroupID = Services::dblookup(
						'Access_Group_Roles','GroupID','ID',$roleid);
				}
			}
		}
		else
		{
			$record = new StdClass;
			$record->ID = NULL;
			if (!$system->MultipleGroups) {
				$record->RoleID = $system->PublicRoleID;
				$record->GroupID = $system->PublicGroupID;
			}
		}
		return $record;
	}
	public function userid()
	{
		return $this->_userid;
	}
	public function user_status()
	{
		return $this->_user_status;
	}
	public function db()
	{
		return $this->_db;
	}
	// *** identical to account rank; should be superclass
	public function rank()
	{
		$owneraccount = $this->model()->activerecord();
		if (!$owneraccount->has_current()) return NULL;
		
		$ownerroleid = $owneraccount->RoleID;
		if (is_null($ownerroleid)) $ownerroleid = System::instance()->DefaultRoleID;
		if (!is_null($ownerroleid)) {
			$record = DB::select('Rank')
				->from('Access_Group_Roles')
				->where('ID','=',$ownerroleid)
				->as_object()
				->execute($this->_db)
				->current();
			if ($record) {
				$rank = (integer) $record->Rank;
			} else {
				$rank = 0;
			}
		} else {
			$rank = 0;
		}
		return $rank;
	}
}
