<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Base\NodeType;

use StdClass;
use Kohana, DB, Session;
use Services;
use Muster\Node\ListNode;
use Base\Node;

defined('SYSPATH') or die('No direct script access.');

class AttachmentAdminList extends ListNode {
	public function __construct($controller)
	{
		$this->_context = $this->_get_context();
		parent::__construct($controller);
	}
	protected function _get_context()
	{
		$context = new StdClass;
		# resources
		$context->tablemodel = 'attachmentsadminschema';
		$context->dataframe = 'attachmentsadmindata';
		$context->controlframe = 'attachmentsadmincontrols';
		$context->messagefile = 'attachmentsadmin';
		# table model
		$context->mastertablename = 'master';
		# recordset
		$context->selectarray = array('AttachmentID','Order','OrderTextRaw','DividerAbove','DividerBelow','InstanceDescription','BlogCode');
		$context->binderscope = NULL;
		$context->foreignkeyname = 'ParentArticleID';
		$context->default_orderby = 'Order';
		# gateway
		$context->mastertableindex = 'attachmentsadmintable';
		$context->masterrecordindex = 'attachmentsadminrecord';
		# node
		$context->get_admindictionary = FALSE;
		$context->get_requestdata = FALSE;
		$context->redirect = 'sublists/attachments/admin';
		$context->datapoolarrays = Kohana::$config->load('recordpartials')->getArrayCopy();
		# security
		$context->tokens = Services::get_token_set('mixed');
		return $context;
	}
	protected function _on_get_dictionary(&$dictionary)
	{
		$parameters = $this->_parameters;
		if ((!empty($parameters->listview)) and ($parameters->listview == 'narrative')) {
			$dictionary['listview'] = 'narrative';
			$dictionary['is_narrativeview'] = TRUE;
		}
		$dictionary['articleid'] = $this->_context->articleid;
	}
	// @returns constant Node::FLOW_CONTINUE, or Node::COMPLETE
	protected function _on_respond($command)
	{ 
		$result = parent::_on_respond($command);
		$datamodel = $this->_datamodel;
		switch ($command) {
			case 'addattachments':
				$result = $this->_addattachments($datamodel); // folder
				break;
			case 'removeattachments':
				$result = $this->_removeattachments($datamodel); // folder
				break;
			case 'saveattachments':
				$result = $this->_saveattachments($datamodel); // folder
				break;
			case 'renumber':
				$result = $this->_renumbercontents($datamodel);
				break;
		}
		return $result;
	}
	protected function _addattachments($datamodel)
	{
		$permissions = $this->_permissions;
		if (!($permissions->canedit))
			return $this->_record_report_access_error($datamodel);
		$context = $this->_context;
		$articleid = $context->articleid;
		$parameters = $this->_parameters;
		$db = $this->_tablemodel->model->db();
		$datapool = $datamodel->datapool;
		if ($context->frozen) {
			$error_info = 'master record is frozen for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		if (!Services::check_article_permission($articleid)) {
			$error_info = 'insufficient permissions for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		$attachmentid = $parameters->newattachmentid;
		if (!$attachmentid and !is_numeric($attachmentid)) {
			$error_info = 'unspecified attachment id';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		$db->lock_tables(array('Articles_Attachments_View'));
		$objecttype = Services::dblookup('Articles','ObjectType','ID',$articleid,$db);
		if (empty($objecttype)) { // invalid recordcode
			$error_info = 'unknown article';
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		$attachmentobjecttype = Services::dblookup('Articles','ObjectType','ID',$attachmentid,$db);
		if (in_array($attachmentobjecttype,array('page','inbox','comment'))) { // invalid objecttype
			$error_info = 'invalid attachment type (' . $attachmentobjecttype . ')';
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		if (empty($attachmentobjecttype)) { // invalid recordcode
			$error_info = 'unknown attachment type';
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		$attachmenttitle = Services::dblookup('Articles','Title','ID',$attachmentid,$db);
		$operation = DB::insert('Articles_Attachments',array('ArticleID','AttachmentID'))
			->values(array($articleid,$attachmentid))
			->prepare()
			->execute();
		$result = $operation->success();
		if (!$result) {
			$error_info = $operation->error_info();
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE; // recycle
		}
		$db->unlock_tables();
		$session = Session::instance();
		$messages = array(
			'messages' => array(
				array('message'=> '[' . $attachmenttitle . '] has been attached to this article')
			),
			'index'=>'messagesection',
		);
		$session->set('listmessages',$messages);
		return Node::FLOW_REDIRECT;
	}
	protected function _removeattachments($datamodel)
	{
		$permissions = $this->_permissions;
		if (!($permissions->canedit))
			return $this->_record_report_access_error($datamodel);
		$parameters = $this->_parameters;
		$datapool = $datamodel->datapool;
		$removearray = $parameters->recordinstances;
		$context = $this->_context;
		$articleid = $context->articleid;
		if ($context->frozen) {
			$error_info = 'master record is frozen for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		if (!Services::check_article_permission($articleid)) {
			$error_info = 'insufficient permissions for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		if (!empty($removearray)) {
			$removearray = array_keys($removearray);
			$operation = db::delete('Articles_Attachments')
				->where('ID','IN',$removearray)
				->prepare()
				->execute();
			$result = $operation->success();
			if (!$result) {
				$error_info = $operation->error_info();
				$this->_set_datapool_error(compact('datapool','error_info'));
				return Node::FLOW_CONTINUE;
			}
			$session = session::instance();
			$messages = array(
				'messages' => array(
					array('message'=> 'Attachments were removed')
				),
				'index'=>'messagesection',
			);
			$session->set('listmessages',$messages);
			return Node::FLOW_REDIRECT;
		} else {
			$error_info = 'no attachments were selected for removal';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		return Node::FLOW_REDIRECT;
	}
	protected function _renumbercontents($datamodel)
	{
		$permissions = $this->_permissions;
		if (!($permissions->canedit))
			return $this->_record_report_access_error($datamodel);
		$tablemodel = $this->_tablemodel;
		$context = $this->_context;
		$db = $tablemodel->model->db();
		$db->lock_tables(array('Articles_Attachments_View'));
		$articleid = $context->articleid;
		if ($context->frozen) {
			$error_info = 'master record is frozen for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		if (!Services::check_article_permission($articleid)) {
			$error_info = 'insufficient permissions for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		if (is_null($articleid)) die('unexpected null articleid in renumbercontents');
		try {
			$operation = DB::select('ID')
				->from('Articles_Attachments_View')
				->where('ParentArticleID','=',$articleid)
				->where('Order','IS',DB::expr('NOT NULL'))
				->order_by('Order','DESC')
				->order_by('Title','DESC')
				->as_object()
				->prepare($db)
				->execute();
			if (!$operation->success()) {
				$db->unlock_tables();
				$messagesection = (object) array(
					'errors' => array(array('error' => 'failure to find records for sort')),
				);
				$datamodel->datapool->variables['messagesection'] = $messagesection;
				return Node::FLOW_CONTINUE;
			}
			$count = $operation->row_count();
			while ($record = $operation->fetch()) {
				if (!is_null($record->ID)) {
					$update = DB::update('Articles_Attachments')
						->set(array('Order'=>$count))
						->where('ArticleID','=',$articleid)
						->where('AttachmentID','=',$record->ID)
						->prepare($db)
						->execute();
					if (!$update->success()) {
						$db->unlock_tables();
						$messagesection = (object) array(
							'errors' => array(array('error' => 'unexpected failure to update record order')),
						);
						$datamodel->datapool->variables['messagesection'] = $messagesection;
						return Node::FLOW_CONTINUE;
					}
				}
				$count--;
			}
		} catch (Exception $e) {}
		$db->unlock_tables();
		$session = Session::instance();
		$messagearray = array(
			'messages' => array(array('message'=>'items have been renumbered')),
			'index'=>'messagesection',
		);
		$session->set('listmessages',$messagearray);
		return Node::FLOW_REDIRECT;
	}
	protected function _saveattachments($datamodel)
	{
		$permissions = $this->_permissions;
		if (!($permissions->canedit))
			return $this->_record_report_access_error($datamodel);
		$parameters = $this->_parameters;
		$is_narrativeview = ($parameters->listview == 'narrative');
		$context = $this->_context;
		$articleid = $context->articleid;
		$datapool = $datamodel->datapool;
		if ($context->frozen) {
			$error_info = 'master record is frozen for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		if (!Services::check_article_permission($articleid)) {
			$error_info = 'insufficient permissions for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		$recordset = $parameters->recordset;
		$db = $this->_tablemodel->model->db();
		$errors = array();
		$warnings = array();
		$posted = 'records ';
		$db->lock_tables(array('Articles_Attachments_View'));
		$is_errors = FALSE;
		foreach ($recordset as $record) {
			$instanceid = $record['InstanceID'];
			if (!is_numeric($instanceid)) continue; // being careful
			unset($record['InstanceID']);
			$nodeid = $record['NodeID'];
			unset($record['NodeID']);
			foreach ($record as &$value) {
				$value = trim($value);
				if ($value === '') $value = NULL;
			}
			unset($value);
			if (!$is_narrativeview) { // order data
				$record['DividerAbove'] = empty($record['DividerAbove'])?FALSE:TRUE;
				$record['DividerBelow'] = empty($record['DividerBelow'])?FALSE:TRUE;
				if (!is_null($record['Order']) and !is_numeric($record['Order'])) {
					$record['Order'] = 0;
					$warnings[] = array('warning' => 'Order for [' . $nodeid . '] was not numeric - changed to "0"');
				}
				if (strlen($record['OrderText']) > 60) {
					$record['OrderText'] = substr($record['OrderText'],0,60);
					$warnings[] = array('warning' => 'Order Text for [' . $nodeid . '] was truncated to 60 characters');
				}
			} else { // narrative data
				if (strlen($record['InstanceDescription']) > 255) {
					$record['InstanceDescription'] = substr($record['InstanceDescription'],0,255);
					$warnings[] = array('warning' => 'Narrative for [' . $nodeid . '] was truncated to 255 characters');
				}
			}
			$operation = DB::update('Articles_Attachments')
				->set($record)
				->where('ID','=',$instanceid)
				->prepare($db)
				->execute();
			$success = $operation->success();
			if (!$success) {
				$is_errors = TRUE;
				$err = $operation->error_info(); $err = $err[2];
				$message = '[' . $nodeid . '] ' . $err;
				$errors[] = array('error' => $message);
			} else {
//				$posted .= '[' . $nodeid . '] ';
			}
		} // end of foreach $record
		$db->unlock_tables();
		if (!empty($posted))
			$messages = array(array('message' => $posted . 'were updated'));
		else
			$messages = array();
		if ($is_errors) {
			$messagesection = (object) array(
				'messages' => $messages,
				'errors' => $errors,
				'warnings' => $warnings,
			);
			$datamodel->datapool->variables['messagesection'] = $messagesection;
			return Node::FLOW_CONTINUE;
		} else { // redirect
			$session = Session::instance();
			$messagearray = array(
				'messages' => $messages,
				'warnings' => $warnings,
				'index'=>'messagesection',
			);
			$session->set('listmessages',$messagearray);
			return Node::FLOW_REDIRECT;
		}
	}
}