<?php 
// Muster Software Copyright (c) Henrik Bechmann, Toronto, Canada 2012. All rights reserved.
// See "musterlicence.txt" for licencing information.
// mustersoftware.net

namespace Base\NodeType;

use StdClass;
use Kohana, DB, Session;
use Services;
use Muster\Node\ListNode;
use Base\Node;

defined('SYSPATH') or die('No direct script access.');

class ContainerList extends ListNode {
	public function __construct($controller,$selector = NULL)
	{
		$this->_context = $this->_get_context($selector);
		parent::__construct($controller);
	}
	protected function _get_context($selector)
	{
		$context = new StdClass;
		# resources
		$context->tablemodel = 'nodecontainerschema';
		$context->dataframe = 'containerdata';
		$context->controlframe = 'containercontrols';
		$context->messagefile = 'containers';
		# tablemodel
		$context->mastertablename = 'containerstable';
		# recordset
		$context->selectarray = array('InstanceID');
		$context->binderscope = NULL;
		$context->foreignkeyname = 'NodeID';
		$context->default_orderby = 'Title';
		$context->default_direction = 'ASC';
		# gateway
		$context->mastertableindex = 'containerstable';
		$context->masterrecordindex = 'containersrecord';
		$context->datapoolarrays = Kohana::$config->load('recordpartials')->getArrayCopy();
		# node
		$context->get_admindictionary = FALSE;
		$context->get_requestdata = FALSE;
		$context->redirect = 'sublists/containers/admin';
		# security
		$context->tokens = Services::get_token_set('mixed'); 
		return $context;
	}
	protected function _on_get_dictionary(&$dictionary)
	{
		$dictionary['articleid'] = $this->_context->articleid;
	}
	// @returns constant Node::FLOW_CONTINUE, or Node::FLOW_REDIRECT
	protected function _on_respond($command)
	{ 
		$result = parent::_on_respond($command);
		$datamodel = $this->_datamodel;
		switch ($command) {
			case 'addcontainer':
				$result = $this->_record_addcontainer($datamodel); // folder
				break;
			case 'removecontainers':
				$result = $this->_record_removecontainers($datamodel); // folder
				break;
		}
		return $result;
	}
	protected function _record_addcontainer($datamodel)
	{
		$permissions = $this->_permissions;
		if (!$permissions->canedit)
			return $this->_record_report_access_error($datamodel);
		$context = $this->_context;
		$parameters = $this->_parameters;
		$db = $this->_tablemodel->model->db();
		$datapool = $datamodel->datapool;
		$articleid = $context->articleid;
		$containerid = $parameters->newcontainerid;
		
		if (!$containerid and !is_numeric($containerid)) {
			$error_info = 'unspecified container id';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		$db->lock_tables(array('Nodes_Instances_View','Access_Tokens','Access_Role_Tokens','Articles_Authors'));
		$objecttype = Services::dblookup('Articles','ObjectType','ID',$articleid,$db);
		if (empty($objecttype)) { // invalid recordcode
			$error_info = 'unknown article';
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		$containerrecord = Services::dbrecordlookup('Articles',array('ObjectType','Frozen','Title'),'ID',$containerid,$db);
		if (empty($containerrecord)) { // invalid recordcode
			$error_info = 'unknown container';
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		$containerobjecttype = $containerrecord->ObjectType;
		$containerfrozen = (bool) $containerrecord->Frozen;
		if ($containerfrozen) { 
			$error_info = 'container is frozen for editing';
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		$canedit = Services::check_article_permission($containerid);
		if (!$canedit) {
			$error_info = 'insufficient permission for editing target container';
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		$containertitle = $containerrecord->Title;
		$allowedcontainers = Kohana::$config->load('allowedcontainers')->getArrayCopy();
		$allowedcontainers = $allowedcontainers[$objecttype];
		if (!in_array($containerobjecttype,$allowedcontainers)) {
			$error_info = 'article cannot be placed in container type: ' . $containerobjecttype;
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE;
		}
		$operation = DB::insert('Nodes_Instances',array('FolderID','NodeID'))
			->values(array($containerid,$articleid))
			->prepare()
			->execute();
		$result = $operation->success();
		if (!$result) {
			$error_info = $operation->error_info();
			$this->_set_datapool_error(compact('datapool','error_info'));
			$db->unlock_tables();
			return Node::FLOW_CONTINUE; // recycle
		}
		$db->unlock_tables();
		$session = Session::instance();
		$messages = array(
			'messages' => array(
				array('message'=> 'Article has been assigned to the container [' . $containertitle . ']')
			),
			'index'=>'messagesection',
		);
		$session->set('listmessages',$messages);
		return Node::FLOW_REDIRECT;
	}
	protected function _record_removecontainers($datamodel)
	{
		$permissions = $this->_permissions;
		if (!$permissions->canedit)
			return $this->_record_report_access_error($datamodel);
		$parameters = $this->_parameters;
		$datapool = $datamodel->datapool;
		$removearray = $parameters->recordinstances;
		if (!empty($removearray)) {
			$db = $this->_tablemodel->model->db();
			$db->lock_tables(array('Nodes_Instances_View','Articles_Authors','Access_Tokens','Access_Role_Tokens'));
			try {
				$removelist = array_keys($removearray);
				$removearray = array();
				// filter removelist items for validity
				foreach ($removelist as $removeitem)
				{
					// get remove container id
					$removecontainerid = Services::dblookup('Nodes_Instances','FolderID','ID',$removeitem);
					// verify it's valid
					if (!is_numeric($removecontainerid)) continue;
					// check if it's frozen
					$frozen = (bool) Services::dblookup('Articles','Frozen','ID',$removecontainerid);
					if ($frozen) continue;
					// check if it's editable
					$editable = Services::get_article_permission($removecontainerid);
					if (!$editable) continue;
					// otherwise add it to list to be removed
					$removearray[] = $removeitem;
				}
				if (count($removearray)<count($removelist))
					$messageaddendum = '; some containers could not be removed for being frozen, or lack of permission';
				$operation = db::delete('Nodes_Instances')
					->where('ID','IN',$removearray)
					->prepare()
					->execute();
				$result = $operation->success();
				if (!$result) {
					$error_info = $operation->error_info();
					$this->_set_datapool_error(compact('datapool','error_info'));
					return Node::FLOW_CONTINUE;
				}
			} catch (Exception $e) {
				$error_info = $e->getMessage();
				$this->_set_datapool_error(compact('datapool','error_info'));
				return Node::FLOW_CONTINUE;
			}
			$db->unlock_tables();
			$session = session::instance();
			$message = 'Containers were removed';
			if (!empty($messageaddendum)) $message.=$messageaddendum;
			$messages = array(
				'messages' => array(
					array('message'=> $message)
				),
				'index'=>'messagesection',
			);
			$session->set('listmessages',$messages);
			return Node::FLOW_REDIRECT;
		} else {
			$error_info = 'no containers were selected for removal';
			$this->_set_datapool_error(compact('datapool','error_info'));
			return Node::FLOW_CONTINUE;
		}
		return Node::FLOW_REDIRECT;
	}
}